// -*- C++ -*-

// MiscPipes.h
//
// Classes for sources and sinks of data
// Various DataPipes.
// for Audio Pattern Recognition Language (aprl)
// dpwe@media.mit.edu 1995feb02
//
/* $Header: /n/crab/da/dpwe/src/aprl/MiscPipes.h,v 1.1 1996/11/18 18:42:29 dpwe Exp $ @media.mit.edu Machine Listening Group
 */

#ifndef _MISCPIPES_H
#define _MISCPIPES_H

#include "DataPipe.h"
#include <math.h>	// for rint() below

class RandSrc : public DataSource {
private:
    int blksize;
    double lasttime;
    double scale;
public:
    RandSrc(double srate=0, int bufsz = 0) : DataSource(srate, bufsz)
        { blksize = bufsz; lasttime = 0; scale = 1.0; };
    virtual ~RandSrc(void)
        { };
    int Update(double /* time */);	// generate some random samples
    double getScale(void)	{ return scale; };
    void setScale(double sc)	{ scale = sc; };
};

class SquareProc : public DATAprocess {
public:
    SquareProc(void)	{ };
    virtual ~SquareProc(void)	{ };
    int process(DATA *input, int istride, int niItems, 
		DATA *output, int ostride, int noItems, 
		int *poutdone);
};

class SquarePipe : public DataPipeProcess {	// copy & square
private:
public:
    SquarePipe(DataSource *in = 0) : DataPipeProcess(in)
        { procFn = new SquareProc(); }
    virtual ~SquarePipe(void)
        { delete procFn; }
    int Update(double t)
    { return DataPipeProcess::Update(t); }
};

class RootProc : public DATAprocess {
public:
    RootProc(void)	{ };
    virtual ~RootProc(void)	{ };
    int process(DATA *input, int istride, int niItems, 
		DATA *output, int ostride, int noItems, 
		int *poutdone);
};

class RootPipe : public DataPipeProcess {	// copy & take sqroot
private:
public:
    RootPipe(DataSource *in = 0) : DataPipeProcess(in)
        { procFn = new RootProc(); }
    virtual ~RootPipe(void)
        { delete procFn; }
    int Update(double t)
    { return DataPipeProcess::Update(t); }
};

class StraightPipe : public DataPipeProcess {	// simple pipe copies i to o
private:
public:
    StraightPipe(DataSource *in = 0) : DataPipeProcess(in)
        { procFn = new DATAcopyClass(); }
    virtual ~StraightPipe(void)
        { delete procFn; }
    int Update(double t)
    {  CircularBuffer::print("StPipe"); return DataPipeProcess::Update(t); }
};

class DummySink : public DataSink {	// dummy sink prints data
public:
    DummySink(DataSource *in=0) : DataSink(in)	{ };
    virtual ~DummySink(void)			{ };
    int Update(double /* time */);
};

#ifndef PI
#define PI (3.1415926535)
#endif /* !PI */

class RmsProc : public DATAprocess {
private:
    CircularBuffer *store;
    CircularBuffer *win;
    int WinLen;
    int Step;
public:
    RmsProc(int winlen, int step);
    virtual ~RmsProc(void);
    int process(DATA *input, int istride, int niItems, 
		DATA *output, int ostride, int noItems, 
		int *poutdone);
};

class RmsPipe : public DataPipeProcess {	// copy & take sqroot
private:
    double opSampleRate;

    void init()  { int decim=max(1,(int)(rint(Input()->SampleRate()/opSampleRate)));
	      procFn = new RmsProc(2*decim, decim); SetIoRatio(decim); }
public:
    RmsPipe(double sr, DataSource *in = 0) 
	: opSampleRate(sr), DataPipeProcess(in)
        { SetSampleRate(opSampleRate); }
    virtual ~RmsPipe(void)
        { if(procFn)  delete procFn; }
    void SetInput(DataSource *in) 
    { DataPipeProcess::SetInput(in);
      if(procFn == 0 && Input())
	  init();  }
    int Update(double t)
    { if(procFn == 0 && Input())
	  init();
      return DataPipeProcess::Update(t); }
};

#define TWENTYONLOG10 8.6858896

class DBProc : public DATAprocess {	// convert a value to dB
private:
    double	ref;	// value to map to 0 dB
public:
    DBProc(double r)	: ref(r)
        { };
    virtual ~DBProc(void)
        { };
    int process(DATA *in, int is, int ni, DATA *out, int os, int no, int *pod)
        { int j, todo = min(ni, no);
	  double e = TWENTYONLOG10*log(ref);
	  for(j = 0; j<todo; ++j)  {
	      *out = TWENTYONLOG10*log(*in)-e;
	      out += os; in += is;
	  }
	  *pod = todo; return todo; }
    double &Ref(void)	{ return ref; }
};

class DBPipe : public DataPipeProcess {	// copy & take sqroot
private:
public:
    DBPipe(double ref, DataSource *in = 0) 
	: DataPipeProcess(in)
        { procFn = new DBProc(ref); }
    virtual ~DBPipe(void)
        { delete procFn; }
    void SetRef(double r)	{ ((DBProc *)procFn)->Ref() = r; }
};

class CompProc : public DATAprocess {	// Compare to a const, return 1 or 0
private:
    double	ref;	// if i<ref, o = 0, else o = 1
public:
    CompProc(double r)	: ref(r)
        { };
    virtual ~CompProc(void)
        { };
    int process(DATA *in, int is, int ni, DATA *out, int os, int no, int *pod)
        { int j, todo = min(ni, no);
	  for(j = 0; j<todo; ++j)  {
	      *out = (*in >= ref);
	      out += os; in += is;
	  }
	  *pod = todo; return todo; }
    double &Ref(void)	{ return ref; }
};

class CompPipe : public DataPipeProcess {	// copy & take sqroot
private:
public:
    CompPipe(double ref, DataSource *in = 0) 
	: DataPipeProcess(in)
        { procFn = new CompProc(ref); }
    virtual ~CompPipe(void)
        { delete procFn; }
    void SetRef(double r)	{ ((CompProc *)procFn)->Ref() = r; }
    double GetRef(void)		{ return ((CompProc *)procFn)->Ref(); }
};

class NotProc : public DATAprocess {	// Logical not
private:
    double	ref;	// if i<ref, o = 0, else o = 1
public:
    NotProc(double r=0)	: ref(r)
        { };
    virtual ~NotProc(void)
        { };
    int process(DATA *in, int is, int ni, DATA *out, int os, int no, int *pod)
        { int j, todo = min(ni, no);
	  for(j = 0; j<todo; ++j)  {
	      *out = (*in <= ref);
	      out += os; in += is;
	  }
	  *pod = todo; return todo; }
    double &Ref(void)	{ return ref; }
};

class NotPipe : public DataPipeProcess {	// copy & take sqroot
private:
public:
    NotPipe(double ref, DataSource *in = 0) 
	: DataPipeProcess(in)
        { procFn = new NotProc(ref); }
    virtual ~NotPipe(void)
        { delete procFn; }
    void SetRef(double r)	{ ((NotProc *)procFn)->Ref() = r; }
    double GetRef(void)		{ return ((NotProc *)procFn)->Ref(); }
};

class DelayPipe : public DataPipe {	// copy input to output after delay
private:
    CircularBuffer	*delayBuffer;
    double		delayTime;
    int 		started;
    void init(DATA v = 0)
        { if(Input())  {
	    int delaySamps = max(1, (int)(delayTime * SampleRate()));
	    int olddlysmps = delayBuffer->Size();
	    delayBuffer->Resize(delaySamps); 
	    if(delaySamps > olddlysmps) // make sure it stays full
		delayBuffer->PushVals(delaySamps-olddlysmps, v); }  }
public:
    DelayPipe(double dlyTime, DATA initval = 0, DataSource *in=0)
	: DataPipe(in), delayTime(dlyTime)
	{  delayBuffer = new CircularBuffer(0); init(initval); started = 0; }
    virtual ~DelayPipe(void)
        {  if(delayBuffer)	delete delayBuffer; }
    void SetDelay(double delay)
        { delayTime = delay; init(); }
    void SetInput(DataSource *in)
        { DataPipe::SetInput(in);  init(); }
    double Delay(void) const 	
        { return delayTime; }
    int Update(double /* time */);
};

#endif /* _MISCPIPES_H */
