// -*- C++ -*-

//
// Ptrack.h
//
// billg's pitch tracker as C++
//
// dpwe 1995feb24
//

#define FMT_SHORT   0	/* 16-bit integers */
#define FMT_FLOAT   1	/* 32-bit floats */
#define FMT_ASCII   2	/* sample, period/pitch pairs */

#include "Xing.h"

class ptrack : public xing {   // ptrack is a subclass of xing, 
                               // so it inherits all of xing etc.
private:

    // were static
/*====================================================================
 * Acquire mode globals.
 *
 * These are the delta times between adjacent events, and combinations
 * of delta times. This is the set of delta times needed for the three
 * state acquisition mode described in the paper.
 */
    long d1;
    long d2;
    long d1p2;
    long d3;
    long d2p3;
    
/*
 * state is the current state of the tracker:
 *
 *  -1	track mode
 *   0	acquire mode, waiting for first delta time
 *   1	acquire mode, waiting for second delta time
 *   2	acquire mode, waiting for third delta time
 *   3	acquire mode, waiting for fourth delta time
 */
    int state;
    double compare_ratio;	/* comparison threshold between delta times */
    double compare_avg;		/* average of last two deltas compared */

/*====================================================================
 * Tracking mode globals.
 */
    double period;	    	/* current period estimate (samples) */
    long cur_delta;	    	/* delta since last tracked event (samples) */
    long last_delta;     	/* last delta time for rough estimate */
    int n_events;	    	/* number events since last tracked event */
    long n_periods;	    	/* number of periods tracked */
    long dropout_timer;  	/* countdown after lvg track mode (samples) */
    double last_period;  	/* last tracked period estimate (samples) */
    int track_periods;   	/* number tracked periods before valid pitch */
    int track_flag;	    	/* TRUE when we achieve track_periods */
    int pitch_flag;	    	/* TRUE when track_flag TRUE or in dropouts */

    // were global
    double compare_cents;	/* above in cents for user interface */
    double track_gain;   	/* averaging gain for period tracking */
    int max_events;	    	/* max events bfr new tracked ev. found */
    int ignore_d1pd2;   	/* TRUE if ignoring d1+d2,d2+d3 comparisons */
    int dropout_periods; 	/* number of periods allowed for a dropout */
    int track_per_ui;    	/*	= track_periods + 2 */

    long total;			/* total number of samples */
/*
 * These parameters are needed if using ptrack_file() for processing.
 */
    FILE *inp_fp;		/* file pointer to input file */
    FILE *out_fp;		/* file pointer to output file */

    int output_fmt;		/* output format, defined below */
    double output_scale; 	/* pitch/period scaling if short output */
    int output_pitch;		/* TRUE if output pitch, else period */

    // private fns
    void init_acquire(void);
    void set_period(double period_param);
    void init_track(double new_period);
    void pitch_on(void);
    void pitch_off(void);
    void track_on(void);
    void track_off(void);
    void proc_track(long delta);
    int  compare(long a, long b);
    int  compare_entry(int index, int state);
    void proc_acquire(long delta);
    void open_files(char *inp_file, char *out_file);

    void new_period_fn(void);

protected:
    void accept_fn(long delta);		// called when new xing accepted 
/*
 * Callback functions for note on and off triggers. 
 */
    virtual void note_on(void)        {};
    virtual void note_off(void)       {};

public:
    ptrack(long samprate);
    virtual ~ptrack(void);

    void default_params(void);
    void speech_params(void);
    void music_params(void);
    void query_params(void);

    void proc_samp(short samp);
    double get_period(void);
    double get_pitch(void);
    void close_files(void);
    void proc_files(char *inp_file, char *out_file);

    // accessors for relevent state
    double& Track_gain(void)	{ return track_gain; };   	
    /* averaging gain for period tracking */
    int& Max_events(void)	{ return max_events; };	    	
    /* max events bfr new tracked ev. found */
    int& Ignore_d1pd2(void)	{ return ignore_d1pd2; };   	
    /* TRUE if ignoring d1+d2,d2+d3 comparisons */
    int& Dropout_periods(void) 	{ return dropout_periods; }; 	
    /* number of periods allowed for a dropout */

    void SetOutputParams(int op_fmt, int op_pitch, double op_scale)
        { output_fmt = op_fmt; output_pitch = op_pitch; 
	  output_scale = op_scale; };
    void SetCompareCents(double compCs);
    void SetTrackPerUI(int tks);
};

