// -*- C++ -*-

// Scheduler.cc
//
// Register and run different processes
// for Audio Pattern Recognition Language (aprl)
// dpwe@media.mit.edu 1995feb02
//

#include <stdio.h>
#include "Scheduler.h"

// declare static data field for SchedProc
int SchedProc::numberRegistered = 0;	// counter of elements made

// --------------- SchedProcList - list of SchedProcs -----------------------
SchedProcList:: ~SchedProcList(void)
{   /* destroy all the nodes in the list */
    SchedProcNode *n, *dpn = head;

    while(dpn) {
	n = dpn->next;
	delete dpn;
	dpn = n;
    }
}

void SchedProcList::insert(SchedProc *proc)
{   /* Insert new SchedProc* into list with a new node.
       Position in list depends on its sequence value, as extracted by 
       overloaded operator<(SchedProc&, SchedProc&) */
    SchedProcNode *n, **pdpn = &head;
    int done = 0;

    while(!done && *pdpn)  { 
	if( *proc < *((*pdpn)->data) )  {
	    n = new SchedProcNode(proc, *pdpn);
	    *pdpn = n;
	    done  = 1;
	} else {
	    pdpn = &((*pdpn)->next);
	}
    }
    if(!done)  {
	n = new SchedProcNode(proc);
	     *pdpn = n;
    }
}

int SchedProcList::remove(SchedProc *proc)
{   /* Search down the list, find and delete the node whose contents 
       is <proc>, close up the list.  Report error to stderr and return 0
       if node not found, else return 1. */
    SchedProcNode *n, **pdpn = &head;
    int done = 0;

    while(!done && *pdpn)  { 
	if( proc == ((*pdpn)->data) )  {
	    n = *pdpn;
	    *pdpn = (*pdpn)->next;
	    n->next = 0;
	    delete n;
	    done  = 1;
	} else {
	    pdpn = &((*pdpn)->next);
	}
    }
    if(!done)  {
	fprintf(stderr, "SchedProcList::remove: not found\n");
	return 0;
    }
    return 1;
}

void SchedProcList::apply(SchedProcFn *dpf, void *dpfdata)
{   /* Map the function to the contents of every node in the list.
       <dpfdata> is a private pointer passed to the client function. */
    SchedProcNode *dpn = head;

    while(dpn)  {
	(*dpf)(dpn->data, dpfdata);
	dpn = dpn->next;
    }
}

// ----------------------------- Scheduler itself ----------------------

// define MAPpable function to call update on each schedule list element
static void DPFupdateFn(SchedProc *dp, void *data)
{
    double time = *(double *)data;
    dp->Update(time);
}

void Scheduler::Update(double time)
{   /* Call Update method for every proc on the list */
    apply(DPFupdateFn, (void *)&time);
}

int Scheduler::Enschedule(SchedProc *proc)	
{   /* add a process to the schedule.  Return 1 on success, -1 on 
       failure because the process is marked as already belonging to 
       a schedule */
    if(proc->sched != 0) { 
	fprintf(stderr, "Enschedule: already scheduled\n"); 
	return -1; 
    }
/*    fprintf(stderr, "Got %lx to enschedule\n", proc);  */
    insert(proc); 	// Inserts by overloaded <, >, ==
    proc->sched = this; 
    return 1; 
}

int Scheduler::Deschedule(SchedProc *proc)
{   /* Remove the specified process from our list, mark it as 
       unscheduled.  Assuming we find it, return 1.  -1 indicates 
       that the process couldn't be found in our schedule list. */
    if(remove(proc)>=0)  { 
	proc->sched = 0; return 1; 
    } else { 
	return -1;
    }
}

int Scheduler::Reschedule(SchedProc *proc)
{   /* suggestion is that proc may have chged its sequence, so re-insert it */
    int rc = -1;
    if(Deschedule(proc)>0)  {
	Enschedule(proc);
	rc = 1;
    }
    return rc;
}

#ifdef MAIN	// ------------------------- TEST CODE ----------------------

class DummyProc : public SchedProc {
private:
    const char *name;
public:
    DummyProc(const char *tag)	{ name = tag; }
    ~DummyProc(void)		{  }
    int Update(double time)
        { fprintf(stderr, "Dummyproc '%s' @%.6f\n", name, time);
	  return 0; }
};

void testSched(void)
{
    DummyProc dp1("one");
    DummyProc dp2("two");
    DummyProc dp3("three");

    Scheduler sc;

    sc.Enschedule(&dp1);
    sc.Enschedule(&dp3);
    sc.Enschedule(&dp2);

    fprintf(stderr, "Update 0\n");
    sc.Update(0);
    dp1.SetSequence(1.5);	// change its order in the sequence
    fprintf(stderr, "Update 1\n");
    sc.Update(1);
    sc.Deschedule(&dp1);
    fprintf(stderr, "Update 2\n");
    sc.Update(2);

}

void main(int argc, char **argv)
{
    testSched();
}

#endif /* MAIN */
