#!./aprlwish
#
# clapper.tcl
# 
# Demonstrate deriving events from the input stream.  
#
# dpwe@media.mit.edu 1995feb14
#

# Enable shell listener
set tcl_interactive 1
# Make window resizable
#wm minsize . 1 1

# Set up the lightbulb icon
label .l -bitmap @lightbulb.xbm
pack  .l -side right

# Initialize the scheduler; each 's step' will process 0.1 sec of sound
scheduler s -timestep 0.1

# Frequency for the BP filter
set frq 3000

# frame for dynamic displays
frame .f -relief groove -bd 2

# VU meter on left-hand edge (displays from -50 to 0 (dB))
vumeter .vu -width 32 -orient v -relief sunken -scale 0.02 -offset 1
pack .vu -side left -fill y -expand y -padx 5 -pady 5

# Set up the audio processing chain.  The audio input buffer determines 
#   the maximum latency of the input, also how well it deals with 
#   interruptions (e.g. from other programs)
audiosrc  as    -sampleRate 8000 -bufferTime 0.2
# Alternatively, I can take input from a file, and play it at the same time
# This is useful for interactive debugging
##soundfsrc  as    -filename "claps.aif"
##audiosink  .

# We're looking for claps - choose a high frequency band
bpfilter  f1    -centerFreq $frq
rms       .     -sampleRate 100
dB        d0    -ref 0.01
# Have one display of the smoothed energy in dB
wavesink  .f.ws1  -scale 0.05 -height 80 \
	-relief sunken -padX 4 -padY 4

# The 'clap' detector samples the energy E(t) at three times (i.e. the 
#   current signal and two delayed versions) and calculates the following 
#   expression:  
#        ( E(t) < -20dB ) & ( E(t-0.1) > 0 dB ) & ( E(t-0.15) < -20dB )
delay     d1    -delayTime 0.1
delay     d2    -delayTime 0.05
# form the logical not of the non-delayed signal, E(t)
# 'nthresh' outputs 1 if the input is *less* than the reference, else 0
nthresh   n0    -input d0 -ref  -20
# threshold E(t-0.1); 'thresh' returns 1 if the input is *above* the reference
thresh    c1    -input d1 -ref    0
# `And' them (multiply)  m1 = delayed(d0) & not(d0)
multer    m1    -input n0 -in2   c1
# multiply in the third term
nthresh   n2    -input d2 -ref  -20
multer    m2    -input n2 -in2   m1

# 'eventer' objects are one-shot flip flops that emit a pulse of
#  length pulseTime every time they see a rising edge in their inputs, 
#  except after triggering, they ignore the input for holdoffTime.
#  Also, whenever their output goes from low to high, they execute 
#  the Tcl expression set by their 'command' option, with the time of 
#  the actual event appended (so this one executes, e.g. "clap 0.151")
eventer   e1    -input m2 -pulseTime 0.05 -holdoffTime 0.05  -command "clap"
# Track the output of the one-shot
wavesink  .f.ws2  -input e1 -height 40 -scale 1.8 -offset -1  \
	-relief sunken -padX 4 -padY 4

# run an rms for the VU meter
rms      vurms   -input as -sampleRate 10
dB       vudb    

# Add a labeled time axis, for yuks
# The timespan of the chartrecorder displays is 200 steps (default)
#  divided by 100 Hz (output of rms etc).
set timelen [expr "200/100.0"]
timeline  .f.tl   -height 20 -tixTopLeft 1 -relief flat -maxVal $timelen \
	-relief sunken -padX 4 -padY 2

# Add labels for the two chartrecorder displays.  I have to put these 
# labels inside individual 'frames' in order to get them left-aligned
# rather than centered.
frame .f.f1
frame .f.f2
label .f.f1.la1 -text "Energy in $frq Hz band"
label .f.f2.la2 -text "Detected claps"

# Make the chart-recorder and timeline displays visible
pack .f .f.f1 .f.ws1 .f.f2 .f.ws2 .f.tl -padx 2 -pady 2 -fill x
pack .f.f1.la1 .f.f2.la2 -side left

### Add "Pause" and "Quit" buttons..
# This command is bound to the Pause button.  It toggles the 
#  running of the scheduler and changes the button label
proc pausebtncmd {w} {
    global flag
    if { $flag } {
	$w config -text "Resume"
	stop
    } else {
	$w config -text "Pause"
	run
    }
}

# Make the buttons and display them
frame .bf 
button .bf.b1 -text "Pause" -command {pausebtncmd .bf.b1}
button .bf.b2 -text "Quit"  -command {exit}
pack .bf 
pack .bf.b1 .bf.b2 -side left -expand yes -padx 10 -pady 10

### Implement the clapper logic - i.e. there must be two claps within 
### some short time

# A global counts the number of 'recent' claps
set claps 0

# This is the function called by the eventer e1 whenever a clap
# trigger is detected
proc clap {time} {
    global claps
    # increment the count of `recent' claps
    incr claps
    # If there have been two, perform the actual 'clapper' function
    if { $claps == 2 } {
	ItsTheClapper
    }
    # This sets the upper limit on how close in time the claps must be.
    # 300 ms after a clap occurs, the recent clap count is cleared to zero.
    after 300 {set claps 0}
}

### The actual effect of the clapper - toggle the highlight of the 
### icon

# Global holds current state
set lightstate 0

# Function called after two close claps detected
proc ItsTheClapper { }  {
    global lightstate
    if { $lightstate == 0 }  {
#	puts "Clap on .."
	.l config -foreground green
	set lightstate 1
    } else {
#	puts "Clap off!"
	.l config -foreground black
	set lightstate 0
    }
}	

### Functions to make the processing network run.

# 'cont' actually executes the 's step' to advance the scheduler by one 
# step.  Then, depending on the global 'flag', it sets itself to 
# be rerun after a short delay, or doesn't.
proc cont {} {
    global flag timelen
    s step
    # update the timeline manually
    set now [as time]
    set then [expr "$now - $timelen"]
    .f.tl config -minVal $then -maxVal $now
    # update the VU meter
    .vu config -val [vudb val]
    # Quit if we've run out of input (for timing tests anyway)
    if { ![as more] } { exit }
    # arrange for us to be called again
    # Note:  I use the 'after' command, which arranges for a callback 
    # to occur approximately <n> milliseconds into the future, to give 
    # the Tk event loop a chance to handle screen updates.  With a 
    # tighter 'while' loop, the sound processing occurs, but the 
    # user interface is not serviced.
    if { $flag } {
	after 10 { cont }
    }
}

# 'run' starts execution by setting the flag and starting the chain 
# of 'cont's
proc run {} {
    global flag
    set flag 1
    cont
}

# 'stop' pauses the processing by clearing the flag, so the next time 
#  'cont' executes, it won't setup the follow-on call
proc stop {} {
    global flag
    set flag 0
}

# Finally, start the processing
run
