// $Header$
//
// $Log$
//
// Object to do various forms of padding on a data stream
//

#include <stdlib.h>
#include <assert.h>

#include "Padding.h"

const CVector
Padding::processBlock(const CVector in,
                      bool& lastTime)
{
  assert(nc == 1);

  if ((state == End) && (points_in_buffer == 0))
    {
      lastTime = true;
      return CVector(0, buffer);
    }

  if (state == Start)
    {
      // Do padding of leading data
      //
      if (nLeading != 0)
        { 
          switch (method)
            {
            case ZERO:
              {
                float* dest = buffer;
                const float* const dest_end = buffer + nLeading;
                while (dest != dest_end) { *dest++ = 0.0f; }
              }
              break;

            case COPY:
              {
                const float val = *in.data;
                float* dest = buffer;
                const float* const dest_end = buffer + nLeading;
                while (dest != dest_end) { *dest++ = val; }
              }
              break;
            }

          points_in_buffer = nLeading;
        }
      if (in.m != 0) { state = Continuing; }
    }
  else if (points_in_buffer != 0)
    {
      // Move any buffered points to the beginning of the buffer
      //
      float* src = buffer + prev_out_len;
      float* src_end = buffer + points_in_buffer;
      float* dest = buffer;
      while (src != src_end) { *dest++ = *src++; }
      points_in_buffer -= prev_out_len;
    }

  if ((state != End) && (in.m != 0))
    {
      const float* src = in.data;
      const float* const src_end = in.data + in.m;
      float* dest = buffer + points_in_buffer;
      while (src != src_end) { *dest++ = *src++; }
      points_in_buffer += in.m;
    }

  if ((state != End) && lastTime)
    {
      if ((nTrailing != 0) && (points_in_buffer != 0))
        {
          switch (method)
            {
            case ZERO:
              {
                float* dest = buffer + points_in_buffer;
                const float* const dest_end =
                  buffer + points_in_buffer + nTrailing;
                while (dest != dest_end) { *dest++ = 0.0f; }
              }
              break;

            case COPY:
              {
                const float val = *(buffer + points_in_buffer - 1);
                float* dest = buffer + points_in_buffer;
                const float* const dest_end =
                  buffer + points_in_buffer + nTrailing;
                while (dest != dest_end) { *dest++ = val; }
              }
              break;
            }

          points_in_buffer += nTrailing;
        }
      state = End;
    }

  if (state == End)
    {
      if (points_in_buffer > ioblocking)
        {
          prev_out_len = ioblocking;
          lastTime = false;
        }
      else
        {
          prev_out_len = points_in_buffer;
        }
    }
  else
    {
      if (points_in_buffer > ioblocking + 1)
        {
          prev_out_len = ioblocking;
          lastTime = false;
        }
      else
        {
          prev_out_len = points_in_buffer - 1;
        }
    }
  
  return CVector(prev_out_len, buffer);
}


const CMatrix
Padding::processBlock(const CMatrix in,
                      bool& lastTime)
{
  assert(in.m == nc);

  if ((state == End) && (points_in_buffer == 0))
    {
      lastTime = true;
      return CMatrix(nc,
                     0,
                     ioblocking + nLeading + nTrailing + 1,
                     buffer);
    }

  if (state == Start)
    {
      // Do padding of leading data
      //
      if (nLeading != 0)
        { 
          switch (method)
            {
            case ZERO:
              {
                float* dest_row_start = buffer;
                float* dest_row_end = buffer + nLeading;
                float* dest_col_end =
                  buffer + nc * (ioblocking + nLeading + nTrailing + 1);
                while (dest_row_start != dest_col_end)
                  {
                    float* dest = dest_row_start;
                    while (dest != dest_row_end) { *dest++ = 0.0f; }
                    dest_row_start += (ioblocking + nLeading + nTrailing + 1);
                    dest_row_end += (ioblocking + nLeading + nTrailing + 1);
                  }
              }
              break;

            case COPY:
              {
                const float* src = in.data;
                float* dest_row_start = buffer;
                float* dest_row_end = buffer + nLeading;
                float* dest_col_end =
                  buffer + nc * (ioblocking + nLeading + nTrailing + 1);
                while (dest_row_start != dest_col_end)
                  {
                    const float val = *src;
                    float* dest = dest_row_start;
                    while (dest != dest_row_end) { *dest++ = val; }
                    dest_row_start += (ioblocking + nLeading + nTrailing + 1);
                    dest_row_end += (ioblocking + nLeading + nTrailing + 1);
                    src += in.s;
                  }
              }
              break;
            }

          points_in_buffer = nLeading;
        }
      
      state = Continuing;
    }
  else if (points_in_buffer != 0)
    {
      // Move any buffered points to the beginning of the buffer
      //
      float* src_row_start = buffer + prev_out_len;
      float* src_row_end = buffer + points_in_buffer;
      float* src_col_end =
        src_row_start + nc * (ioblocking + nLeading + nTrailing + 1);
      float* dest_row_start = buffer;
      while (src_row_start != src_col_end)
        {
          float* src = src_row_start;
          float* dest = dest_row_start;
          while (src != src_row_end) { *dest++ = *src++; }
          src_row_start += (ioblocking + nLeading + nTrailing + 1);
          src_row_end += (ioblocking + nLeading + nTrailing + 1);
          dest_row_start += (ioblocking + nLeading + nTrailing + 1);
        }
      points_in_buffer -= prev_out_len;
    }

  if ((state != End) && (in.n != 0))
    {
      // Move input data into buffer
      //
      const float* src_row_start = in.data;
      const float* src_row_end = in.data + in.n;
      const float* src_col_end = src_row_start + in.m * in.s;
      float* dest_row_start = buffer + points_in_buffer;
      while (src_row_start != src_col_end)
        {
          const float* src = src_row_start;
          float* dest = dest_row_start;
          while (src != src_row_end) { *dest++ = *src++; }
          src_row_start += in.s;
          src_row_end += in.s;
          dest_row_start += (ioblocking + nLeading + nTrailing + 1);
        }
      points_in_buffer += in.n;
    }

  if ((state != End) && lastTime)
    {
      if ((nTrailing != 0) && (points_in_buffer != 0))
        {
          switch (method)
            {
            case ZERO:
              {
                float* dest_row_start = buffer + points_in_buffer;
                float* dest_row_end = 
                  buffer + points_in_buffer + nTrailing;
                float* dest_col_end =
                  buffer + points_in_buffer +
                  nc * (ioblocking + nLeading + nTrailing + 1);
                while (dest_row_start != dest_col_end)
                  {
                    float* dest = dest_row_start;
                    while (dest != dest_row_end) { *dest++ = 0.0f; }
                    dest_row_start += (ioblocking + nLeading + nTrailing + 1);
                    dest_row_end += (ioblocking + nLeading + nTrailing + 1);
                  }
              }
              break;

            case COPY:
              {
                float* dest_row_start = buffer + points_in_buffer;
                float* dest_row_end = 
                  buffer + points_in_buffer + nTrailing;
                float* dest_col_end =
                  buffer + points_in_buffer +
                  nc * (ioblocking + nLeading + nTrailing + 1);
                while (dest_row_start != dest_col_end)
                  {
                    const float val = *(dest_row_start - 1);
                    float* dest = dest_row_start;
                    while (dest != dest_row_end) { *dest++ = val; }
                    dest_row_start += (ioblocking + nLeading + nTrailing + 1);
                    dest_row_end += (ioblocking + nLeading + nTrailing + 1);
                  }
              }
              break;
            }

          points_in_buffer += nTrailing;
        }
      state = End;
    }

  if (state == End)
    {
      if (points_in_buffer > ioblocking)
        {
          prev_out_len = ioblocking;
          lastTime = false;
        }
      else
        {
          prev_out_len = points_in_buffer;
        }
    }
  else
    {
      if (points_in_buffer > ioblocking + 1)
        {
          prev_out_len = ioblocking;
          lastTime = false;
        }
      else
        {
          prev_out_len = points_in_buffer - 1;
        }
    }
  
  return CMatrix(nc,
                 prev_out_len,
                 ioblocking + nLeading + nTrailing + 1,
                 buffer);
}

#ifdef TEST

#include <stdio.h>

int
main()
{
  Vector v(10);
  for (size_t i = 0; i < 10; i++) { v.data[i] = i+1; }

  for (size_t ioblocking = 1; ioblocking < 13; ioblocking++)
    {
      for (size_t ld = 0; ld < 13; ld++)
        {
          for (size_t tr = 0; tr < 13; tr++)
            {
              Padding p(ioblocking, 1, Padding::ZERO, ld, tr);

              bool EOProc = false;
              int in_offset = 0;
              int length = 0;

              printf("%d %d %d : ",ioblocking, ld, tr);

              while (!EOProc)
                {
                  if (in_offset < 10 - (int) ioblocking)
                    {
                      EOProc = false;
                      length = ioblocking;
                    }
                  else if (in_offset < 10)
                    {
                      EOProc = true;
                      length = 10 - in_offset;
                    }
                  else
                    {
                      EOProc = true;
                      length = 0;
                    }
      
                  const CVector v1(length, v.data + in_offset);
                  const CVector v2 = p.processBlock(v1, EOProc);
                  
                  for (size_t j = 0; j < v2.m; j++)
                    {
                      printf("%g ", v2.data[j]);
                    }

                  in_offset += length;
                }
              printf("\n");
            }
        }
    }

  for (size_t ioblocking = 1; ioblocking < 13; ioblocking++)
    {
      for (size_t ld = 0; ld < 13; ld++)
        {
          for (size_t tr = 0; tr < 13; tr++)
            {
              Padding p(ioblocking, 1, Padding::COPY, ld, tr);

              bool EOProc = false;
              int in_offset = 0;
              int length = 0;

              printf("%d %d %d : ",ioblocking, ld, tr);

              while (!EOProc)
                {
                  if (in_offset < 10 - (int) ioblocking)
                    {
                      EOProc = false;
                      length = ioblocking;
                    }
                  else if (in_offset < 10)
                    {
                      EOProc = true;
                      length = 10 - in_offset;
                    }
                  else
                    {
                      EOProc = true;
                      length = 0;
                    }
      
                  const CVector v1(length, v.data + in_offset);
                  const CVector v2 = p.processBlock(v1, EOProc);
                  
                  for (size_t j = 0; j < v2.m; j++)
                    {
                      printf("%g ", v2.data[j]);
                    }

                  in_offset += length;
                }
              printf("\n");
            }
        }
    }

  Matrix m(2,5);
  for (size_t i = 0; i < 10; i++) { m.data[i] = i+1; }

  for (size_t ioblocking = 1; ioblocking < 8; ioblocking++)
    {
      for (size_t ld = 0; ld < 8; ld++)
        {
          for (size_t tr = 0; tr < 8; tr++)
            {
              Padding p(ioblocking, 2, Padding::ZERO, ld, tr);

              bool EOProc = false;
              int in_offset = 0;
              int length = 0;

              printf("%d %d %d : ",ioblocking, ld, tr);

              while (!EOProc)
                {
                  if (in_offset < 5 - (int) ioblocking)
                    {
                      EOProc = false;
                      length = ioblocking;
                    }
                  else if (in_offset < 5)
                    {
                      EOProc = true;
                      length = 5 - in_offset;
                    }
                  else
                    {
                      EOProc = true;
                      length = 0;
                    }
      
                  const CMatrix m1(2, length, 5, m.data + in_offset);
                  const CMatrix m2 = p.processBlock(m1, EOProc);
                  
                  for (size_t j = 0; j < m2.n; j++)
                    {
                      printf("%g ", m2.data[m2.s+j]);
                    }

                  in_offset += length;
                }
              printf("\n");
            }
        }
    }

  for (size_t ioblocking = 1; ioblocking < 8; ioblocking++)
    {
      for (size_t ld = 0; ld < 8; ld++)
        {
          for (size_t tr = 0; tr < 8; tr++)
            {
              Padding p(ioblocking, 2, Padding::COPY, ld, tr);

              bool EOProc = false;
              int in_offset = 0;
              int length = 0;

              printf("%d %d %d : ",ioblocking, ld, tr);

              while (!EOProc)
                {
                  if (in_offset < 5 - (int) ioblocking)
                    {
                      EOProc = false;
                      length = ioblocking;
                    }
                  else if (in_offset < 5)
                    {
                      EOProc = true;
                      length = 5 - in_offset;
                    }
                  else
                    {
                      EOProc = true;
                      length = 0;
                    }
      
                  const CMatrix m1(2, length, 5, m.data + in_offset);
                  const CMatrix m2 = p.processBlock(m1, EOProc);
                  
                  for (size_t j = 0; j < m2.n; j++)
                    {
                      printf("%g ", m2.data[m2.s+j]);
                    }

                  in_offset += length;
                }
              printf("\n");
            }
        }
    }
}

#endif
