// $Header: /n/yam/da/dpwe/tmp/bedk_frontend/RCS/SOSFilter.cc,v 1.2 1997/08/08 02:09:08 bedk Exp dpwe $
//
// $Log: SOSFilter.cc,v $
// Revision 1.2  1997/08/08 02:09:08  bedk
// Thought I needed doubles as intermediates, but floats will do.
//
// Revision 1.1  1997/08/08 01:57:37  bedk
// Initial revision
//
//
// General ARMA filter object, implemented as a cascade of
// second-order sections
//

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>

#include "SOSFilter.h"

static void swab_uibuf(unsigned int *ptr, int n) {
  // Swap bytes in a buffer of uint32s, if needed to make them bigendian
#ifndef WORDS_BIGENDIAN
  unsigned int d;
  int i;

  assert(sizeof(unsigned int) == 4);

  for (i=0; i<n; ++i) {
    d = *ptr;
    *ptr++ = ((d & 0x000000FF) << 24) | ((d & 0x0000FF00) << 8) | ((d & 0x00FF0000) >> 8) | (0x000000FF & ((d & 0xFF000000) >> 24));
  }
#endif /* WORDS_BIGENDIAN */
}

static void swab_fbuf(float *buf, int n) {
  // Swap bytes in a buffer of floats, if needed to make them bigendian
  unsigned int *ptr = (unsigned int *)buf;
  assert(sizeof(unsigned int) == sizeof(float));
  swab_uibuf(ptr, n);
}

// This uses a temporary file format, until a reasonably portable one
// is found
//
SOSFilter::SOSFilter(const size_t a_ioblocking,
                       const size_t a_nc,
                       const char* const a_filename)
  : ioblocking(a_ioblocking),
    nc(a_nc)
{
  FILE* f;
  unsigned int buf;
  char magic[7];

  if ((f = fopen(a_filename, "r")) == NULL)
    {
      fprintf(stderr,
              "Unable to open file %s for reading.\n",
              a_filename);
      exit(1);
    }

  // Check file type
  //
  if (fread(magic, sizeof(char), 6, f) != 6)
    {
      fprintf(stderr,
              "Error reading %s\n",
              a_filename);
      exit(1);
    }
  *(magic + 6) = '\000';
  if (strcmp(magic, "SOSFLT") != 0)
    {
      fprintf(stderr,
              "%s is not a filter file.\n",
              a_filename);
      exit(1);
    }

  // Now read the file
  //
  if (fread(&sf, sizeof(float), 1, f) != 1)
    {
      fprintf(stderr,
              "Error reading %s\n",
              a_filename);
      exit(1);
    }
  swab_fbuf(&sf, 1);

  if (fread(&gd, sizeof(unsigned int), 1, f) != 1)
    {
      fprintf(stderr,
              "Error reading %s\n",
              a_filename);
      exit(1);
    }
  swab_uibuf(&gd, 1);

  if (fread(&buf, sizeof(unsigned int), 1, f) != 1)
    {
      fprintf(stderr,
              "Error reading %s\n",
              a_filename);
      exit(1);
    }
  swab_uibuf(&buf, 1);
  N = (size_t) buf;

  c = new float[5 * N];
  state_store = new float[2 * N * nc];
  out_buffer = new float[ioblocking * nc];

  if (fread(c, sizeof(float), 5 * N, f) != 5 * N)
    {
      fprintf(stderr,
              "Error reading %s\n",
              a_filename);
      exit(1);
    }
  swab_fbuf(c, 5*N);

  // Initialize filter state to 0.0
  //
  float* dest = state_store;
  const float* const dest_endp = dest + 2 * N * nc;
  while (dest != dest_endp) { *dest++ = 0.0; }

  fclose(f);
}

const CVector
SOSFilter::processBlock(const CVector in,
                         bool& lastTime)
{
  assert(nc == 1);
  assert(in.m <= ioblocking);

  if (state == End)
    {
      lastTime = true;
      return CVector(0, out_buffer);
    }

  if (state == Start) { state = Continuing; }

  // Now do the filtering
  //
  const float* xp = in.data;
  float* yp = out_buffer;
  const float* const y_block_endp = out_buffer + (in.m & ~0x03);
  const float* const y_endp = out_buffer + in.m;
  const float* const c_endp = c + 5 * N;
  const float* cp = c;
  float* sp = state_store;

  // On the initial pass through, read from input and write to output
  // buffer 
  //
  {
    const float b0 = *(cp + 0);
    const float b1 = *(cp + 1);
    const float b2 = *(cp + 2);
    const float a1 = *(cp + 3);
    const float a2 = *(cp + 4);
    float w2 = *(sp + 0);
    float w3 = *(sp + 1);
    while (yp != y_block_endp)
      {
        float x0 = *(xp + 0);
        float x1 = *(xp + 1);
        float x2 = *(xp + 2);
        float x3 = *(xp + 3);
        float w0 = x0 + a1 * w3 + a2 * w2;
        x0 = b0 * w0 + b1 * w3 + b2 * w2;
        float w1 = x1 + a1 * w0 + a2 * w3;
        x1 = b0 * w1 + b1 * w0 + b2 * w3;
        w2 = x2 + a1 * w1 + a2 * w0;
        x2 = b0 * w2 + b1 * w1 + b2 * w0;
        w3 = x3 + a1 * w2 + a2 * w1;
        x3 = b0 * w3 + b1 * w2 + b2 * w1;
        *(yp + 0) = x0;
        *(yp + 1) = x1;
        *(yp + 2) = x2;
        *(yp + 3) = x3;
        xp += 4;
        yp += 4;
      }
    while (yp != y_endp)
      {
        float x0 = *xp++;
        float w0 = x0 + a1 * w3 + a2 * w2;
        x0 = b0 * w0 + b1 * w3 + b2 * w2;
        *yp++ = x0;
        w2 = w3;
        w3 = w0;
      }
    *(sp + 0) = w2;
    *(sp + 1) = w3;
    sp += 2;
    cp += 5;
  }

  // On subsequent passes, read from and write to output buffer 
  //
  while (cp != c_endp)
    {
      yp = out_buffer;
      const float b0 = *(cp + 0);
      const float b1 = *(cp + 1);
      const float b2 = *(cp + 2);
      const float a1 = *(cp + 3);
      const float a2 = *(cp + 4);
      float w2 = *(sp + 0);
      float w3 = *(sp + 1);
      while (yp != y_block_endp)
        {
          float x0 = *(yp + 0);
          float x1 = *(yp + 1);
          float x2 = *(yp + 2);
          float x3 = *(yp + 3);
          float w0 = x0 + a1 * w3 + a2 * w2;
          x0 = b0 * w0 + b1 * w3 + b2 * w2;
          float w1 = x1 + a1 * w0 + a2 * w3;
          x1 = b0 * w1 + b1 * w0 + b2 * w3;
          w2 = x2 + a1 * w1 + a2 * w0;
          x2 = b0 * w2 + b1 * w1 + b2 * w0;
          w3 = x3 + a1 * w2 + a2 * w1;
          x3 = b0 * w3 + b1 * w2 + b2 * w1;
          *(yp + 0) = x0;
          *(yp + 1) = x1;
          *(yp + 2) = x2;
          *(yp + 3) = x3;
          yp += 4;
        }
      while (yp != y_endp)
        {
          float x0 = *yp;
          float w0 = x0 + a1 * w3 + a2 * w2;
          x0 = b0 * w0 + b1 * w3 + b2 * w2;
          *yp++ = x0;
          w2 = w3;
          w3 = w0;
        }
      *(sp + 0) = w2;
      *(sp + 1) = w3;
      sp += 2;
      cp += 5;
    }

  // If we are finished, set the state to End, otherwise copy final
  // filter state to the state store
  //
  if (lastTime) { state = End; }
  
  return CVector(in.m, out_buffer);
}

const CMatrix
SOSFilter::processBlock(const CMatrix in,
                         bool& lastTime)
{
  assert(in.m == nc);
  assert(in.n <= ioblocking);

  if (state == End)
    {
      lastTime = true;
      return CMatrix(nc, 0, ioblocking, out_buffer);
    }

  if (state == Start) { state = Continuing; }

  // Loop over the input channels
  //
  const float* x_chan_startp = in.data;
  const float* const c_endp = c + 5 * N;
  float* sp = state_store;
  float* y_chan_startp = out_buffer;
  const float* const y_chan_end = out_buffer + nc * ioblocking;

  while (y_chan_startp != y_chan_end)
    {
      const float* xp = x_chan_startp;
      float* yp = y_chan_startp;
      const float* const y_block_endp = y_chan_startp + (in.n & ~0x03);
      const float* const y_endp = y_chan_startp + in.n;
      const float* cp = c;

      // On the initial pass through, read from input and write to output
      // buffer 
      //
      {
        const float b0 = *(cp + 0);
        const float b1 = *(cp + 1);
        const float b2 = *(cp + 2);
        const float a1 = *(cp + 3);
        const float a2 = *(cp + 4);
        float w2 = *(sp + 0);
        float w3 = *(sp + 1);
        while (yp != y_block_endp)
          {
            float x0 = *(xp + 0);
            float x1 = *(xp + 1);
            float x2 = *(xp + 2);
            float x3 = *(xp + 3);
            float w0 = x0 + a1 * w3 + a2 * w2;
            x0 = b0 * w0 + b1 * w3 + b2 * w2;
            float w1 = x1 + a1 * w0 + a2 * w3;
            x1 = b0 * w1 + b1 * w0 + b2 * w3;
            w2 = x2 + a1 * w1 + a2 * w0;
            x2 = b0 * w2 + b1 * w1 + b2 * w0;
            w3 = x3 + a1 * w2 + a2 * w1;
            x3 = b0 * w3 + b1 * w2 + b2 * w1;
            *(yp + 0) = x0;
            *(yp + 1) = x1;
            *(yp + 2) = x2;
            *(yp + 3) = x3;
            xp += 4;
            yp += 4;
          }
        while (yp != y_endp)
          {
            float x0 = *xp++;
            float w0 = x0 + a1 * w3 + a2 * w2;
            x0 = b0 * w0 + b1 * w3 + b2 * w2;
            *yp++ = x0;
            w2 = w3;
            w3 = w0;
          }
        *(sp + 0) = w2;
        *(sp + 1) = w3;
        sp += 2;
        cp += 5;
      }

      // On subsequent passes, read from and write to output buffer 
      //
      while (cp != c_endp)
        {
          yp = y_chan_startp;
          const float b0 = *(cp + 0);
          const float b1 = *(cp + 1);
          const float b2 = *(cp + 2);
          const float a1 = *(cp + 3);
          const float a2 = *(cp + 4);
          float w2 = *(sp + 0);
          float w3 = *(sp + 1);
          while (yp != y_block_endp)
            {
              float x0 = *(yp + 0);
              float x1 = *(yp + 1);
              float x2 = *(yp + 2);
              float x3 = *(yp + 3);
              float w0 = x0 + a1 * w3 + a2 * w2;
              x0 = b0 * w0 + b1 * w3 + b2 * w2;
              float w1 = x1 + a1 * w0 + a2 * w3;
              x1 = b0 * w1 + b1 * w0 + b2 * w3;
              w2 = x2 + a1 * w1 + a2 * w0;
              x2 = b0 * w2 + b1 * w1 + b2 * w0;
              w3 = x3 + a1 * w2 + a2 * w1;
              x3 = b0 * w3 + b1 * w2 + b2 * w1;
              *(yp + 0) = x0;
              *(yp + 1) = x1;
              *(yp + 2) = x2;
              *(yp + 3) = x3;
              yp += 4;
            }
          while (yp != y_endp)
            {
              float x0 = *yp;
              float w0 = x0 + a1 * w3 + a2 * w2;
              x0 = b0 * w0 + b1 * w3 + b2 * w2;
              *yp++ = x0;
              w2 = w3;
              w3 = w0;
            }
          *(sp + 0) = w2;
          *(sp + 1) = w3;
          sp += 2;
          cp += 5;
        }

      x_chan_startp += in.s;
      y_chan_startp += ioblocking;
    }
  
  if (lastTime) { state = End; }

  return CMatrix(nc, in.n, ioblocking, out_buffer);
}

#ifdef TEST

#endif
