/*******************************************************************************

	Bigram.C:
		
		Contains class:	 Bigram

		Author:		Dan Jurafsky
		Date:		May 27, 1993
		History:	DSJ	5-27-93	Initial version


	Don't confuse this file with the various programs which manipulate bigrams
	and which are kept in ~jurafsky/berp/src/bigram.  This one is used in the
	parser only when i'm running tests which require knowledge of what the
	bigram grammar would do at some point, to see how much smarter the CFG is
	(or isn't).  It reads in a bigram grammar, and has a method for filtering
	the CFG follow set by the bigram follow set.

******************************************************************************/

#include "grammar.h"
#include "parser.h"
#include "bigram.h"

extern void RemoveStars(char *);
extern char *wp_cont; // name of bigram middle file.  variable from dp_pipe.

/*******************************************************************************

	class Bigram: 
	
	Data:
		char		Tokens[MAXSYMBOLS][MAXSYMLEN];
		int		iTokens;
		link*  		list[MAXSYMBOLS];
		Grammar 	*g;

	Constructor:
	        Bigram(Grammar *g);

	Methods:
		void	AddFollow(char *,int);
		int 	FilterFollow(int *, int ,int *);

	Also uses a quick-and-dirty linked-list class.  Gee, I should really check
	and see if this compiler has parameterized types, and write a nice clean
	linked-list class somewhere else.  Oh well.

	class link {
		int	index;
		link *	next;
		friend 	Bigram;
	};

*******************************************************************************/

Bigram::Bigram(Grammar *gr)
{
	int 	i;
	char 	name[MAXSYMLEN];
	int	symbolindex,followindex;
	FILE	*fp = NULL;
	char	buf[BUFSIZ];
	float	prob;

	g = gr;
	iTokens = 0;
	for (i = 0; i < MAXSYMBOLS; i++) {
		Tokens[i][0] = '\0';
		list[i] = '\0';
	}
        fprintf(stderr,"reading backoff bigram file %s\n",wp_cont);
	fp = fopen(wp_cont,"r");

	if (fp == NULL) {
		printf("Error, could not open bigram.out\n");
		exit(1);
	} else {
		//printf("Opening bigram file\n");
		//fflush(stdout);
	}
	//g->Symbols.AddSymbol("'BEG'");
	//g->Symbols.AddSymbol("'END'");
	//g->Symbols.AddSymbol("'.'");
			//printf("added . \n");
	//fflush(stdout);
	//g->Symbols.AddSymbol("'[uh]'");
	//g->Symbols.AddSymbol("'[um]'");
	//g->Symbols.AddSymbol("'[mm]'");
	//g->Symbols.AddSymbol("'[er]'");

	while(fgets(buf,BUFSIZ,fp) != NULL){
		if(buf[0] == '>'){
			for(i=1;buf[i]!='\n';i++) ;
			buf[i]=0;
			//printf("read in %s\n",buf);
			//fflush(stdout);
			symbolindex = g->Symbols.AddSymbol(&buf[1])->ID();
		} else {
			sscanf(buf,"%s %f",name,&prob);
			//printf("read in %s, %f\n",name,prob);
			//fflush(stdout);
			//for(i=1;buf[i]!=' ';i++) ;
			//buf[i]=0;
			//followindex = g->Symbols.IsSymbol(name);
			followindex = g->Symbols.AddSymbol(name)->ID();
			add_to_follow(followindex,symbolindex,prob);
		}
	}
	fclose(fp);
}

// Add bigram to this word's follow set
// NOTE: bigrams are added at the beginning of the linked list
// so that if they appear in the bigram file in least likely
// to most likely order, they will appear in this list in
// most likely to least likely order.

Bigram::~Bigram()
{
	int i;
	for (i = 0; i < MAXSYMBOLS; i++) {
		clear_list(list[i]);
	}
	
}

void Bigram::clear_list(link * lp)
{
	if (lp == NULL)
		return;
	if (lp->next == NULL)
		delete lp;
	else {
		clear_list(lp->next);
		delete lp;
	}
}


void Bigram::add_to_follow(int followword,int firstword, float prob)
{
	//printf("Adding Follow %d for word %d\n",followword,firstword);
	//fflush(stdout);

	link *bp = list[firstword];
	list[firstword] = new link;
	list[firstword]->index = followword;
	list[firstword]->prob = prob;
	list[firstword]->next = bp;
}
	
float Bigram::get_follow_prob(int firstword, int followword)
{
	link *bp = list[firstword];
	for (; bp; bp = bp->next) {
		if (bp->index == followword)
			return (bp->prob);
	}
	return (0.0);
}

float Bigram::bigram_parse()
{
	int i;
	float bprob = 1.0;

	for (i=0;i < iInputLen-1; i++) {
		bprob = bprob *  get_follow_prob(
							InputTokens[i]->ID(),
							InputTokens[i+1]->ID());
	}
	return bprob;
}

float Bigram::get_bigram_followpair(Symbol * word1,Symbol * word2)
{
	int j = 0;
	link *bp = list[word1->ID()];
	for (; bp; bp = bp->next) {
		if (bp->index == word2->ID())
			return bp->prob;
	}
	return (double) -1;
}



int NotBigramWord(char* s)
{
	int slen = strlen(s);
	return ((s[0] == '/') || 
		(s[0] == '[') || 
		(s[0] == '-') || 
		(s[0] == '<') || 
		((slen == 12) &&
			(s[8] == '.')) ||
		((slen == 9) &&
			(s[5] == '.')) ||
		(s[slen-1] == '-'));
}

void Bigram::tokenizer(char*& inp)
{
	int i;
	char buffer[MAXSYMLEN];
	char *p = inp;
	
	iInputLen = 0;

	for (i=0; i< MAX_SENTENCE_LEN; i++) {
		buffer[0] = '\'';
		while (TRUE) {
			if (GetNextToken(buffer+1,p)) {
				if (NotBigramWord(&buffer[1])) 
					continue;
				 else  
					break;
			} else
				return;
		}
		RemoveStars(buffer);
		buffer[strlen(buffer)+1] = '\0';
		buffer[strlen(buffer)] = '\'';
		//printf("tokenizer: buffer is [%s]\n",buffer);
		InputTokens[i] = g->Symbols.AddSymbol(buffer);
		iInputLen++;
	}
	return;
}

int Bigram::get_bigram_vector_new(Symbol * word, float *bigramprobs)
{
	int j = 0;
	link *bp = list[word->ID()];
	//printf("looking up word %s\n",word->NoQuotesStringify());
	//fflush(stdout);
	for (; bp; bp = bp->next,j++) {
		if (bp->index >= MAXSYMBOLS) {
			fprintf(stderr,"Danger Will Robinson: bp->index is %d\n",bp->index);
			abort();
		}
		bigramprobs[bp->index] = bp->prob;
	}
	return j;
}

int Bigram::get_bigram_vector(Symbol * word,char ** followarray, float *followprobs)
{
	int j = 0;
	link *bp = list[word->ID()];
//	printf("looking up word %s\n",word->NoQuotesStringify());
	//fflush(stdout);
	for (; bp; bp = bp->next) {
		//followarray[j] = word->NoQuotesStringify();
		followarray[j] = g->Symbols.GetSymbol(bp->index)->NoQuotesStringify();
		followprobs[j++] = bp->prob;
	}
	followarray[j] = NULL;
	followprobs[j] = 0;
	return j;
}
		
