/*******************************************************************************

	follow.C:
		
		Contains class:	 FollowSet

		Author:		Dan Jurafsky
		Date:		May 27, 1993
		History:	DSJ	5-27-93	Initial version

******************************************************************************/

#include "follow.h"
#include "grammar.h"
#include "parser.h"
#include "bigram.h"
#include <assert.h>

/********************************* GLOBAL VARIABLES ************************/

extern int      News;
extern int      Deletes;
int             NodeDeletes = 0;
int             NodeCreates = 0;
Grammar        *gr;
ChartParser    *p;
//char          CacheTab[HASHMAX][150];

/*  these are arguments, settable via icsiargs */

float   mu = 0.5;            /* set mu=0.0 for bigram only, 1.0 for cfg only */
int dumbmixture = 1;                      // do dumb, mu -based mixture if true

/*******************************************************************************

	class Follow:	Computes top-down or bottom-up followsets for
			a prefix string passes from the recognizer.
			
	Data:
		float	**termarray;	one int for each terminal,
						passed back by the parser to
						specify the follow set
	Methods:
			FollowSet();
			~FollowSet();
		int	BottomFollow(char * ,char **);
		int 	TopFollow(char * ,char **);

******************************************************************************/

/*******************************************************************************
		FollowSet:  Constructor creates and reads in grammar,
				initializes a parser.
******************************************************************************/
#ifndef START_SYMBOL
#define START_SYMBOL "ss"
#endif

FollowSet::FollowSet()
{
	termarray = (float **) malloc(sizeof(float *) * INITHASH);
	hashsize = INITHASH;
	if (!termarray)
		error("no mem for cache!");

	for (int i = 0; i < hashsize; i++) {
		termarray[i] = (float *) malloc(sizeof(float) * MAXSYMBOLS);
		if (!termarray[i])
			error("no mem for cache!");
	}
	cfg_probs = bigram_probs = NULL;
	cfg_probs = (float *) malloc(sizeof(float) * MAXSYMBOLS);
	if (!cfg_probs)
		error("no mem for cfg_probs!");
	bigram_probs = (float *) malloc(sizeof(float) * MAXSYMBOLS);
	if (!bigram_probs)
		error("no mem for bigram_probs!");
	gr = new Grammar(START_SYMBOL, 0);
	p = new ChartParser(gr);
	gr->ReadFile("berp.grammar");
	p->bg = new Bigram(gr);
	p->ComputeLCArray();
	Tcl_InitHashTable(&HashTable, TCL_STRING_KEYS);
	iCacheLines = 0;
}

FollowSet::~FollowSet()
{
	for (int i = 0; i < hashsize; i++) {
		free((char *) termarray[i]);
	}
	free((char *) termarray);
	free((char *) cfg_probs);
	free((char *) bigram_probs);
	gr->Reset();
	delete          gr;
	p->Reset();
	delete          p->bg;
	delete          p;
	Tcl_DeleteHashTable(&HashTable);
}
/*******************************************************************************

		BottomFollow:  Compute the follow set assuming a bottom-up parse
	
******************************************************************************/

int FollowSet::BottomFollow(char *buf, char **followarray, double *pr)
{
	int             i;
	int             j = 0;
	float          *termp;
	char           *s;
	int             x;
	termp = Cache(buf, BOTTOMFOLLOW, pr);

	for (i = 0; i < MAXSYMBOLS; i++) {
		if (termp[i]) {
			followarray[j++] = gr->Symbols.GetSymbol(i)->NoQuotesStringify();
#ifdef NOTDEF
			s = gr->Symbols.GetSymbol(i)->NoQuotesStringify();
			x = strlen(s);
			assert(x < 100);
			//followarray[j++] = (char *) malloc(strlen(s) + 1);
			followarray[j++] = new char[x + 1];
			strcpy(followarray[j - 1], s);
			free(s);
#endif
		}
	}
	followarray[j] = NULL;
	return j;
}

/*******************************************************************************

		TopFollow:  Compute the follow set assuming a top-down parse
	
******************************************************************************/

int FollowSet::TopFollow(char *buf, char **followarray,
									 float *followprobs, double *pr, int* nwords)
{
	int             i;
	int             j = 0;
	float          *termp;
	char           *s;
	int             x;
	float           tmp;
	double          cfg_prefix = 0;
	double          bigram_prefix = 0;
	float			cfgweight,bigramweight;

	for (i = 0; i < MAXSYMBOLS; i++) {
		cfg_probs[i] = 0.0;
		bigram_probs[i] = 0.0;
	}

	//fprintf(stderr, "entering TopFollow with %s\n", buf);
	//fflush(stdout);

	if (mu == 0.0) {

		/*  If mu is zero, we are doing only bigram */

		p->last_token = p->Tokenizer(buf);
		if (p->last_token == p->EndWord)
			return 0;
		bigramweight = 1.0;
		cfgweight = 0.0;

		*nwords = p->ntokens;

	} else {

		/* either CFG only, or mixture */

		p->last_token = p->Tokenizer(buf);

		if (p->last_token == p->EndWord)
			return 0;
		//fprintf(stderr, "Last token is '%s'\n", p->last_token);

		*nwords = p->ntokens;


		if (mu != 1.0) {

		    //Do the bigram tokenization and parse;

		    p->bg->tokenizer(buf);
		    bigram_prefix = p->bg->bigram_parse();
		    //cout << "bigram parse prob is " << bigram_prefix << "\n" << flush;
		}

		termp = Cache(buf, TOPFOLLOW, &cfg_prefix);
		*pr = cfg_prefix;

		for (i = 0; i < MAXSYMBOLS; i++) {
			if (termp[i] != 0.0)
				cfg_probs[i] = termp[i];
		}

                if (dumbmixture) 
                {
                  cfgweight = mu;
                  bigramweight = 1-mu;
                  /* note: this does the good "kludge" of always multiplying

                     the BG weight by 0.5, even if the CFG fails to parse... */
                }
                else
                {
                  cfgweight = mu * cfg_prefix / 
				(((1 - mu) * bigram_prefix) + (mu * cfg_prefix));
                  bigramweight = (1 - mu) * bigram_prefix /
				 (((1 - mu) * bigram_prefix) + (mu * cfg_prefix));
                  if ((mu == 1.0) && (cfg_prefix == 0.0)) { 
                    cfgweight = 0.0;
                    bigramweight = 1.0;
                  }
                }
                
	    /* the following line is our as-yet-theoretically-unmotivated
		    divide-the-bigram-by-two kludge */
		//if (cfg_prefix == 0.0) 
			//bigramweight = (1 - mu) * bigram_weight;
	}

	if (p->last_token) {
/* #define BEFOREPAUSE  */
#ifdef BEFOREPAUSE
		if (p->ntokens && (p->last_token == p->Pause))
			j = p->bg->get_bigram_vector_new(p->tokens[p->ntokens - 1], bigram_probs);
		else
#endif
			j = p->bg->get_bigram_vector_new(p->last_token, bigram_probs);
		//fprintf(stderr, "Returning size %d fset for word '%s' of string '%s'\n", j, p->last_token->NoQuotesStringify(), buf);
		if (j == 0)
			printf("No bigram follow set for word %s\n", p->last_token->NoQuotesStringify());
	} else {
		error("Last Token is not set for input '%s'", buf);
	}

	//printf("cfgweight is %f\n",cfgweight);
	//printf("bigramweight is %f\n",bigramweight);

	for (i = 0, j = 0; i < MAXSYMBOLS; i++) {
		if (cfg_probs[i] || bigram_probs[i]) {
			followarray[j] = gr->Symbols.GetSymbol(i)->NoQuotesStringify();
			followprobs[j++] = (cfgweight * cfg_probs[i]) + (bigramweight * bigram_probs[i]);
			if (followprobs[j - 1] == 0)
				j--;
		}
	}
	followarray[j] = NULL;
	followprobs[j] = 0;
	return j;
}
int FollowSet::TopFollow(char *buf, char **followarray,
									 float *followprobs, double *pr)
{
  int foo;
  return TopFollow(buf,followarray,followprobs,pr,&foo);
}

    

/*******************************************************************************

		Cache:  If we already have a result for the string passed us,
			return it.
	
******************************************************************************/
float *FollowSet::Cache(char *buf, int topdown, double *pr)
{

	Tcl_HashEntry  *ptr;
	int             newptr = 0;
	int             index;
	int             flags = 0;
	int             guessflag = 0;
	int             followflag = 0;
	int             guesslist[30];
	char           *s;
	char            string[BUFSIZ];

	strcpy(string, buf);
	string[strlen(string) + 1] = 0;
	if (topdown == BOTTOMFOLLOW)
		string[strlen(string)] = 'B';
	else
		string[strlen(string)] = 'T';
	ptr = Tcl_CreateHashEntry(&HashTable, string, &newptr);
	if (newptr == 0) {
		//this entry already there;
		return termarray[(int) Tcl_GetHashValue(ptr)];
	} else {
		//nothing there, compute a new one
			index = iCacheLines++;
		//printf("index is %d\n", index);
		//fflush(stdout);
		if (index >= hashsize)
			bumphashsize();
		//printf("about to call parser with '%s'\n", buf);
		//fflush(stdout);
		if (topdown == BOTTOMFOLLOW)
			s = p->Parser(buf, pr, 0, 0, &guessflag, guesslist);
		else
			s = p->TopDownParser(buf, pr, 0, 0);

		//printf("returned from parser, index is %d \n", index);
		p->CopyOutTermArray(termarray[index]);
		p->Reset();
		Tcl_SetHashValue(ptr, index);
		if (s)
			free(s);
		return termarray[index];
	}
}

void FollowSet::Reset()
{
	Tcl_DeleteHashTable(&HashTable);
	Tcl_InitHashTable(&HashTable, TCL_STRING_KEYS);
	iCacheLines = 0;
}

/* ------------------------bumphashsize------------------------ */
/*
 * bumphashsize will increase the size of the hastable
 */
void FollowSet:: bumphashsize()
{
	int             i;
	int             newsize = hashsize + INITHASH;
	termarray = (float **) realloc((char *)termarray,
				       sizeof(float *) * newsize);
	if (!termarray)
		error("no mem for cache!");
	for (i = hashsize; i < newsize; i++) {
		termarray[i] = (float *) malloc(sizeof(float) * MAXSYMBOLS);
		if (!termarray[i])
			error("no mem for cache!");
	}
	hashsize = newsize;
}

