/*******************************************************************************

	grammar.C:
		
		Contains class:	 Grammar

		Author:		Dan Jurafsky
		Date:		September, 1992
		History:	DSJ	6-3-93	Added header

******************************************************************************/

#include "grammar.h"

#ifdef SUNOS41
/* no rand* protos on sunos41?? */
double drand48(void);
void srand48(long seedval);
#endif /* SUNOS41 */


/********************************* GLOBAL VARIABLES ************************/

const int MAXRHSRULES = 100;
const int FIRSTRHS = 1;
const int LEFTMOST = TRUE;
const int RIGHTMOST = FALSE;

int IndexCount[MAXSYMBOLS];
int IndexTable[MAXSYMBOLS][MAXINDEX];
int LeftIndexStart[MAXSYMBOLS];
int LeftIndexEnd[MAXSYMBOLS];
SymbolCollection *sc;

inline double my_log2(double arg) {
  return log(arg)/log(double(2.0));
}

/*******************************************************************************

	class Grammar:	
			
	Data:
		RuleCollection	Productions;
		Symbol		*Start;
		SymbolCollection Symbols;
		char		LastSentence[200];
		double		LastProb;
		int		LastLength;
		int 		visitlist[MAXRULES];

	Constructor:

		Grammar(char *, ...);
	Methods:
		void		StoreRule(Rule r);
		Rule		*GetRandomRule(Symbol*);
		void	print();
		int 	NRules();
		Rule * 	GetRule(int );
		Symbol	*GetStart();
		void	ReadFile(char*);
		void	AddRule(char *);
		double	EmpiricalEntropy();
		double	GrammarPerplexity();
		void	GenerateRandomSentence();
		void	GetSentence(char *);
		double	GetProb();
		int	GetRulesWithRhs(Symbol*,Rule**);
		int	GetRulesWithLhs(Symbol*,Rule**);
		void	CheckRules();
		void	ComputeTerminalExpansions(Symbol *, int * , int*);
		void 	Reset();
		void 	InitIndexTable();

*************************************************************************************/

/*******************************************************************************

		Grammar:  Constructor creates a grammar with a distinguished
				start symbol.  For small grammars, you can pass
				in the rules as arguments, but usually we just pass
				in the start symbol and use ReadFile to read in all
				the rules.

******************************************************************************/
Grammar::Grammar(char *start ...)
{
	va_list ap;
	char *p;
	Rule * r;

	va_start(ap,start);

	sc = &Symbols;
	Symbols.iNonTerms = 0;
	Symbols.iTerms = 0;
	LeftCorner = NULL;
	Start = Symbols.AddSymbol(start);
	srand48( (long) time(NULL));

	while (p = va_arg(ap,char *)) {
		r = new Rule(p);
		Productions.AddRule(r);
	}
	va_end(ap);
}

Grammar::~Grammar()
{
	Productions.Reset();
	if (LeftCorner) 
		matrix_free(LeftCorner);
}


/*******************************************************************************

		ReadFile:  Reads in the lines of the grammar file, calling
				AddRule() for each line that is not a comment or blank.
		
			Arguments:	char *	filename

******************************************************************************/
char *scanline(char *,int,FILE *);
int GrammarFileLineNumber;

int Grammar::ReadFile(char *filename)
{
	char buf[BUFSIZ];
	FILE *fp = fopen(filename,"r");

	if (fp == NULL) {
	    fprintf(stderr, "Grammar::Cannot open file %s\n",filename);
	    return 0;
	}

	GrammarFileLineNumber=0;
	if (scanline(buf,BUFSIZ,fp) != NULL) {
		/* Read in header line */
		if (buf[0] == '%')
			strcpy(Header,buf);
		else {
			Header[0] = 0;
			if (((buf[0] != '/') || (buf[1] != '/')) && (strlen(buf) > 5))
			  AddRule(buf);
		}
	}
		

	while (scanline(buf,BUFSIZ,fp) != NULL) {
		// ignore comment or blank lines in grammar
		if (((buf[0] != '/') || (buf[1] != '/')) && (strlen(buf) > 5))
		  AddRule(buf);
	}
	fclose(fp);
	GrammarFileLineNumber=0;
	InitIndexTable();
	
	return 1;
}

char *scanline(char *buf, int bufmax, FILE *fp) {
  char *p=buf, *q;
  int pmax=bufmax;
  int parencount=0;
  int bracketcount=0;
  int error=0;
  static char *retval=buf;

  p[0]='\0';
  while (pmax > 1 && (retval=fgets(p,pmax,fp))) {
    GrammarFileLineNumber++;
    // must allow fgets more than 0 bytes to read, or you get an infinite loop
    for(q=p;*q && *q!='\n';q++) {
      switch(*q) {
      case '(':
	parencount++;
	break;
      case ')':
	parencount--;
	break;
      case '{':
	bracketcount++;
	break;
      case '}':
	bracketcount--;
	break;
      }
      if (parencount < 0) {
	fprintf(stderr,"Error, grammar file line %d:\n\tmismatched () in grammar file.  Ignoring trailing garbage.\n",GrammarFileLineNumber);
	*q='\0';
	return buf;
      } else if (parencount == 0 && bracketcount < 0) {
	fprintf(stderr,"Error, grammar file line %d:\n\tmismatched {} in grammar file.  Ignoring trailing garbage.\n",GrammarFileLineNumber);
	*q='\0';
	return buf;
      }
    }
    *q='\0';
    p=q;
    pmax=bufmax-(p-buf);
    if (parencount==0 && bracketcount==0)
      break;
  }
  if (retval || buf[0])
    return buf;
  else
    return NULL;
}

		
	  

/*******************************************************************************

		AddRule:  Calls the Rule constructor to make a rule, and then
				adds it to the grammar's production list.
		
			Arguments:	char *	rule - a string which the Rule() constructor
					will use to make the rule.

******************************************************************************/
void Grammar::AddRule(char *rule)
{
	Rule * r = new Rule(rule);
	Productions.AddRule(r);
}

/*******************************************************************************

		GetRule:  Get rule by rule number
		
			Arguments:	int	num  - the rule number

******************************************************************************/
Rule * Grammar::GetRule(int num)
{
	return Productions.GetRule(num);
}

/*******************************************************************************

		NRules:  returns  the number of rules in the grammar

******************************************************************************/
int Grammar::NRules()
{
	return Productions.NRules();
}

/*******************************************************************************

		GetRandomRule:  Gets a random rule which expands the symbol s.
				(i.e. whose left hand side is s). Picks a random probability
				and then uses a RuleIterator to find the rule whose probability
				lies in that region.

******************************************************************************/

Rule *Grammar::GetRandomRule(Symbol *s)
{
	double d;
	Rule *p, *prev = NULL;
	double probspace = 0;
	RuleIterator next(&Productions,s);


	d = drand48();
	
	while (p = next(s)) {
		prev = p;
		probspace += p->RuleProb();
		if (d < probspace) 
			break;
	}
	if (prev == NULL) 
		error("Nonterminal [%s] has not been defined\n",s->NoQuotesStringify());
	
	return prev;
}

/*******************************************************************************

		CheckRules:  Just makes sure all the probabilities add up to 1.
				Not currently used.

******************************************************************************/

void Grammar::CheckRules()
{
	int i;
	Rule *p;
	double prob = 0;
	RuleIterator next(&Productions);
	NonTermIterator nonterm = &Symbols;
	Symbol *symp;

	for (i = 0; i < Symbols.iNonTerms; i++) {
		symp = nonterm.Next();
		next.Reset(symp);
		prob = 0;
		while (p = next(symp)) {
			prob += p->RuleProb();
		}
		if ((float)prob != (float)1)  {
			cout << "Bad sum" <<  prob << " for LHS";
			symp->print();
			cout << "\n";
		}
	}
}

/*******************************************************************************

		print:  prints out all the rules

******************************************************************************/

void Grammar::print()
{
	RuleIterator next = &Productions;
	Rule *p;

	while (p = next())
		p->print();
}

/*******************************************************************************

		GenerateRandomSentence:  Generates a random sentence according
					to its probability of occurrence according to the
					rule probabilities.  Does this by building a tree
					top-down, expanding each node by stochastically 
					choosing a grammar rule which expands it, until
					every node is expanded to a terminal node.

******************************************************************************/

void Grammar::GenerateRandomSentence()
{
	Node n(Start,1.0,-1);    // top node is grammar's start symbol
	Tree t(&n);    // top node is grammar's start symbol
	Symbol *symCurrent = Start;
	Rule *r;
	int ret;

	do {
		r = GetRandomRule(symCurrent);
		ret = t.ExpandNode(LEFTMOST,r);
		symCurrent = t.GetNonTerminal(LEFTMOST);
	} while (ret && symCurrent);

	(void)t.GetSentence(LastSentence);
//	cout << LastSentence << flush;
//	cout << " [" << t.GetProb();
//	cout <<  "]\n" << flush;
	LastProb = t.GetProb();
	LastLength = t.GetLen();
//	t.print();
}

/*******************************************************************************

		ComputeTerminalExpansions:  given a symbol 's', return the list
					of all words which could be the leftmost
					prefix of some expansion of s.  That is,
					given that we know only that the next non-terminal
					symbol is 's', compute the followset
					of terminals.

******************************************************************************/

void Grammar::ComputeTerminalExpansions(Symbol * s, int * TermList)
{
	RuleIterator ri(&Productions,s);
	Rule * rp;

	/* Go through all rules which have LHS 's', and recursively check all their
		leftmost children, adding them to the "TermList" array.  It's possible
		to do this in time which is linear in the number of rules, by
		keeping a "visitlist" array of rules which we've seen, and if we
		try to expand a rule which we've already checked, we just return. */

	if (s->IsTerminal()) {
		TermList[s->ID()] = 1;
	} else while (rp = ri(s)) {
		if (visitlist[rp->RuleNum]) return;
		visitlist[rp->RuleNum] = 1;
		if (rp->iNumChildren && rp->rhs[0]->IsTerminal()) {
			TermList[rp->rhs[0]->ID()] = 1;
		} else if (rp->iNumChildren)
			ComputeTerminalExpansions(rp->rhs[0],TermList);
	}
}

/*******************************************************************************

		GrammarPerplexity:  returns 2 to the entropy.

******************************************************************************/


double Grammar::GrammarPerplexity()
{
	return pow(2, EmpiricalEntropy());
}

/*******************************************************************************

		EmpiricalEntropy:  compute entropy for a grammar by generating
					10000 random sentences, and taking the
					log of the geometric mean of their probabilities.

******************************************************************************/

double Grammar::EmpiricalEntropy()
{
	int i;
	int len = 0;
	double entropy = 0;

	for (i=0;i<01000;i++) {
		GenerateRandomSentence();
		entropy = entropy + my_log2(LastProb);
		if (LastProb == 0.0)
			error("Sentence with 0 prob\n");
		len = len + LastLength;
	}

	// i can't remember why i use 'i' in here at all, since it cancels
	// out.  I guess i'll double check this, and then remove it.

	entropy = entropy / (double)i;

	entropy = - entropy / ((double)len / (double)(i));
	return entropy;

}

/*******************************************************************************

		GetProb:  return the prob of the last sentence generated

******************************************************************************/

double Grammar::GetProb()
{
	return LastProb;
}

/*******************************************************************************

		GetStart:  return the distinguished start symbol

******************************************************************************/

Symbol *Grammar::GetStart()
{
	return Start;
}

/*******************************************************************************

		GetSentence:  return the last sentence generated

******************************************************************************/

void Grammar::GetSentence(char * buf)
{
	strcpy(buf,LastSentence);
}

/*******************************************************************************

		GetRulesWithRhs,
		GetRulesWithLhs:  A couple of methods used by the bottom-up and
				top-down parsers to access rules given a symbol.
				GetRulesWithRhs returns all the rules whose
				first right-hand constituent is the specificed symbol
				GetRulesWithLhs returns all the rules whose
				left-hand-side is the symbol.
				Both use RuleIterators to do all the work,
				and both return their values by filling an array
				of pointers to rules.  They are passed the array as input.

******************************************************************************/

int Grammar::GetRulesWithRhs(Symbol *s,Rule** r)
{
	Rule *p;
	RuleIterator nextrule(&Productions);
	uint i = 0, icount = 0;

	while (p = nextrule.NextRuleByRhs0(s)) {
		if (icount > MAXRHSRULES)
			error("Too many (%d) rules with rhs[0] = '%s'\n",icount,s->NoQuotesStringify());
		r[icount++] = p;
	}
	return icount;
}

int Grammar::GetRulesWithLhs(Symbol *s, Rule ** r)
{
	Rule *p;
	RuleIterator iter(&Productions,s);
	uint i = 0, icount = 0;

	while (p = iter.NextLexLhs(s)) {
		r[icount++] = p;
		if (icount > 30)
			printf("Oh my god %d rules\n",icount);
		if (icount > MAXRHSRULES)  
			error("Too many (%d) rules with lhs = '%s'\n",icount,s->NoQuotesStringify());
	}
	
	return icount;
}

/*******************************************************************************

		Reset:  Clear the grammar.

******************************************************************************/

 
void Grammar::Reset()
{
	Productions.Reset();
}


/*******************************************************************************

		InitIndexTable:  This initializes all the indexing tables which are
				used to speed up rule access in the parser(s).
				
				IndexTable[symbols][rules]  For each symbol s, a list
					of the rules for which s is the first child, 
					i.e. the leftmost constituent. Used to speed
					up bottom-up rule access.

				LeftIndexStart[symbols] For each symbol s, the rule
					number of the rule r which is the first rule
					to expand s, i.e. the first rule with lhs = s.
					
				
				LeftIndexEnd[symbols] For each symbol s, the rule
					number of the rule r which is the LAST rule
					to expand s, i.e. the LAST rule with lhs = s.

				Since in general I try to keep together all the rules 
					which expand a given symbol, if we use LeftIndexStart
					and LeftIndexEnd to constrain our search for
					expansions, we will save some serious time.
					We don't REQUIRE that rules be contiguous, since
					we check anyhow, but it does speed things up a lot,
					so try to keep doing it.

******************************************************************************/

void Grammar::InitIndexTable()
{
	int i,j;
	Rule *rp;
	int symno;
        NonTermIterator nonterm = &Symbols;
        Symbol *symp;

	for (i = 0; i < Symbols.iSymbols; i++) {
		IndexCount[i] = 0;
		for (j=0;j<MAXINDEX;j++) {
			IndexTable[i][j]=0;
		}
		LeftIndexStart[i] = -1;
		LeftIndexEnd[i] = 0;
	}
	for (i=0;i<Productions.NRules();i++) {
		rp = Productions.GetRule(i);
		symno = rp->rhs[0]->ID();
		IndexTable[symno][IndexCount[symno]++] = rp->RuleNum;
		symno = rp->lhs->ID();
		if (LeftIndexStart[symno] == -1) {
			LeftIndexStart[symno] = i;
			LeftIndexEnd[symno] = i;
		} else
			LeftIndexEnd[symno] = i;
	}
	while (symp = nonterm.Next()) {
		if (LeftIndexStart[symp->ID()] == -1) 
			error("No expansion for symbol '%s'\n",symp->NoQuotesStringify());
	}
	Symbols.SetLastSym(Symbols.iSymbols);
	//Symbols.AddSymbol("'???'");
}

/*******************************************************************************

		ResetVisitList:  Clear the visit list, used in ComputeTerminalExpansions
				to do linear searches of parse arcs.

******************************************************************************/


void Grammar::ResetVisitList()
{
	int i;
	for (i=0;i<MAXRULES;i++)
		visitlist[i] = 0;
}

