/*******************************************************************************

	trees.C:
		
		Contains classes:
			Node
			Tree
			Stack
			LeafIterator
			PreOrderTreeIterator
			PostOrderTreeIterator

		Author:		Dan Jurafsky
		Date:		September, 1992
		History:	DSJ	5-27-93	Added header
		                JEF     12-04-94 Added unification

******************************************************************************/
#include "grammar.h"

/************************************************************************

	class Node: Each node has the following private data:
	
		Symbol	*val;		// The (lhs) value of the node
		Node **  children;	// an array of ptrs to children
						this array is "malloc"ed,
						not "new"'d
		int	iNumChildren;	// the number of children
		double	prob;		// The bottom-up product of all the rules
						used in building the tree to here
		int	RuleNum;	// the index of the rule that	
						expands this node
	Plus four constructors: 

		Node();
		Node(Symbol*,int);
		Node(Symbol*,double,int);
	        Node(Node*);
		~Node();

	And these useful functions:

	void	AddChildren(Symbol**,int);
	Node	*Copy();
	Node 	*GetLowestNode(int);
	Node	*LeftChild();
	Node	*RightChild();
	Node	*GetNonTerminal(int);
	int 	CountRulesUsed(int iRule);
	Symbol	*GetVal();
	int	IsLeafNode();

************************************************************************/

extern int NodeDeletes;		// These are for debugging because
extern int NodeCreates;		// memory leaks seem to occur most often here
extern int Unify;

Node::Node()
{
	NodeCreates++;
	iNumChildren = 0;
	prob = 0;
	RuleNum = 0;
}

Node::Node(Symbol *sym,int iRule)
{
	NodeCreates++;
	val = sym;
	iNumChildren = 0;
	prob = 0;
	RuleNum = iRule;
}

Node::Node(Symbol *sym, double p,int iRule)
{
	NodeCreates++;
	val = sym;
	iNumChildren = 0;
	prob = p;
	RuleNum = iRule;
}

Node::Node(Node*n)
{
	NodeCreates++;
	val = n->val;
	prob = n->prob;
	RuleNum = n->RuleNum;
	iNumChildren = n->iNumChildren;
}

Node::~Node()
{
	uint i;

	NodeDeletes++;
	if (iNumChildren) {
		for (i = 0; i < iNumChildren; i++)  {
			if (children[i])
				delete children[i];
		}
		free((char *)children);
	}
	children = 0;
	iNumChildren = 0;
}


Node * Node::Copy()
{
	int i;
	Node * p = new Node(this);

	if (iNumChildren)
	  p->children = (Node **)malloc(iNumChildren*sizeof(Node *)); 

	for (i = 0; i < iNumChildren;i++) {
		if (children[i]) 
		  p->children[i] = children[i]->Copy();
		else
		  p->children[i] = NULL;
	}

	return p;
}


void Node::AddChildren(Symbol **p, int iCount)
{
	uint i;

	if (iCount)
	  children = (Node **)malloc(iCount*sizeof(Node *));
	for (i = 0; i < iCount; i++)
		children[i] = new Node(p[i],NORULE);
	iNumChildren = iCount;
}


Node * Node::LeftChild()
{
	if (iNumChildren)
		return children[0];
	else
		return 0;
}

Node * Node::RightChild()
{
	if (iNumChildren)
		return children[iNumChildren-1];
	else
		return 0;
}

Symbol *Node::GetVal()
{
	return val;
}

/************************************************************************

	class Tree: Each tree has the following private data:

		Node	*Root;
		double	TreeProb;
		int	SentenceLength;
		Dag     *ParentDag;
		Dag     **ChildDag;
		SubstList *SL;
		ConstraintList *Constraints;
		ConstraintIterator *ConstraintIndex;

	Plus four constructors: 

		Tree();
		Tree(Symbol*);
		Tree(Node*);
                Tree(Symbol*, Tree *, int, int);
	        ~Tree();

	And these useful functions:

	void	ConjoinChild(int dot, Tree *t);
	Tree 	*Copy();
	Symbol	*GetNonTerminal(int);
	int	GetSentence(char *);
	double	GetProb();
	int	GetLen();
	Node	*GetRoot();
	int	ExpandNode(int, Rule*);
	void	print();
	int	CountRulesUsed(int iRule);
	int     checkConstraints();
	int     unifyChildren();

************************************************************************/

Tree::~Tree()
{
	int i;
	if (Root)
		delete Root;
	Root = 0;
	if (Unify) {
	  //	if (ParentDag)       // using Dag::flush_the_toilet() instead
	  //	  delete ParentDag;
	  ParentDag=0;
	  if (ChildDag)
	    delete[] ChildDag;
	  ChildDag=0;
	  if (SL)
	    delete SL;
	  SL=0;
	  Constraints=NULL;
	  if (ConstraintIndex)
	    delete ConstraintIndex;
	  ConstraintIndex=NULL;
	  //contstraints can be found under the rules, so do not delete
	}
}

Tree::Tree()
{
	TreeProb = 1;
	SentenceLength = 0;
	Root = 0;
	ParentDag = 0;
	ChildDag = 0;
	SL = 0;
	Constraints=NULL;
	ConstraintIndex=NULL;
}

Tree::Tree(Symbol *s)
{
	Root = 	new Node(s,NORULE);
	TreeProb = 1;
	SentenceLength = 0;
	ChildDag = 0;
	ParentDag = 0;
	SL = 0;
	Constraints=NULL;
	ConstraintIndex=NULL;
}

Tree::Tree(Node *n)
{
	Root = 	n;
	TreeProb = 1;
	SentenceLength = 0;
	// ?? Is this the right thing to do?
	ParentDag = 0;
	ChildDag = 0;
	SL = 0;
	Constraints=NULL;
	ConstraintIndex=NULL;
}

Tree::Tree(Symbol *s, Tree * t, int numKids, int iRule,ConstraintList *c)
{
	int i;
	extern int Unify;

	SentenceLength = 0;
	TreeProb = 1;
	Root =  new Node(s,iRule);
	Root->children = (Node **)malloc(numKids*sizeof(Node *)); 
	for (i=0; i< numKids; i++)
	  Root->children[i] = NULL;
	if (t)
		Root->children[0] = (t->Root)->Copy();
	Root->iNumChildren = numKids;

	// unification stuff
	if (Unify) {
	  char *temp = s->Stringify();
	  ParentDag=make_nonterminal(temp);
	  free(temp);
	  ChildDag=new Dag *[numKids+1];
	  //a little trickery here
	  ChildDag[0]=ParentDag;
	  if (t)
	    ChildDag[1]=t->ParentDag;
	  SL=new SubstList;
	  Constraints=c;
	  ConstraintIndex=new ConstraintIterator(c);
	} else {
	  ParentDag = 0;
	  ChildDag = 0;
	  Constraints = 0;
	  ConstraintIndex = 0;
	  SL = 0;
	}
 
}


double Tree::GetProb()
{
	return TreeProb;
}
int Tree::GetLen()
{
	return SentenceLength;
}

Symbol *Tree::GetNonTerminal(int LeftMost)
{
	Node * n = Root->GetNonTerminal(LeftMost);

	if (n)
		return n->val;
	else
		return 0;
}


void Tree::ConjoinChild(int dot, Tree *t)
{
  extern int Unify;

	//cout << "conjoining into " << dot << " child of ";
	//this->print();
	//cout << " with ";
	//t->print();
	Root->children[dot] = t->Root->Copy();
	if (Unify)
	  ChildDag[dot+1]=t->ParentDag;

	//cout << " answer piece is: ";
	//Root->children[dot]->val->print();
	//if (dot != 0)
		//Root->children[dot-1]->val->print();
	//cout << " answer is: ";
	//this->print();
}

Tree *Tree::Copy()
{
  int i;
  extern int Unify;

	Tree * p = new Tree();
	if (p == NULL)
	  error("Tree:Copy: no more memory");

	p->SentenceLength = SentenceLength;
	p->TreeProb = TreeProb;

	p->Root = Root->Copy();

  if (Unify) {
    p->ParentDag= new Dag;
    p->ParentDag->copy(ParentDag);
    p->ChildDag = new Dag *[Root->iNumChildren+1];
    for(i=1;i<=Root->iNumChildren;i++)
      p->ChildDag[i]=ChildDag[i];
    // a little trickery here
    p->ChildDag[0]=p->ParentDag;

    p->Constraints=Constraints;
    p->ConstraintIndex=new ConstraintIterator(p->Constraints);
    p->SL=SL->copy();
  } else {
    p->ParentDag = 0;
    p->ChildDag = 0;
    p->Constraints = 0;
    p->ConstraintIndex = 0;
    p->SL = 0;
  }

  return p;
}

int Tree::ExpandNode(int LeftMost, Rule *r)
{
	int i;
	Node *n = Root->GetNonTerminal(LeftMost);

	if (n) {
		n->RuleNum = r->RuleNum;
		n->AddChildren(r->rhs, r->iNumChildren);
//		for ( i = 0; i < r->iNumChildren; i++ )
//			n->children[i]->prob = n->prob * r->prob;
			n->children[0]->prob = n->prob * r->prob;
		for ( i = 1; i < r->iNumChildren; i++ )
			n->children[i]->prob = 1.0;
		return 1;
	}
	return 0;
}

void Tree::CountRulesUsed(int * RuleCounts)
{
	PreOrderTreeIterator next = Root;
	Node *p;
	int count = 0;
		
	while (p = next())  {
		if (!p->IsLeafNode())
			RuleCounts[p->RuleNum]++;
	}
}
#ifdef NOTDEF
int Tree::CountRulesUsed(int iRule, int * RuleCounts)
{
	PreOrderTreeIterator next = Root;
	Node *p;
	int count = 0;
		
	while (p = next())  {
		if (!p->IsLeafNode() &&  (p->RuleNum == iRule))
			count++;
	}
	return count;
}
#endif


void Tree::print()
{
	PreOrderTreeIterator next = Root;
	Node *p;

	cout << "{p=" << TreeProb << "} ";

	if (Root == NULL)
		return;
	while (p = next()) {
		if (next.IsParen(p))
		  cout << ")";
		else if (p->IsLeafNode()) {
		  p->val->print();
		  cout << " ";
		} else {
			cout << "( " ;
			p->val->print();
			cout  << " ";
			cout << "[Rule " << p->RuleNum << "]";
		}
	}
	cout << "\n" << flush;
}


int Tree::checkConstraints(int dot,int rulenum) { 
  Dag *d1, *d2;
  Constraint *c;

  if(Constraints) {
    // should probably do a check around this
    if(ConstraintIndex->end())
      ConstraintIndex->reset(Constraints);
    
    for(;!ConstraintIndex->end() && ConstraintIndex->cur()->head<=dot;
	ConstraintIndex->next()) {
      c=ConstraintIndex->cur();
      d1=(c->head1)?ChildDag[c->head1]:ParentDag;
      d2=(c->head2)?((c->head2==-1)?c->dag2:ChildDag[c->head2]):ParentDag;
      
      if (!d1 || !d2) {
	cerr << "Warning... empty dag for rule [" << rulenum << "]\n";
	c->write(cerr);
      }

      if (!(SL=d1->unify_with_paths_and_sl(d2,c->path1,c->path2,SL)))
	return 0;
    }
  }
  return 1; 
}

int Tree::unifyChildren(int rulenum) {
  Dag *d1, *d2;
  Constraint *c;
  extern int onlineUnify;
  extern int UnifyYes, UnifyNo;

  //cout << "+++++++++++++++Entering unifyChildren\n";
  //ParentDag->write(cout);

  if (!onlineUnify) {
    if (Constraints) {
      // should probably do a check around this
      if(ConstraintIndex->end())
	ConstraintIndex->reset(Constraints);
      
      for(;!ConstraintIndex->end();ConstraintIndex->next()) {
	c=ConstraintIndex->cur();
	d1=(c->head1)?ChildDag[c->head1]:ParentDag;
	d2=(c->head2)?((c->head2==-1)?c->dag2:ChildDag[c->head2]):ParentDag;

	//d1->write(cout,0);
	//c->path1->write(cout);
	//cout << c->head1 << "\n";
	//cout << "---------------\n";
	//d2->write(cout,0);
	//c->path2->write(cout);
        //cout << c->head2 << "\n";
	//SL->write(cout);
	//cout << "===============\n";

	if (!d1 || !d2) {
	  cerr << "Warning... empty dag for rule [" << rulenum << "]\n";
	  c->write(cerr);
	}

	if (!(d1->unify_with_paths_and_sl(d2,c->path1,c->path2,SL))) {
	  UnifyNo++;
	  return 0;
	}
      }
    }
  }
  //cout << "--> AFTER UNI SL\n";
  //SL->write();
  ParentDag = ParentDag->fill_in(SL);
  //cout << "--> AFTER FILL SL\n";
  //SL->write();
  //cout << "\nAFTER\n";
  //ParentDag->write(cout);
  delete SL;
  SL=0;
  UnifyYes++;
  return 1;
}

#ifdef TATEYAMA
void Tree::output2ptw(FILE *s4ptw)
{
	PreOrderTreeIterator next = Root;
	Node *p;
	char *valstr;

	if (Root == NULL) {
	  fprintf(s4ptw, "0: notree\n");
	  return;
	}
	while (p = next()) {
		if (next.IsParen(p))
		  // end children
		  fprintf(s4ptw, "3:EC\n");
		else if (p->IsLeafNode()) {
		  // leaf node.
		  valstr = p->val->Stringify();
		  fprintf(s4ptw, "2:LN: %s\n", valstr);
		  free(valstr);
		} else {
		  valstr = p->val->Stringify();
		  fprintf(s4ptw, "1:PN: %5d: %s\n", p->RuleNum,
			  valstr);
		  free(valstr);
		}
	}
	// End Tree
	fprintf(s4ptw, "8:ET ------------------------------\n");
	//fflush(s4ptw);
}
#endif

Stack::Stack()
{
	index = 0;
	TheStack[index] = 0;
}

void Stack::Push(STACKTYPE el)
{
	if (index == MAXSTACK - 1)
		error("stack overflow");
	TheStack[++index] = el;
}

STACKTYPE Stack::Pop()
{
	if (index == 0)
		return 0;
	else
		return TheStack[index--];
}

STACKTYPE Stack::Peek()
{
	return TheStack[index];
}

	

PostOrderTreeIterator::PostOrderTreeIterator(Node *n)
{
	s.Push(n);
}

PostOrderTreeIterator::PostOrderTreeIterator()
{
}

void PostOrderTreeIterator::operator=(Node * n)
{
	s.Push(n);
}

Node * PostOrderTreeIterator::operator()()
{
	Node * tmp;
	int i;

	if (s.Peek() == 0)
		return 0;
	if (s.Peek() == &marker) {
		(void) s.Pop();
		return s.Pop();
	}
	if (s.Peek()->IsLeafNode())
		return s.Pop();
	while(!(tmp = s.Peek())->IsLeafNode()) {
		s.Push(&marker);
		for ( i = tmp->iNumChildren - 1; i >= 0; i-- ) 
			s.Push(tmp->children[i]);
	}
	return s.Pop();
}

PreOrderTreeIterator::PreOrderTreeIterator(Node *n)
{
	s.Push(n);
}

PreOrderTreeIterator::PreOrderTreeIterator()
{
}

void PreOrderTreeIterator::operator=(Node * n)
{
	s.Push(n);
}

Node * PreOrderTreeIterator::operator()()
{
	Node * tmp;
	int i;

	if (s.Peek() == 0)
		return 0;
	if (s.Peek() == &marker) {
		return s.Pop();
	}
	if (s.Peek()->IsLeafNode())
		return s.Pop();

	tmp = s.Pop();
	s.Push(&marker);

	for ( i = tmp->iNumChildren - 1 ; i >= 0; i-- ) 
		s.Push(tmp->children[i]);

	return tmp;
}

int PreOrderTreeIterator::IsParen(Node *p)
{
	return p == &marker;
}

LeafIterator::LeafIterator(Node *n)
{
	next = n;
}

Node * LeafIterator::operator()()
{
	Node * tmp;

	/* repeat until we find a node which is NOT null,
		and which is a leaf node */

	while ((tmp = next()) && !tmp->IsLeafNode());

	return tmp;
}

Node * Node::GetLowestNode(int LeftMost)
{
	Node * prevn = this;
	Node *n = this;

	if (LeftMost) {
		while (n = n->LeftChild())
			prevn = n;
	} else {
		while (n = n->RightChild())
			prevn = n;
	}

	return prevn;
}

Node * Node::GetNonTerminal(int LeftMost)
{
	int i;
	Node *NodeTmp;

	if (iNumChildren == 0) {
		if (val->IsNonTerminal())
			return this;
		else
			return 0;
	} else if (LeftMost) {
		for (i = 0; i < iNumChildren; i++) {
			if (NodeTmp = children[i]->GetNonTerminal(LeftMost))
				return NodeTmp;	
		}
		return 0;
	} else {
		for (i = iNumChildren; i > 0; i++) {
			if (NodeTmp = children[i]->GetNonTerminal(LeftMost))
				return NodeTmp;	
		}
		return 0;
	}
}


Node * Tree::GetRoot()
{
	return Root;
}

int Tree::GetSentence(char *buf)
{	
	LeafIterator nextleaf = Root;
	Node *p;
	char *bufptr = buf;
	TreeProb = 1;
	SentenceLength = 0;
//	char buf[BUFSIZ];
	char *x;

	while (p = nextleaf()) {
		x = p->val->NoQuotesStringify();
		strcpy(bufptr,x);
		bufptr +=  strlen(x);
		//free(x);
		*bufptr++ = ' ';
		TreeProb *= p->prob;
		SentenceLength++;
	}
	*bufptr = '\0';
	return TRUE;
}

	

