#include "unify.h"
#include <ctype.h>

#define MAXLEN 100

bool do_occurs_check = FALSE;
// set to TRUE to avoid those pesky infinite loops in unification
// slower though.

int dag_arc_compare(Dag_Arc *a, Dag_Arc *b) {
  return (int) b->attribute - (int) a->attribute;
}

Atom& gen_var() {
  char buff[100];
  static int counter = 0;
  sprintf(buff, "_%d", counter++);
  Atom my_atom(buff);
  return my_atom;
}

Subst::Subst(Atom var, Dag *val)  : variable(var), value(val) {
}

Dag_Arc::Dag_Arc(Dag_Arc &old) : attribute(old.attribute), value(old.value) {
}

void Dag_Arc::write(ostream &os, int indent) {
  os << attribute << ':' << '\t';
  if (value->is_atomic())
    value->write(os, indent);
  else {
    os << endl;
    value->write(os, indent);
  }    
}

Dag_Arc::~Dag_Arc() {
//  delete value;  // flush_the_toilet instead
}

void Subst::write(ostream &os) { 
  variable.write(os); os << endl; value->write(os); 
}

void SubstList::write(ostream &os) { 
  if (!this) {
    cout << "NULL SUBSTITUTION LIST\n";
    return;
  }
  for (reset(); !end(); next()) {
    Subst *s = cur();
    s->write(os);
  }
}

Dag::Dag(Dag &old) : arcs(dag_arc_compare), atomic(old.atomic) {
  if (atomic)
    sym = old.sym;
  else {
    // using sym as remainder
    sym = gen_var();
    for (old.arcs.reset(); !old.arcs.end(); old.arcs.next()) {
      Dag_Arc *oa = old.arcs.cur();
      add_arc(new Dag_Arc(*oa));
    }
  }
}

// deep copy of arcs and child dags - keeps remainder variables the same
void Dag::copy(Dag *d)  {
  if (!d) 
    return;
  atomic = d->atomic;
  sym = d->sym;
  if (!atomic) {
    for (d->arcs.reset(); !d->arcs.end(); d->arcs.next()) {
      Dag_Arc *oa = d->arcs.cur();
      Dag *ndag = new Dag;
      ndag->copy(oa->value);
      add_arc(new Dag_Arc(oa->attribute, ndag));
    }
  }
}

static void _indent(ostream &os, int indent)
{   /* print <indent> spaces to ostream (no os.form() under Solaris CC -dpwe) */
    int i;
    for(i = 0; i < indent; ++i) {
	os << " ";
    }
}

void Dag::write(ostream &os, int indent,int shouldend /* = 1 */) {
  const static int ind_incr = 11;
//  os << "Address: " << this << "\n";
  if (atomic) {
    sym.write(os);
    if (shouldend)
      os << endl;
  } else {
    for (arcs.reset(); !arcs.end(); arcs.next()) {
      //os.form("%*s", indent, "");
      _indent(os, indent);
      Dag_Arc *da = arcs.cur();
      da->write(os, indent + ind_incr);
    }
    //os.form("%*s", indent, "");
    _indent(os, indent);
    os << "REMAINDER: "; 
    sym.write(os);
    if (shouldend)
      os << endl;
  }
}

void Dag::read(istream &is /* = cin */) {
  char c;
  char buff[MAXLEN];
  is >> ws >> c;
  if (c != '(') {
    // if dag doesn't start with '(', it is atomic
    int i;
    for (i = 0; (!isspace(c)) && c !=  ')' && c != ']'; i++) {
      buff[i] = c;
      is.get(c);
    }
    buff[i] = '\0';
    //is.unget();
    is.putback(c);
    atomic = TRUE;
    sym = Atom(buff);
    return;
  }
  atomic = FALSE;
  // loop to read in individual features
  while (is >> ws >> c, c == '[') {
    is >> buff;
    char *colon = buff + strlen(buff) - 1;
    if (*colon != ':') {
      cerr << "Expected colon at end of feature name\n";
      return;
    }
    *colon = '\0';
    is >> ws;
    Dag *val = new Dag;
    val->read(is);
    is >> ws >> c;
    if (c != ']') {
      cerr << "Expected ']' at end of feature value\n";
      return;
    }
    add_arc(new Dag_Arc(buff, val));
  }
  if (c != ')') {
    cerr << "Expected ')' at end of feature list\n";
  }
}


bool Dag::destructive_unify(Dag *d) {
  if (is_atomic() ^ d->is_atomic())
    return FALSE;
  for (d->arcs.reset(); !d->arcs.end(); d->arcs.next()) {
    Dag_Arc *a = d->arcs.cur();
    Dag *val = value_of(a->attribute); // value of attr in this DAG
    if (val) {
      if (!val->destructive_unify(a->value))
	return FALSE;
    }
    else
      add_arc(new Dag_Arc(*a));
  }
  return TRUE;
}

/*
SubstList *Dag::unify(Dag *d) {
  SubstList *sl = new SubstList;
  if (combine_values(d, sl))
    return sl;
  else {
    delete sl;
    return NULL;
  }
}
*/

SubstList *Dag::unify(Dag *d) {
  SubstList *sl = new SubstList;
  return unify_with_sl(d, sl);
}

SubstList *Dag::unify_with_sl(Dag *d, SubstList *sl) {
/*  SubstList *result = sl->copy(); */
  SubstList *result = sl;
  Dag *d1 = get_real(sl);
  Dag *d2 = d->get_real(sl);
  if (d1 == d2)
    return result;
  if (do_occurs_check) {
    if ((d->is_var() && !occurs_check(d->sym, sl)) ||
	is_var() && !d->occurs_check(sym, sl))
      return FALSE;
  }
  if (!d1->atomic && d1->arcs.empty()) {  // handle empty dag
    d1->atomic = TRUE;
    d1->sym = gen_var();
    result->add_subst(d1->sym, d2);   /*****/
    return result;
  }
  if (!d2->atomic && d2->arcs.empty()) {  // handle empty dag
    d2->atomic = TRUE;
    d2->sym = gen_var();
    result->add_subst(d2->sym, d1);
    return result;
  }
  if (d1->combine_values(d2, result) && d2->combine_values(d1, result)) {
    if (!d1->atomic && !d2->atomic) {
      Atom oursym = d1->get_real_rem(result)->sym;
      Atom theirsym = d2->get_real_rem(result)->sym;
      if (oursym != theirsym)
	result->add_subst(oursym, new Dag(theirsym));
    }
    return result;
  } else {
//    delete result;
    return NULL;
  }
}

SubstList *Dag::unify_with_paths_and_sl(Dag *d, Path *p1, Path *p2, 
					SubstList *sl) {
  Dag *d1, *d2;
  d1 = make_path(p1, sl);
  d2 = d->make_path(p2, sl);

  if (!d1 || !d2)
    return NULL;     // tried to make path through an atom

  return d1->unify_with_sl(d2, sl);
}

// add attribute/value pair in da to remainder.
//  create new binding for deepest for remainder
void Dag::add_to_sl(char *attr, Dag *val, SubstList *sl) {
  Dag *rem_dag = get_real_rem(sl);
  Dag *ndag = new Dag;
  ndag->add_arc(new Dag_Arc(attr, val));
  sl->add_subst(rem_dag->sym, ndag);
}

// add attribute/value pairs in lda to remainder.
//  create new binding for deepest for remainder
// NOT CURRENTLY USED
void Dag::add_many_to_sl(List<Dag_Arc *> *lda, SubstList *sl) {
  Dag *rem_dag = get_real_rem(sl);
  Dag *ndag = new Dag;
  for (lda->reset(); !lda->end(); lda->next()) {
    Dag_Arc *da = lda->cur();
    ndag->add_arc(new Dag_Arc(da->attribute, da->value));
  }
  sl->add_subst(rem_dag->sym, ndag);
}


// control whether occurs check happens by setting 
// global variable do_occurs_check true or false
bool Dag::combine_values(Dag *d, SubstList *sl) {
  Dag *d1 = get_real(sl);
  Dag *d2 = d->get_real(sl);
  // d1 and d2 should be used for everything except occurs check
  if (d1->is_var()) {
    if (!d2->is_var() || d2->sym != d1->sym)
      sl->add_subst(d1->sym, d2);
    return TRUE;
  } else
    if (d2->is_var()) {
      sl->add_subst(d2->sym, d1);
      return TRUE;
    }
  if (d1->atomic || d2->atomic)
    // if one is a constant, both must be same constant
    return d1->atomic && d2->atomic && d1->sym == d2->sym;
  d1->arcs.reset();
  for (d1->arcs.reset(); !d1->arcs.end(); d1->arcs.next()) {
    Dag_Arc *a1 = d1->arcs.cur();
    Dag *val2 = d2->value_of_with_sl(a1->attribute, sl);
    if (val2) {
      if (!a1->value->unify_with_sl(val2, sl))
	// try to recursively unify values of same attribute
	return FALSE;
    } else {
      d2->add_to_sl(a1->attribute, a1->value, sl);
    }
/*
    int compare_result = dag_arc_compare(a1, a2);
    if (compare_result == 0) {
      if (!a1->value->combine_values(a2->value, sl))
	return FALSE;
      d->arcs.next();
      arcs.next();
    } else if (compare_result > 0) {
      // attribute not present in d, must add
      d->add_to_sl(a1, sl);
      arcs.next();
    } else if (compare_result < 0) {
      // attribute not present in this dag, must add
      add_to_sl(a2, sl);
      d->arcs.next();
    }
*/
  }
  Dag *rem_dag;
  if (rem_dag = sl->lookup(d1->sym)) {
    rem_dag = rem_dag->get_real(sl);
    if (!rem_dag->atomic)
      // also check features specified in remainder for unification
      rem_dag->combine_values(d2, sl);
  }
  return TRUE;
}

// follow chain of remainder bindings
Dag *Dag::get_real_rem(SubstList *sl) {
  Dag *next;
  if (next = sl->lookup(sym))
    return next->get_real_rem(sl);
  else
    return this;
}

Dag *Dag::get_real(SubstList *sl, int depth /* = 0 */) {
  Dag *next;
  if (atomic && (next = sl->lookup(sym)))
    return next->get_real(sl, depth + 1);
  else
    return this;
}  

bool Dag::copy_unify(Dag *d1, Dag *d2) {
  Dag_Arc *a1, *a2;
  Dag *ndag;
  if (d1 == NULL || d2 == NULL)
    return FALSE;
  if (d1->is_atomic()) {
    if (d2->is_atomic() && d1->sym == d2->sym) {
      atomic = TRUE;
      sym = d1->sym;
      return TRUE;
    } else {
      return FALSE;
    }
  } else if (d2->is_atomic()) {
    return FALSE;
  }
  atomic = FALSE;
  d1->arcs.reset();
  d2->arcs.reset();
  while (!d1->arcs.end() && !d2->arcs.end()) {
    a1 = d1->arcs.cur();    
    a2 = d2->arcs.cur();
    ndag = new Dag;
    int compare_result = dag_arc_compare(a1, a2);
    if (compare_result == 0) {
      if (!ndag->copy_unify(a1->value, a2->value))
	return FALSE;
      add_arc(new Dag_Arc(a1->attribute, ndag));
      d1->arcs.next();
      d2->arcs.next();
    } else if (compare_result < 0) {
      // attribute not present in d2, must add
      ndag->copy(a1->value);
      add_arc(new Dag_Arc(a1->attribute, ndag));
      d1->arcs.next();
    } else if (compare_result > 0) {
      // attribute not present in d1, must add
      ndag->copy(a2->value);
      add_arc(new Dag_Arc(a2->attribute, ndag));
      d2->arcs.next();
    }
  }
  while (!d1->arcs.end()) {
    // go through any leftovers in d1
    a1 = d1->arcs.cur();
    ndag = new Dag;
    ndag->copy(a1->value);
    add_arc(new Dag_Arc(a1->attribute, ndag));
    d1->arcs.next();
  }
  while (!d2->arcs.end()) {
    // go through any leftovers in d2
    a2 = d2->arcs.cur();
    ndag = new Dag;
    ndag->copy(a2->value);
    add_arc(new Dag_Arc(a2->attribute, ndag));
    d2->arcs.next();
  }
  return TRUE;
}

Dag *Dag::fill_in(SubstList *sl, bool debug) {
  Dag *real = get_real(sl);
  if (real->atomic)
    return real;
  if (debug) {
    cout << "fill_n:\n";
    write();
  }
  Dag *ndag = new Dag;           /* deep copy so as not to alter other dags */
  ndag->sym = real->sym;
  real->fill_remainder(ndag, sl, debug); 
  /* add stuff down chain of remainders */
  for (ndag->arcs.reset(); !ndag->arcs.end(); ndag->arcs.next()) {
    Dag_Arc *da = ndag->arcs.cur();
    da->value = da->value->fill_in(sl, debug);
  }
  for (real->arcs.reset(); !real->arcs.end(); real->arcs.next()) {
    /* now add in attributes from real */
    Dag_Arc *da = real->arcs.cur();
    ndag->add_arc(new Dag_Arc(da->attribute, da->value->fill_in(sl, debug)));
  }
  return ndag;
}

void Dag::fill_remainder(Dag *add_to, SubstList *sl, bool debug) {
  Dag *rem_dag = sl->lookup(sym);
  if (!rem_dag)
    return;
  rem_dag = rem_dag->get_real(sl);
  if (debug) { cout << " fill  remainder ";
	       sym.write(); cout << endl; }
  for (rem_dag->arcs.reset(); !rem_dag->arcs.end(); rem_dag->arcs.next()) {
    Dag_Arc *da = rem_dag->arcs.cur();
    if (!add_to->value_of(da->attribute)) {  // prevent duplicates
      add_to->add_arc(new Dag_Arc(da->attribute, da->value));
    } else
      cerr << "Attempt to fill in with duplicate attribute\n";
  }
  add_to->sym = rem_dag->sym;
  rem_dag->fill_remainder(add_to, sl, debug);
}

Dag *Dag::value_of(char *attr) {
  for (arcs.reset(); !arcs.end(); arcs.next()) {
    Dag_Arc *da = arcs.cur();
    if (da->attribute == attr)
      return da->value;
  }
  return NULL;
}

Dag *Dag::value_of_with_sl(char *attr, SubstList *sl) {
  Dag *result;
  if (result = value_of(attr)) {
    return result->get_real(sl);
  }
  if (result = sl->lookup(sym))
    return result->value_of_with_sl(attr, sl);
  else 
    return NULL;           // no value found
}    

Dag *Dag::get_path(Path *p) {
  Dag *d = this;
  for (p->reset(); d != NULL && !p->end(); p->next()) {
    d = d->value_of(p->cur());
  }
  return d;
}

Dag *Dag::make_path(Path *p, SubstList *sl) {
  Dag *last = this, *next = this;
  bool bail = FALSE;
  for (p->reset(); !bail && !p->end(); p->next()) {
    if (last->atomic && !last->sym.is_var())
      return NULL;              // can't make path through an atomic dag
    next = last->value_of_with_sl(p->cur(), sl);
    if (!next) {
      next = new Dag;
      last->add_to_sl(p->cur(), next, sl);
      bail = TRUE;
    }
    last = next;
  }
  if (bail) {
    for (; !p->end(); p->next()) {
      next = new Dag;
      Dag_Arc *da = new Dag_Arc(p->cur(), next);
      last->add_arc(da);
      last = next;
    }
    last->atomic = TRUE;  // make into variable rather than empty dag
  }
  return last;
}

bool Dag::copy_path_unify(Dag *d1, Dag *d2, Path *p1, Path *p2) {
  d1 = d1->get_path(p1);
  d2 = d2->get_path(p2);
  if (!d1)
    d1 = new Dag;
  if (!d2)
    d2 = new Dag;
  return copy_unify(d1, d2);
}

// check if a occur in this dag (this dag is a)
// behaves identically whether or not a is a variable
bool Dag::occurs_check(Atom a, SubstList *sl) {
  //cout << "occurs check"; a.write(); cout << endl; write();
  if (sym == a)
    return FALSE;
  for (arcs.reset(); !arcs.end(); arcs.next()) {
    if (!arcs.cur()->value->occurs_check(a, sl))
      return FALSE;
  }
  Dag *rem_dag;
  if (rem_dag = sl->lookup(sym)) {   // handles both vars and remainders
    return rem_dag->occurs_check(a, sl);
  }
  return TRUE;
}

Dag *make_nonterminal(char *cat) {
  Dag *result = new Dag;
  Dag *cat_atom = new Dag(Atom(cat));
  Dag_Arc *da = new Dag_Arc("cat", cat_atom);
  result->add_arc(da);
  return result;
}

// memory routines follow
DStack<Dag *> Dag::free_list(1000);
DStack<Dag *> Dag::busy_list(1000);

void *Dag::operator new(size_t n) {
  Dag *result;
  if (free_list.empty())
    result = (Dag *) malloc(sizeof(Dag));
  else
    result = free_list.pop();
  busy_list.push(result);
  return result;
}

// to be called by flush_the_toilet only!!
void Dag::operator delete(void *s, size_t n) {
  if (s != NULL) {
    if (free_list.count() < 10000)
      free_list.push((Dag *) s);
    else
      free(s);
  }
}

void Dag::flush_the_toilet() {
  while (busy_list.count())
    delete busy_list.pop();
}


void Dag::clobber_free_list() {
  while (!free_list.empty())
    free(free_list.pop());
}


void Path::read(istream &is /* = cin */) {
  char c;
  char buff[MAXLEN];
  is >> ws >> c;
  if (c != '(') {
    cerr << "Expected '(' at beginning of path\n";
    return;
  }
  // loop to read in individual edges
  while (is >> ws >> c, c != ')') {
    int i;
    for (i = 0; (!isspace(c)) && c !=  ')' && c != ']'; i++) {
      buff[i] = c;
      is.get(c);
    }
    buff[i] = '\0';
    //is.unget();
    is.putback(c);
    add(buff);  // add to end of list of feature names
  }
  if (c != ')') {
    cerr << "Expected ')' at end of feature list\n";
  }
}

char *Path::stringread(char *string) {
  char *p=string;
  char buff[MAXLEN];
  
  if (*p != '(') {
    cerr << "Expected '(' at beginning of path\n";
    return NULL;
  }
  p++;
  while(isspace(*p)) p++;
  // loop to read in individual edges
  while (*p != ')') {
    int i;
    for (i = 0; (!isspace(*p)) && *p !=  ')' && *p != ']'; i++) {
      buff[i] = *p;
      p++;
    }
    buff[i] = '\0';
    add(buff);  // add to end of list of feature names
    while(isspace(*p)) p++;
  }
  if (*p != ')') {
    cerr << "Expected ')' at end of feature list\n";
    return NULL;
  }
  p++;
  return p;
}

void Path::write(ostream &os /* = cout */) {
  os << "( ";
  for (reset(); !end(); next()) {
    os << cur() << ' ';
  }
  os << ")\n";
}

void Path::write(int num,ostream &os /* = cout */) {
  os << "( " << num << " ";
  for (reset(); !end(); next()) {
    os << cur() << ' ';
  }
  os << ")";
}


DStack<void *> Link_free_list(10000);
/* used by Link class within List template */
