#ifndef _UNIFY_H_
#define _UNIFY_H_

#include <stdio.h>
#include <string.h>
#include "unify_list.h"
#include "unify_hash.h"

class Atom {
  char *name;
public:
  char *get_name() { return name; }
  bool is_var() { return *name == '_'; }
  void write(ostream &os = cout) { os << name; }
  Atom() : name(NULL) {}
  Atom(char *s) { name = StringTableEntry(s); }
  Atom(Atom &a) { name = a.name; }
  bool operator == (Atom &s) { return name == s.name; }
  bool operator != (Atom &s) { return name != s.name; }
  operator const char * () { return name; } 
};

Atom& gen_var();

class Dag;
class Path;

class Dag_Arc {
public:
  char *attribute;
  Dag *value;
  void write(ostream &os = cout, int indent = 0);
  Dag_Arc() {}
  Dag_Arc(Dag_Arc &old);
  Dag_Arc(char *att, Dag *v) : attribute(StringTableEntry(att)), value(v) {}
  ~Dag_Arc();
};

int dag_arc_compare(Dag_Arc *a, Dag_Arc *b);

class Subst {
public:
  Atom variable;
  Dag *value;
  Subst(Subst &s) : variable(s.variable), value(s.value) {}
  Subst(Atom var, Dag *val);
  void write(ostream &os = cout);
}; 

class SubstList : private List<Subst *> {
public:
  Dag *lookup(Atom &var) {
    for (reset(); !end(); next()) {
      Subst *s = cur();
      if (s->variable == var)
	return s->value;
    }
    return NULL;
  }
  void add_subst(Atom var, Dag *val) { 
    if (!var.is_var())
      cout << "Error: adding a non-variable to the substitution list.\n";
    add(new Subst(var, val)); }
  SubstList *copy() {
    if (!this) return NULL;
    SubstList *result = new SubstList;
    for (reset(); !end(); next())
      result->add(new Subst(*cur()));
    return result;
  }
  ~SubstList() { clobber(); }
  void write(ostream &os = cout);
};

class Dag {
 public:
  char *get_sym() { return sym.get_name(); }
  void add_arc(Dag_Arc *da) {
    arcs.add(da);
  }

  void write(ostream &os = cout, int indent = 0, int shouldend=1);
  
  void read(istream &is = cin);
  bool is_atomic() { return atomic; }
  bool is_var() { return atomic && sym.is_var(); }
  Dag(char *s) : arcs(dag_arc_compare),
                atomic(TRUE), sym(s) {  busy_list.pop(); /* not to be freed */}
  Dag(Atom s) : arcs(dag_arc_compare), sym(s), atomic(TRUE) {}
  Dag(Dag &old);
  Dag() : arcs(dag_arc_compare), sym(gen_var()) 
    { atomic = FALSE; }
  ~Dag() { arcs.clobber(); }
  void copy(Dag *d);                           // recursive deep copy
  bool destructive_unify(Dag *d);
  SubstList *unify(Dag *d);                    
     // unify this dag with d nondestructively, return sub. list
  SubstList *unify_with_sl(Dag *d, SubstList *sl);
     // as above, but with sl to look things up in
  SubstList *unify_with_paths_and_sl(Dag *d, Path *p1, Path *p2, 
				     SubstList *sl);
     // the sl version of copy_path_unify
  bool copy_unify(Dag *d1, Dag *d2);
  Dag *fill_in(SubstList *sl, bool debug = FALSE);
  void fill_remainder(Dag *add_to, SubstList *sl, bool debug = FALSE);
  Dag *value_of(char *attr);
  Dag *value_of_with_sl(char *attr, SubstList *sl);
       // follow remainder chains and mulitple substitutions
  Dag *get_path(Path *p);
  Dag *make_path(Path *p, SubstList *sl);
       // like get_path, but add to sl if path not in dag
  bool copy_path_unify(Dag *d1, Dag *d2, Path *p1, Path *p2);
  Dag *get_real(SubstList *sl, int depth = 0);
  static void flush_the_toilet();
  void *operator new(size_t n);
  void operator delete(void *s, size_t n);
  static void clobber_free_list();
 private:
  void add_to_sl(char *attr, Dag *val, SubstList *sl);
       // add to remainder, making new binding for remainder
  void add_many_to_sl(List<Dag_Arc *> *lda, SubstList *sl);
       // add to remainder, making new binding for remainder
  bool combine_values(Dag *d, SubstList *sl);
  Dag *get_real_rem(SubstList *sl);
  bool occurs_check(Atom a, SubstList *sl);
       // follow chain of remainder bindings
  // DATA
  bool atomic;
  Atom sym;         // value if atomic, remainder if not
  SortedList<Dag_Arc *> arcs;
  static DStack<Dag *> free_list;
  static DStack<Dag *> busy_list;
};

Dag *make_nonterminal(char *cat);
// make new dag with one arc - cat <parameter>

class Path : public List<char *> {
public:
  void add(char *sym) { List<char *>::add(StringTableEntry(sym)); }
  void write(ostream &os = cout);
  void write(int head, ostream &os = cout);
  void read(istream &is = cin);
  char *stringread(char *);
};

extern int ConstraintNews;
extern int ConstraintDeletes;

class Constraint {
 public:
  int head;
  int head1;
  Path *path1;
  int head2;
  Path *path2;
  Dag *dag2;

  Constraint(int h1, int h2, Path *p1, Path *p2, Dag *d2) : 
    head1(h1), head2(h2), path1(p1), path2(p2), dag2(d2) {
      ConstraintNews++;
      head=(head1>head2)?head1:head2;
    }
  friend int Constraint_compare(Constraint *c1, Constraint *c2);
  ~Constraint() {
    ConstraintDeletes++;
    if (path1) {delete path1; path1=NULL;}
    if (path2) {delete path2; path2=NULL;}
    if (dag2) {delete dag2; dag2=NULL;}
  }
  void write(ostream &os=cout) {
    os << "(";
    path1->write(head1,os);
    os << " = ";
    if (head2>=0)
      path2->write(head2,os);
    else
      dag2->write(os,0,0);
    os << ")";
    return;
  }
      
};

extern int ConstraintListNews;
extern int ConstraintListDeletes;

class ConstraintList : public SortedList<Constraint *> {
 public:
  ConstraintList(int (*cmpf)(Constraint *,Constraint *), char *string,
		 int max);
  ~ConstraintList() { ConstraintListDeletes++;}
  friend class ConstraintIterator;
};

class ConstraintIterator : public Iterator<Constraint *> {
 public:
  ConstraintIterator(ConstraintList *c) {if (c) current=c->first; else current=NULL;};
  ConstraintIterator() {};
};


#endif /* _UNIFY_H_ */
