#
# treecmds.tcl
#
# Define some new Tcl commands to simulate the Tcl extensions in 'treesh'
#
# 1996sep20 dpwe@icsi.berkeley.edu
#
# $Header: /u/drspeech/src/berpdemo/share/RCS/treecmds.tcl,v 1.2 1999/05/12 00:34:02 dpwe Exp $
#

# There's this part of the BeRP demo that displays the parse tree, 
# but it uses a special version of wish with a set of built-ins 
# for displaying trees on a canvas.  But how hard is it to do this 
# in Tcl?  Well, let's try it and find out.

# Looking at the code in ptbrowser, the commands we need appear to be:
#
# tree_getCurrent $tree - return a name for a tree structure
# $tree rmlink $link
# $tree draw
# tree_addNode $tree $parent $child $bitmap $arcname
# (tree_center $tree) not actually used
# tree_create $tree $text_proc $bitmap_proc
# $tree configure -layout vertical

# What is to be our data structure?  I suppose a nested list of the
# form (name bitmap (child child child child ...)) where child can be the same
# So, then, how do we draw it?

# fns from dpweutils.tcl (a/k/a ~dpwe/share/lib/tcl/utilfns.tcl)
proc CanvTextDims {cnv {text "N"} {font {}}} {
    # return the w and h for the given text & font if drawn in the given 
    # canvas.  If omitted, font uses default canvas font; 
    # With no text specified, returns the size of an upper-case N.
    if { $font == "" }  {
	# no font specified, allow default to happen
	set tag [$cnv create text 10 10 -text $text]
    } else {
	set tag [$cnv create text 10 10 -text $text -font $font]
    }
    set bbx [$cnv bbox $tag]
    $cnv delete $tag
    return [list [expr "[lindex $bbx 2]-[lindex $bbx 0]"] \
	    [expr "[lindex $bbx 3]-[lindex $bbx 1]"]]
}

proc lassign {list args} {
    # take a list and some number of variable names; set the 
    # variables to the corresponding elements of the list
    set i 0
    foreach arg $args {
	set line [lindex $list $i]
	# escape quotes
        uplevel 1 "set $arg \{$line\}"
	incr i
    }
}

proc max args {
    # Treats args as a list of numbers, returns the largest
    set list [eval "concat $args"]
    set rslt [lindex $list 0]
    foreach el $list {
	if { $el > $rslt } {
	    set rslt $el
	}
    }
    return $rslt
}

proc min args {
    # Treats args as a list of numbers, returns the smallest
    set list [eval "concat $args"]
    set rslt [lindex $list 0]
    foreach el $list {
	if { $el < $rslt } {
	    set rslt $el
	}
    }
    return $rslt
}


# Here's an array to keep our trees in 
set tree_priv(.) {}

proc tree_create {tree textProc iconProc} {
    # Create a new tree 
    global tree_priv
    set tree_priv($tree) {}
    # Define a tree command
    proc $tree {args} "
	set cmd \[lindex \$args 0\]
	if {\$cmd == \"draw\"} {
	    tree_draw $tree
	} elseif {\$cmd == \"rmlink\"} {
	    tree_rmlink $tree \[lindex \$args 1\]
	} elseif {\$cmd == \"configure\"} {
	} else {
	    puts stderr \"$tree: unknown cmd: \$cmd\"
	}
    "
}

proc tree_addNode {tree parent child bitmap arcname} {
    # Add a new node to a tree
    global tree_priv
    if {$parent == ""} {
	# Rewriting root is special case - reset tree
	set tree_priv($tree) [list $child $bitmap $arcname {}]
    } else {
	set tree_priv($tree) [tree_add_sub $tree_priv($tree) $parent $child \
	                   $bitmap $arcname]   
    }
}

proc tree_add_sub {list parent childname bitmap arcname} {
    # return a copied subtree, but add the named node if you find its parent
    # Arcname currently ignored
    lassign $list name bitm labl chil
    set newchil {}
    foreach child $chil {
	lappend newchil [tree_add_sub $child $parent $childname \
		                      $bitmap $arcname]
    }
    # Add the new node if we are the parent
    if {$parent == $name} {
	lappend newchil [list $childname $bitmap $arcname {}]
    }
    return [list $name $bitm $labl $newchil]
}

proc tree_getCurrent {tree} {
    # return the name of the specified tree
    global tree_priv
    set list $tree_priv($tree)
    lassign $list name bitm labl chil
    return $labl
}

proc tree_rmlink {tree link} {
    # add a link to a tree
    global tree_priv
    set tree_priv($tree) [tree_rm_sub $tree_priv($tree) $link]
}

proc tree_rm_sub {list rmname} {
    # return the tree but without the named node, if any (recursive)
    lassign $list name bitm labl chil
    if {$name == $rmname} {
	return ""
    } else {
	set newchil {}
	foreach child $chil {
	    set childout [tree_rm_sub $child $rmname]
	    if {$childout != ""} {
		lappend newchil $childout
	    }
	}
	return [list $name $bitm $labl $newchil]
    }
}

proc tree_draw {tree} {
    # Render a tree in a canvas
    global tree_priv
    # the tree definition comes from our internal structure
    set list $tree_priv($tree)
    # the canvas name is the stem of the tree name
    regsub {\.[^.]+$} $tree "" cnv
    # delete the existing tree
    $cnv delete $tree
    # Recursively draw the tree
    tree_draw_sub $tree $list {10 10}
    $cnv config -scrollregion [$cnv bbox all]
}

proc tree_draw_sub {tree list {pos {0 0}} {minctr -999}} {
    # Part of the recursive tree-drawing business
    # If we're at a terminal node, draw & return
    # else open first node.
    # $pos gives the top-left of where to draw the node
    # Returns two x values, for the center and extreme right of drawn node
#puts stderr "tree_draw_sub '$list' '$pos'"
    # the canvas name is the stem of the tree name
    regsub {\.[^.]+$} $tree "" cnv
    lassign $list name bitm labl chil
    lassign $pos x y
    # Vertical height of generations
#    set genheight 70
    set genheight 40
    set theight 20
    set bmheight 20
    set wgap 10
    set childy [expr $y+$genheight]
    set childcenterxs {}
    # Width of this label
    set mytwidth [expr [lindex [CanvTextDims $cnv $labl] 0]+$wgap]
    # leftmost acceptable center for child
    set minchctr [max $minctr [expr $x+$mytwidth/2]]
    if {$chil != {} } {
	# Not a terminal node.  Draw all the children
	set childx $x
	foreach child $chil {
	    set rpos [tree_draw_sub $tree $child [list $childx $childy] \
		                    $minchctr]
	    lappend childcenterxs [lindex $rpos 0]
	    set childx [lindex $rpos 1]
	}
	# Make the actual width that of the child, or me
	set mywidth [max $mytwidth [expr $childx - $x]]
	set centerx [expr ([min $childcenterxs]+[max $childcenterxs])/2]
    } else {
	# Terminal node - render the thing
	set mywidth $mytwidth
	set centerx [max $minctr [expr $x+$mywidth/2]]
    }
    # All the children are drawn.  Draw my node, connect to children
#set bitm "" ;# No bitmaps for now
    if {$bitm != ""} {
	set bb [$cnv bbox \
	  [$cnv create bitmap $centerx $y -bitmap @$bitm -anchor n \
	  -tag [list $tree $name]]]
	incr y [expr [lindex $bb 3] - [lindex $bb 1]]
    }
    if {$labl != ""} {
	set bb [$cnv bbox \
	       [$cnv create text $centerx $y -text $labl -anchor n \
	       -tag [list $tree $name]]]
	incr y [expr [lindex $bb 3] - [lindex $bb 1]]
    }
    # If we had children, draw links to them all
    set i 0
    foreach childx $childcenterxs {
	set childname [lindex [lindex $chil $i] 0]
	incr i
	$cnv create line $centerx $y $childx $childy \
		-tag [list $tree $childname]
    }
#puts "**: $labl $x $centerx [expr $x+$mywidth]"
    return [list $centerx [expr $centerx+$mywidth/2]]
}

proc tree_node_color {tree node color} {
    # Use the tags we set up to color a particular node
    # "node" is the "name" field of the created node
    regsub {\.[^.]+$} $tree "" cnv
    set ids [$cnv find withtag $node]
    foreach id $ids {
	set type [$cnv type $id]
	if {$type == "bitmap"} {
	    $cnv itemconfigure $id -foreground $color
	} else {
	    $cnv itemconfigure $id -fill $color
	}
    }
}
