/*
   $Header: /u/spert/src/fltvec/testsuite/RCS/vecele_fvf_vf_test.c,v 1.2 1995/07/05 23:47:30 davidj Exp $

   Test element wise vector routines of the form op_vfvf_vf.
*/

#include <stddef.h>
#include <stdlib.h>

#include "fltvec_test.h"

typedef void (*rtn_t)(size_t len, float a, const float* b,
		      float* res);

/* Performs one test given input values and expected output. */

void
onetest(rtn_t rtn, size_t len,
	float a, const float* b, const float* correct)
{
    /* Create padded copies of inputs. */
    float* c_b = rtst_padvec_clone_vf(len, b);
    float* c_res = rtst_padvec_new_vf(len);

    /* Run routine. */
    rtn(len, a, c_b, c_res);

    /* Check result. */
    rtst_checknear_fvfvf(len, 1e-6, c_res, correct);

    /* Check input not corrupted. */
    rtst_checkeq_vfvf(len, b, c_b);

    /* Free and check padding. */
    rtst_padvec_del_vf(c_b);
    rtst_padvec_del_vf(c_res);
}

void
sub_1(rtn_t rtn)
{
    const float a = 3.0f;
    const float b[] = { 2.0f };
    const float correct[] = { 1.0f };

    onetest(rtn, 1, a, b, correct);
}

void
sub_3(rtn_t rtn)
{
    const float a = 2.1f;
    const float b[] = { 3.0f, 2.0f, 1.0f };
    const float correct[] = { -0.9f, 0.1f, 1.1f };

    onetest(rtn, 3, a, b, correct);
}

void
div_1(rtn_t rtn)
{
    const float a = 3.0f;
    const float b[] = { 2.0f };
    const float correct[] = { 1.5f };

    onetest(rtn, 1, a, b, correct);
}

void
div_3(rtn_t rtn)
{
    const float a = -0.5f;
    const float b[] = { 2.0f, 0.2f, -1.0f };
    const float correct[] = { -0.25f, -2.5f, 0.5f };

    onetest(rtn, 3, a, b, correct);
}

/* Compare the results of two vector routines */

void
random_test(rtn_t nv_rtn, rtn_t test_rtn, int nozero)
{
    const int numtests = rtst_numtests;
    const size_t maxlen = rtst_sizetests;
    float a;
    float *b = malloc_vf(maxlen);
    float *res = malloc_vf(maxlen);

    int i;
    for (i=0; i<numtests; i++)
    {
        size_t len = (size_t) rtst_urand_i32i32_i32(1, maxlen);

	a = rtst_urand_ff_f(-1e8, 1e8);
        rtst_urand_ff_vf(len, -1e8, 1e8, b);
	if (nozero)
	{
	    size_t j;

	    for (j=0; j<maxlen; j++)
	    {
		if (b[j]==0.0)
		    b[j] = 0.01;
	    }
	}
        nv_rtn(len, a, b, res);

        rtst_log("len=%d\n", len);

        onetest(test_rtn, len, a, b, res);
    }

    free_vf(b);
    free_vf(res);
}

void
test_sub(void)
{
    rtst_start("sub_fvf_vf");
    rtst_log("testing nv.\n");
    sub_1(nv_sub_fvf_vf);
    sub_3(nv_sub_fvf_vf);
#ifdef CHECK_TOR
    rtst_log("testing tor.\n");
    sub_1(tor_sub_fvf_vf);
    sub_3(tor_sub_fvf_vf);
    random_test(nv_sub_fvf_vf, tor_sub_fvf_vf, 0);
#endif
    rtst_passed();
}

void
test_div(void)
{
    rtst_start("div_fvf_vf");
    rtst_log("testing nv.\n");
    div_1(nv_div_fvf_vf);
    div_3(nv_div_fvf_vf);
#ifdef CHECK_TOR
    rtst_log("testing tor.\n");
    div_1(tor_div_fvf_vf);
    div_3(tor_div_fvf_vf);
    random_test(nv_div_fvf_vf, tor_div_fvf_vf, 1);
#endif
    rtst_passed();
}


int
main(int argc, char *argv[])
{
    rtst_args(argc, argv);

    test_sub();
    test_div();

    rtst_exit();

    return 0;
}
