/*
   $Header: /u/spert/src/fltvec/testsuite/RCS/vecele_vff_vf_test.c,v 1.2 1995/07/05 23:47:41 davidj Exp $

   Test element wise vector routines of the form op_vfvf_vf.
*/

#include <stddef.h>
#include <stdlib.h>

#include "fltvec_test.h"

typedef void (*rtn_t)(size_t len, const float* a, float b,
		      float* res);

/* Performs one test given input values and expected output. */

void
onetest(rtn_t rtn, size_t len,
	const float* a, float b, const float* correct)
{
    /* Create padded copies of inputs. */
    float* c_a = rtst_padvec_clone_vf(len, a);
    float* c_res = rtst_padvec_new_vf(len);

    /* Run routine. */
    rtn(len, c_a, b, c_res);

    /* Check result. */
    rtst_checknear_fvfvf(len, 1e-6, c_res, correct);

    /* Check input not corrupted. */
    rtst_checkeq_vfvf(len, a, c_a);

    /* Free and check padding. */
    rtst_padvec_del_vf(c_a);
    rtst_padvec_del_vf(c_res);
}

void
add_1(rtn_t rtn)
{
    const float a[] = { 1.0f };
    const float b = 2.0f;
    const float correct[] = { 3.0f };

    onetest(rtn, 1, a, b, correct);
}

void
add_3(rtn_t rtn)
{
    const float a[] = { -1.0f, 2.0f, 3.5f };
    const float b = 2.5f;
    const float correct[] = { 1.5f, 4.5f, 6.0f };

    onetest(rtn, 3, a, b, correct);
}

void
sub_1(rtn_t rtn)
{
    const float a[] = { 3.0f };
    const float b = 2.0f;
    const float correct[] = { 1.0f };

    onetest(rtn, 1, a, b, correct);
}

void
sub_3(rtn_t rtn)
{
    const float a[] = { 3.0f, 2.0f, 1.0f };
    const float b = 2.1f;
    const float correct[] = { 0.9f, -0.1f, -1.1f };

    onetest(rtn, 3, a, b, correct);
}

void
mul_1(rtn_t rtn)
{
    const float a[] = { 2.0f };
    const float b = 3.0f;
    const float correct[] = { 6.0f };

    onetest(rtn, 1, a, b, correct);
}

void
mul_3(rtn_t rtn)
{
    const float a[] = { -3.0f, 0.0f, 0.1f };
    const float b = 2.0f;
    const float correct[] = { -6.0f, 0.0f, 0.2f };

    onetest(rtn, 3, a, b, correct);
}

void
div_1(rtn_t rtn)
{
    const float a[] = { 3.0f };
    const float b = 2.0f;
    const float correct[] = { 1.5f };

    onetest(rtn, 1, a, b, correct);
}

void
div_3(rtn_t rtn)
{
    const float a[] = { 3.0f, 2.0f, 0.0f };
    const float b = -0.5f;
    const float correct[] = { -6.0f, -4.0f, -0.0f };

    onetest(rtn, 3, a, b, correct);
}

/* Compare the results of two vector routines */

void
random_test(rtn_t nv_rtn, rtn_t test_rtn, int nozero)
{
    const int numtests = rtst_numtests;
    const size_t maxlen = rtst_sizetests;
    float *a = malloc_vf(maxlen);
    float b;
    float *res = malloc_vf(maxlen);

    int i;
    for (i=0; i<numtests; i++)
    {
        size_t len = (size_t) rtst_urand_i32i32_i32(1, maxlen);

        rtst_urand_ff_vf(len, -1e8, 1e8, a);
	b = rtst_urand_ff_f(-1e8, 1e8);
	if (nozero)
	{
	    if (b==0.0)
		b = 0.01;
	}
        nv_rtn(len, a, b, res);

        rtst_log("len=%d\n", len);

        onetest(test_rtn, len, a, b, res);
    }

    free_vf(a);
    free_vf(res);
}

void
test_add(void)
{
    rtst_start("add_vff_vf");
    rtst_log("testing nv.\n");
    add_1(nv_add_vff_vf);
    add_3(nv_add_vff_vf);
#ifdef CHECK_TOR
    rtst_log("testing tor.\n");
    add_1(tor_add_vff_vf);
    add_3(tor_add_vff_vf);
    random_test(nv_add_vff_vf, tor_add_vff_vf, 0);
#endif
    rtst_passed();
}

void
test_sub(void)
{
    rtst_start("sub_vff_vf");
    rtst_log("testing nv.\n");
    sub_1(nv_sub_vff_vf);
    sub_3(nv_sub_vff_vf);
#ifdef CHECK_TOR
    rtst_log("testing tor.\n");
    sub_1(tor_sub_vff_vf);
    sub_3(tor_sub_vff_vf);
    random_test(nv_sub_vff_vf, tor_sub_vff_vf, 0);
#endif
    rtst_passed();
}

void
test_mul(void)
{
    rtst_start("mul_vff_vf");
    rtst_log("testing nv.\n");
    mul_1(nv_mul_vff_vf);
    mul_3(nv_mul_vff_vf);
#ifdef CHECK_TOR
    rtst_log("testing tor.\n");
    mul_1(tor_mul_vff_vf);
    mul_3(tor_mul_vff_vf);
    random_test(nv_mul_vff_vf, tor_mul_vff_vf, 0);
#endif
    rtst_passed();
}

void
test_div(void)
{
    rtst_start("div_vff_vf");
    rtst_log("testing nv.\n");
    div_1(nv_div_vff_vf);
    div_3(nv_div_vff_vf);
#ifdef CHECK_TOR
    rtst_log("testing tor.\n");
    div_1(tor_div_vff_vf);
    div_3(tor_div_vff_vf);
    random_test(nv_div_vff_vf, tor_div_vff_vf, 1);
#endif
    rtst_passed();
}


int
main(int argc, char *argv[])
{
    rtst_args(argc, argv);

    test_add();
    test_sub();
    test_mul();
    test_div();

    rtst_exit();

    return 0;
}
