// -*- C++ -*-
//
// nSurface.H
// 
// Substitute for Surface.H, no weird stuff
//
// Delcares this single hierarchy:   floatRef     len, *data, step
//                                   floatVec     new/delete of *data
//                                   floatArray   rows, cols
//				     Surface      (time) base
//				     Patch        (same as surface for now)
// 
// Object that holds a 2-D array of data, corresponding to an energy surface
//
// dpwe@media.mit.edu  1994nov29
// $Header: /u/drspeech/src/guitools/farray_otcl/RCS/nSurface.H,v 1.5 1998/08/20 01:53:44 dpwe Exp $

#ifndef _SURFACE_H
#define _SURFACE_H

#include "dpwedefs.H"	// for min, max etc

#include "FloatVecOps.H"	// defines floatRef + floatRef etc.

#include <string.h>	// for memset etc
#include <assert.h>

// Should we abort on out-of-range accesses?
#ifdef ABORT_ON_OOR
#define ABORT_OR_RTZ	abort()
#else /* !ABORT_ON_OOR */
#define ABORT_OR_RTZ	return 0
#endif /* ABORT_ON_OOR */

// dB-type constants
#define TWENTYONLN10	(8.68588963806504)
#define LN10ONTWENTY	(0.115129254649702)

// Dimensions of the basic data set (??)
#define NUM_FRQ_CHANS    (46)
#define NUM_TIME_STEPS   (1000)

#include <iostream.h>	// for Print 
#include <math.h>    	// for log, exp, sqrt
#include <stdlib.h>	// for abort()

typedef float floatCombinerFn(float, float);

class probRef;	// forward reference

class floatRef {
friend class probRef;
protected:
    int       len;
    float     *data;
    int       step;
static float  dummy;	// for OOR returns

    void      _Swap(floatRef& fr) 
                { float *td = data; data = fr.data; fr.data = td;
		  int ti = len; len  = fr.len;  fr.len  = ti;
		  ti = step; step = fr.step; fr.step = ti; }
    int	      _Check(int i) const
                { if(i<0 || i>=len)  { cerr << "floatRef: " \
		     << i << " bad for len " << len << endl; ABORT_OR_RTZ; }
		  return 1;}
    void      _Init(int l, float *d, int s=1)
                { len = l; data = d;  step = s; }
public:
    floatRef(int l=0, float *d=0, int s=1) : len(l), data(d), step(s)
                { }
    floatRef(const floatRef& fr)
                { len = fr.len; data = fr.data; step = fr.step; }
    ~floatRef(void)
                { }
    floatRef& operator = (const floatRef& fr)   // copy referred data 
                { if(this != &fr)  Copy(fr);  return *this; }

    // Create a new ref, returned by val, for subrange
    floatRef  Ref(int start = 0, int olen = -1) {
	floatRef _fr;
	if (olen < 0)	olen = len - start;
	assert(olen >= 0 && olen+start <= len);
	_fr.SetLen(olen); _fr.SetStep(step); _fr.SetData(&_El(start));
	return _fr; }

    float     Val(int i) const	   { float v = 0; if(_Check(i))
                                     v = *(data+step*i); return v;}
    float&    _El(int i)	   { return *(data+step*i); }
    float&    El(int i) 	   { if(_Check(i)) return _El(i); 
                                     else return floatRef::dummy; }
    float&    elem(int i) 	   { return El(i); }	// GNU compat
    float&    operator [] (int i)  { return El(i); }
    float     operator [] (int i) const   { return Val(i); }
    void      Set(int i, float f)  { El(i) = f; }
    void      Copy(const floatRef& fr)
		{ assert(len>0 || fr.len==0); // catches assignmt to `new' frs
		  int i, n = min(len, fr.len);
		  for(i=0; i<n; ++i)  {  _El(i) = fr.Val(i); } }
    int       CopyAt(const floatRef& fr, int start) {  // write into middle
	          int i, n = min(len-start, fr.len);
		  for(i=0; i<n; ++i) {  _El(i+start) = fr.Val(i);} 
                  return n; }
    int       CopyFrom(int start, const floatRef& fr) {  // read from middle
	          int i, n = min(fr.len-start, len);
		  for(i=0; i<n; ++i) {  _El(i) = fr.Val(i+start);} 
                  return n; }
    void      CloneRef(const floatRef& fr)  { len = fr.len; step = fr.step;
					      data = fr.data; }
    void      CloneRefFrom(int start, const floatRef& fr)  
                                            { start = min(fr.len, start);
					      len = fr.len-start; 
					      step = fr.step;
					      data = fr.data + step*start; }
    int       Size(void) const     { return len; }
    int       size(void) const     { return len; }	// GNU Vec compat
    int       Len(void)	const	   { return len; }
    int       Step(void) const     { return step; }
    float*    Data(void) const	   { return data; }

    void      SetLen(int l)	   { len = l; }
    void      SetData(float *d)	   { data = d; }
    void      SetStep(int s)	   { step = s; }

    // Overloaded operators
    // don't need operator == () fns here; they're equivalent to ones
    // in FloatVecOps.H

    // arithmetic overloading
    floatRef& operator += (const floatRef& fr)
              { Add(fr); return *this; }
    floatRef& operator += (const double d)
              { Offset(d); return *this; }

    floatRef& operator -= (const floatRef& fr)
              { Sub(fr); return *this; }
    floatRef& operator -= (const double d)
              { Offset(-d); return *this; }

    floatRef& operator *= (const floatRef& fr)
              { Mul(fr); return *this; }
    floatRef& operator *= (const double d)
              { Scale(d); return *this; }

    floatRef& operator /= (const floatRef& fr)
              { Div(fr); return *this; }
    floatRef& operator /= (const double d)
              { Scale(1.0/d); return *this; }

    floatRef& operator ^= (const floatRef& fr)
              { int i; for(i=0; i<len; ++i)  El(i) = pow(El(i),fr.Val(i));
	        return *this; }
    floatRef& operator ^= (const double d)
              { int i; for(i=0; i<len; ++i)  El(i) = pow(El(i),d);
	        return *this; }

    // Eq returns a floatVec, whereas == returns a single boolean
    floatRef& Eq(const floatRef& fr)
              { int i; for(i=0; i<len; ++i)  El(i) = (El(i)==fr.Val(i));
	        return *this; }
    floatRef& Eq(double d)
              { int i; for(i=0; i<len; ++i)  El(i) = (El(i)==d);
	        return *this; }

    // Collapsing functions (return a single value from just the vector)

    double    Dot(const floatRef& fr) const
                { double r=0;int i; assert(fr.Size()==Size());
		  for(i=0;i<len;++i) r+=Val(i)*fr.Val(i); 
		  return r; }
    double    Sum(void)	const
                { double r=0; int i; for(i=0;i<len;++i) r+=Val(i); return r; }
    double    Mean(void)	const
                { return Sum()/len; }
    double    SumSq(void) const
                { return Dot(*this); }
    double    MeanSq(void) const
                { return Dot(*this)/len; }
    double    Norm(void) const
                { return sqrt(Dot(*this)); }
    double    AvNorm(void) const
                { return sqrt(Dot(*this)/len); }

    double    Min(void)	const
		{ float r=0; int i; if(len) r=Val(0); 
		  for(i=1;i<len;++i) if(Val(i)<r) r=Val(i); return r; }
    double    Max(void)	const
		{ float r=0; int i; if(len) r=Val(0);
		  for(i=1;i<len;++i) if(Val(i)>r) r=Val(i); return r; }

    void      Scale(double f)   { int i; for(i=0;i<len;++i) El(i)*=f; }
    void      Offset(double f)  { int i; for(i=0;i<len;++i) El(i)+=f; }

    void      Log(void) { int i; for(i=0;i<len;++i) El(i) = log(El(i)); }
    void      Exp(void) { int i; for(i=0;i<len;++i) El(i) = exp(El(i)); }
    void      Sq(void)  { int i; for(i=0;i<len;++i) El(i) *= El(i); }
    void      Sqrt(void)     { int i; for(i=0;i<len;++i) El(i) = sqrt(El(i)); }
    void      Pow(double d)  { int i; for(i=0;i<len;++i) El(i) = pow(El(i),d);}
    void      dB(void)  { int i; for(i=0;i<len;++i) El(i) = TWENTYONLN10*log(El(i)); }
    void      idB(void) { int i; for(i=0;i<len;++i) El(i) = exp(LN10ONTWENTY*El(i)); }

    void      Add(const floatRef& fr) 
                  { int i; for(i=0;i<len;++i) El(i)+=fr.Val(i); }
    void      Sub(const floatRef& fr) 
                  { int i; for(i=0;i<len;++i) El(i)-=fr.Val(i); }
    void      Mul(const floatRef& fr) 
                  { int i; for(i=0;i<len;++i) El(i)*=fr.Val(i); }
    void      Div(const floatRef& fr) 
                  { int i; for(i=0;i<len;++i) El(i)/=fr.Val(i); }

    // Hypot is Add in squared domain (hypotenuese)
    void      Hypot(const floatRef& fr) 
                  { int i; for(i=0;i<len;++i) 
			      El(i) = sqrt(El(i)*El(i)+fr.Val(i)*fr.Val(i)); }

    void      HWRectify()  { int i; for(i=0;i<len;++i) if(El(i)<0) El(i)=0;}
    void      FWRectify()  { int i; for(i=0;i<len;++i) if(El(i)<0) El(i)*=-1;}
    
    void      Floor(double d) { int i; for(i=0;i<len;++i) if(El(i)<d) El(i)=d;}
    void      Ceil(double d) { int i; for(i=0;i<len;++i) if(El(i)>d) El(i)=d;}

    void      Fill(double d)  { int i; for(i=0;i<len;++i) El(i)=d; }
    void      Clear(void)    { Fill(0); }
    void      Enumerate(int base=0) {  int i; for(i=0;i<len;++i) El(i)=base+i; }

    void      Map(floatCombinerFn *cfn, double v)     
                  { int i; for(i=0;i<len;++i) El(i)=(*cfn)(El(i),v); }
    void      Map(floatCombinerFn *cfn, const floatRef& fr) 
                  { int i; for(i=0;i<len;++i) El(i)=(*cfn)(El(i),fr.Val(i)); }

    void      Print(ostream &os=cout, char *name=0) const
                  {  if(name) { os << "Ref '" << name << "'@" << (void *)this \
				    << endl; }
		     int i; for(i = 0; i<len; ++i) { os << Val(i) << " "; }
		     os << endl; 	}
};

inline ostream& operator << (ostream& os, const floatRef& fr)
    { fr.Print(os); return os; }

// Alternate invocation of collapsing methods
inline double Sum(const floatRef& fr)	 { return fr.Sum(); }
inline double Mean(const floatRef& fr)   { return fr.Mean(); }
inline double SumSq(const floatRef& fr)	 { return fr.SumSq(); }
inline double MeanSq(const floatRef& fr) { return fr.MeanSq(); }
inline double Norm(const floatRef& fr)	 { return fr.Norm(); }
inline double AvNorm(const floatRef& fr) { return fr.AvNorm(); }
inline double Max(const floatRef& fr)	 { return fr.Max(); }
inline double Min(const floatRef& fr)	 { return fr.Min(); }
inline double dot(const floatRef& a, const floatRef& b) { return a.Dot(b); }

class floatVec : public floatRef {	// floatRef with allocation
public:
    floatVec(int l=0) : floatRef(l)
              { DBG((cout << "floatVec("<<l<<")@"<<(void*)this<<endl));
		data = new float[len]; 
		/* Clear(); */ memset(data, 0, len*sizeof(float)); }
    floatVec(const floatVec& fv)
              { DBG((cout << "floatVec(fv"<<(void*)&fv<<")@"\
		     <<(void*)this<<endl));
		len = fv.Size(); data = new float[len]; 
		/* Copy(fv); */ memcpy(data, fv.data, len*sizeof(float)); }
    // a floatVec casts to a floatRef, but I think we need the above
    // to pre-empt the copy constructor.  Not sure if adding a floatRef
    // special case works, tho'
    floatVec(const floatRef& fr)
              { DBG((cout << "floatVec(fv"<<(void*)&fr<<")@"\
		     <<(void*)this<<endl));
                len = fr.Size(); data = new float[len]; floatRef::Copy(fr); }
    ~floatVec(void)
              { DBG((cout << "~floatVec()@"<<(void*)this<<endl));
		delete [] data; }
    floatVec& operator = (const floatRef& fr)    // assignment to fv clones
              { DBG((cout << "floatVec@"<<(void*)this<<"=fr"\
		     <<(void*)&fr<<endl)); 
                if((floatRef*)this != &fr) CloneVec(fr); return *this; }
    // g++ uses builtin copy assignment in preference to above - be explicit
    floatVec& operator = (const floatVec& fv)    // assignment to fv clones
              { DBG((cout << "floatVec@"<<(void*)this<<"=fv"\
		     <<(void*)&fv<<endl)); 
                if(this != &fv) CloneVec(fv); return *this; }
    void    CloneVec(const floatVec& fv)   // can copy faster if *know* a vec
              { delete [] data; len = fv.Size();
	        data = new float[len];
		memcpy(data, fv.data, len*sizeof(float)); }
    void    Copy(const floatVec& fv)   // can copy faster if *know* a vec
              { int n = min(len, fv.len);
	        memcpy(data, fv.data, n*sizeof(float)); }
    void    CloneVec(const floatRef& fr)
              { delete [] data; len = fr.Size(); 
		data = new float[len]; Copy(fr); }
    void    Resize(int l)
              { floatVec tmp(l); tmp.Fill(0); tmp.Copy(*this); _Swap(tmp); }
    void    resize(int l)		// GNU vec compatibility
              { floatVec tmp(l); tmp.Fill(0); tmp.Copy(*this); _Swap(tmp); }
    void    Resample(int l)
	      { floatVec tmp(l); int i,x; double fx;
	        for(i=0; i<l; ++i)  { 
		    fx = (i*(len-1.0))/(l-1.0); x = (int)fx; fx -= x;
		    tmp[i] = (1-fx)*El(x);
		    if(x<len-1)	tmp[i] += fx*El(x+1); }
	        _Swap(tmp); }
};

class probArray;	// fwd ref for friend, permit _Init()
class floatArray;       // fwd ref for return from Mult

class arrayRef {	// lightweight info for accessing 2D data
friend class probArray;	// needed .. later .. ?
protected:
    int rows;
    int cols;
    float *adata;	// separate pointer to the data, alloc'd elsewhere
    floatRef fr;	// fakeup ref for returns - reused
static float  dummy;	// for OOR returns

    void     _Swap(arrayRef& ar) 
	       { int t;  t = rows; rows = ar.rows; ar.rows = t;
		         t = cols; cols = ar.cols; ar.cols = t; 
	         float *tp; tp = adata; adata = ar.adata; ar.adata = tp; }
    int	     _Check(int r, int c)  const
	       { if(r<0 || r>=rows || c<0 || c>=cols)  { 
		   cerr << "arrayRef: " << r << "," << c << " bad for " \
		       << rows << "," << cols << endl; ABORT_OR_RTZ; } 
	         return 1;}    
    void     _Init(int rs, int cs, float *d=0)
               { rows = rs; cols = cs; adata = d; }

public:
    arrayRef(int rs=0, int cs=0, float *d=0)
        { rows = rs; cols = cs; adata = d; }
    ~arrayRef(void)
        { }

    int       Rows(void) const		{ return rows; }
    int       Cols(void) const		{ return cols; }
    const float* Adata(void) const      { return (const float*) adata; }

    float     Val(int r, int c) const { float v = 0; if(_Check(r,c)) 
	                                v = *(adata+r*cols+c); return v; }
    float&    _El(int r, int c)	      { return *(adata+r*cols+c);}
    float&    El(int r, int c)        { if(_Check(r,c)) return _El(r,c); 
                                        else return dummy; }
    void      Set(int r, int c, float d)	{ El(r, c) = d; }

    floatRef ColRef(int c) const
                { floatRef _fr;
		  if(_Check(0,c)) {
		   _fr.SetLen(rows); _fr.SetStep(cols); _fr.SetData(adata+c);
		  } return _fr; }
    floatRef RowRef(int r) const
                { floatRef _fr;
		  if(_Check(r,0)) {
		   _fr.SetLen(cols); _fr.SetStep(1); _fr.SetData(adata+r*cols);
		  } return _fr; }
    floatRef operator[] (int r)
                { return RowRef(r); }
    void      Copy(const arrayRef& fa);
    void      CopyAt(const arrayRef& fa, int tr, int lc);
    void      AddAt(const arrayRef& fa, int tr, int lc);

    void      Print(ostream &os=cout, char *name=0) const
                  {  if(name) { os << "aRef '" << name << "'@" <<(void *)this \
				    << endl; }
		     int i,j; for(i = 0; i<rows; ++i) { 
			 for(j=0; j<cols; ++j)  {
			     os << Val(i,j) << " "; }
			 os << endl; 	} }
    int operator == (const arrayRef& ar) const
                  { int rc = 1; if(rows == ar.Rows() && cols == ar.Cols()) {
		      int i, j; for(i=0; i<rows; ++i) for(j=0; j<cols; ++j) 
			  if(Val(i,j)!=ar.Val(i,j)) { rc = 0; break; }
		  } else rc = 0; return rc; }

    void	ScaleRows(const floatRef& scale)
                  { if(scale.Size()!=rows) 
		      cerr << "ScaleRows: size mismatch " << endl;
		  else {  int r; for(r=0; r<rows; ++r)  
	 	                 RowRef(r).Scale(scale.Val(r)); } }
    void	NormRows(const floatRef& maxes)
                  { if(maxes.Size()!=rows) 
		      cerr << "MaxesRows: size mismatch " << endl;
		  else {  int r; for(r=0; r<rows; ++r)  
	 	                 RowRef(r).Scale(1.0/maxes.Val(r)); } }
    floatVec	WeightedRowSum(const floatRef& weights) const
                  { floatVec rslt; if(weights.Size()!=rows) 
		      cerr << "MaxesRows: size mismatch " << endl;
		  else {  rslt.Resize(cols);
		          int r; for(r=0; r<rows; ++r)  
			      rslt += weights.Val(r) * RowRef(r); } 
		  return rslt; }
    floatVec	SumRows(void) const
                  { floatVec rslt(cols); 
		  int r; for(r=0; r<rows; ++r)  
			      rslt += RowRef(r);
		  return rslt; }
    floatArray	Mult(const arrayRef& a0) const;

};

class floatArray : public arrayRef, public floatVec {
friend class probArray;
protected:
    void     _Swap(floatArray& fa) 
	       { floatRef::_Swap(fa);  arrayRef::_Swap(fa); }
    void     _Init(int rs, int cs, float *d)
               { floatRef::_Init(rs*cs, d); arrayRef::_Init(rs,cs,d); }
public:
    floatArray(int rs=0, int cs=0) : floatVec(rs*cs)
                { arrayRef::_Init(rs, cs, data);
                  DBG((cout << "floatArray(" << rs << "," << cs << ")@" \
		       << (void *)this << endl)); };
    floatArray(const floatArray& fa) : floatVec(fa.rows*fa.cols)
                { arrayRef::_Init(fa.rows,fa.cols,data); floatVec::Copy(fa); }
    ~floatArray(void)
                { DBG((cout << "~floatArray()@" << (void *)this << endl)); }
    floatArray& operator = (const floatArray& fa)
                 { CloneArray(fa); return *this; }	
    floatArray& operator = (const arrayRef& fa)
                 { CloneArray(fa); return *this; }	
    void      CloneArray(const arrayRef& ar)
                 { delete [] data; rows = ar.Rows(); cols = ar.Cols();
		   _Init(rows,cols,new float[rows*cols]);
		   /* arrayRef::Copy(ar); */
		   memcpy(adata, ar.Adata(), rows*cols*sizeof(float)); }
    void      Copy(const floatArray &fa) // disamb floatVec::Copy and arrR:C
                 { arrayRef::Copy(fa); }
    int operator == (const floatArray& fa) 
                 { return arrayRef::operator == (fa); }
    floatRef operator[] (int r)		// otherwise genuinely ambiguous
                { return RowRef(r); }
    // Never mind that the floatVec and arrayRef versions of these 
    // take different numbers of args - have to disambiguate here (hmph!!)
    float     Val(int r, int c) const { return arrayRef::Val(r,c); }
    float&    _El(int r, int c)	      { return arrayRef::_El(r,c); }
    float&    El(int r, int c)        { return arrayRef::El(r,c); }
    void      Set(int r, int c, float d)     { arrayRef::Set(r,c,d); }

    void      Resize(int nr, int nc);
    void      OuterProd(const floatRef& rowMarg, const floatRef& colMarg);
    void      Transpose(void);
    void      FlipUD(void);	// make top row at bottom etc
    void      FlipLR(void);	// make first col last etc
    void      Reshape(const arrayRef &fa)	// match the shape
                { Resize(fa.Rows(), fa.Cols()); }
    void      Resample(int nr, int nc);
    int       Load(const char *filename, int dataset=0);
    int       Save(const char *filename);

    // more disambiguation/unveiling of protected
    void      Print(ostream &os=cout, char *name=0) const
                { arrayRef::Print(os, name); }

};

//class Surface : public floatArray {
//protected:
//    int       base;
//public:
//    Surface(int st = NUM_TIME_STEPS, int chs = NUM_FRQ_CHANS, int bs = 0)
//	: base(bs), floatArray(chs, st)
//	        { }
//    Surface(char *filename, int dataset=0)	// read from a DAT file
//	: base(0), floatArray(0, 0)
//                { Load(filename, dataset); }
//    ~Surface(void)
//                { }
//    Surface& operator = (const Surface& su)
//          { floatArray::operator = (su); base = su.TimeBase(); return *this; }
//
//    void      SetTimeBase(int t)	{ base = t; }
//    int       TimeBase(void) 	const	{ return base; }
//    int       TimeSteps(void)	const	{ return cols; }
//    int       FrqChans(void) 	const	{ return rows; }
//};

// Patch is now in Patch.H

#endif /* _SURFACE_H */
