#!/bin/sh
# -*- Tcl -*-   The next line restarts using drspeech_tclsh \
	exec drspeech_tclsh $0 ${1+"$@"}

#!/usr/local/bin/tclsh
#
# test-bwritefvec.tcl
#
# Script to test basic function of binary floatVec IO in libfarray_otcl.so
#
# 1998dec05 dpwe@icsi.berkeley.edu
# $Header: /u/drspeech/src/guitools/farray_otcl/RCS/test_bwritefvec.tcl,v 1.1 1998/12/06 02:41:47 dpwe Exp $

if {$argv != ""} {
    set libname $argv
} else {
    set libname "[file dirname [info script]]/H-$env(MACHTYPE)/libfarray_otcl.so"
}

if {![string match "*/*" $libname]} {
    # No path components - prepend cwd
    set libname "[pwd]/$libname"
}

# Farray requires otcl - hope it's in TCLLIBPATH
package require Otcl

puts "loading $libname ..."
load $libname Farray

# Set to "be" or "le" depending on the natural endianness of the machine
if {[lindex [array get tcl_platform "byteOrder"] 1] == "littleEndian"} {
    set nat "le"
} else {
    set nat "be"
}

proc ResultCode code {
    switch $code {
        0 {return TCL_OK}
        1 {return TCL_ERROR}
        2 {return TCL_RETURN}
        3 {return TCL_BREAK}
        4 {return TCL_CONTINUE}
    }
    return "Invalid result code $code"
}

proc OutputTestError {id command expectCode expectResult resultCode result} {
    puts stderr "======== Test $id failed ========"
    puts stderr $command
    puts stderr "==== Result was: [ResultCode $resultCode]:"
    puts stderr $result
    puts stderr "==== Expected : [ResultCode $expectCode]:"
    puts stderr $expectResult
    puts stderr "===="
}

# Test Procedure used by all tests
# id is the test identifier
# code is the test scenario
# optional -dontclean argument will stop the test classes being cleaned out
# alternatively -cleanup {script} will execute script before cleaning out

proc Test {id command expectCode expectResult args} {
    # debug
    puts "------------------------- Test $id:"

    set resultCode [catch {uplevel $command} result]

    if {($resultCode != $expectCode) ||
        ([string compare $result $expectResult] != 0)} {
        OutputTestError $id $command $expectCode $expectResult $resultCode \
                $result
    }

   if {[llength $args] == 0 || [lindex $args 0] != "-dontclean"} {
      if { [llength $args] > 0 && [lindex $args 0] == "-cleanup" } {
	  # post-completion cleanup specified
	  if { [llength $args] != 2 }  {
	      puts stderr "Test: -cleanup specified without cleanup script"
          } else {
	      set cscrpt [lindex $args 1]
 	      if {[catch {uplevel $cscrpt} rslt]} {
	          puts stderr "**Cleanup string \"$cscrpt\" returned \"$rslt\""
	      }
	  }
      }
   }
}

global results

set tmpfile "tmp.tst"

Test 1.1 {
    # Write & read floatvec
    floatVec fv 4
    fv Set 0 1
    fv Set 1 -1.5
    fv Set 2 123456789
    fv Set 3 0
    set f [open $tmpfile "w"]  
    # Return val is bytecount
    set results [bwritefvec $f fv]
    close $f
    fv Clear
    set f [open $tmpfile "r"]
    # return is bytecount again
    lappend results [breadfvec $f fv]
    close $f
    lappend results [expr 0.0+[fv Val 0]]
    lappend results [expr 0.0+[fv Val 1]]
    lappend results [expr 0.0+[fv Val 2]]
    lappend results [expr 0.0+[fv Val 3]]
    set results
} 0 {16 16 1.0 -1.5 1.23457e+08 0.0} -cleanup {otclDelete fv}

Test 1.2 {
    # Write & read floatvec as little-endian doubles
    floatVec fv 4
    fv Set 0 1
    fv Set 1 -1.5
    fv Set 2 123456789
    fv Set 3 0
    set f [open $tmpfile "w"]  
    # Return val is bytecount
    set results [bwritefvec $f fv f8l]
    close $f
    lappend results [file size $tmpfile]
    fv Clear
    set f [open $tmpfile "r"]
    # return is bytecount again
    lappend results [breadfvec $f fv f8l]
    close $f
    lappend results [expr 0.0+[fv Val 0]]
    lappend results [expr 0.0+[fv Val 1]]
    lappend results [expr 0.0+[fv Val 2]]
    lappend results [expr 0.0+[fv Val 3]]
    set results
} 0 {32 32 32 1.0 -1.5 1.23457e+08 0.0} -cleanup {otclDelete fv}

Test 1.3 {
    # Write & read floatvec as ints
    floatVec fv 4
    fv Set 0 1
    fv Set 1 -1.5
    fv Set 2 123456789
    fv Set 3 0
    set f [open $tmpfile "w"]  
    set results [bwritefvec $f fv i]
    close $f
    fv Clear
    set f [open $tmpfile "r"]
    # return is bytecount again
    lappend results [breadfvec $f fv i]
    close $f
    lappend results [expr 0.0+[fv Val 0]]
    lappend results [expr 0.0+[fv Val 1]]
    lappend results [expr 0.0+[fv Val 2]]
    lappend results [expr 0.0+[fv Val 3]]
    set results
} 0 {16 16 1.0 -1.0 1.23457e+08 0.0} -cleanup {otclDelete fv}

Test 1.4 {
    # Write & read floatvec as unsigned 2 byte swapped ints
    floatVec fv 4
    fv Set 0 1
    fv Set 1 -1.5
    fv Set 2 56789.10
    fv Set 3 65536.123
    set f [open $tmpfile "w"]  
    set results [bwritefvec $f fv u2s]
    close $f
    fv Clear
    set f [open $tmpfile "r"]
    # return is bytecount again
    lappend results [breadfvec $f fv u2s]
    close $f
    lappend results [expr 0.0+[fv Val 0]]
    lappend results [expr 0.0+[fv Val 1]]
    lappend results [expr 0.0+[fv Val 2]]
    lappend results [expr 0.0+[fv Val 3]]
    set results
} 0 {8 8 1.0 65535.0 56789.0 0.0} -cleanup {otclDelete fv}

Test 2.1 {
    # Check that writing byteswapped floats doesn't hurt the vector
    floatVec fv 4
    fv Set 0 1
    fv Set 1 -1.5
    fv Set 2 56789.10
    fv Set 3 65536.123
    set f [open $tmpfile "w"]  
    set results [bwritefvec $f fv f4s]
    close $f
    lappend results [expr 0.0+[fv Val 0]]
    lappend results [expr 0.0+[fv Val 1]]
    lappend results [expr 0.0+[fv Val 2]]
    lappend results [expr 0.0+[fv Val 3]]
    fv Clear
    set f [open $tmpfile "r"]
    # return is bytecount again
    lappend results [breadfvec $f fv f4s]
    close $f
    lappend results [expr 0.0+[fv Val 0]]
    lappend results [expr 0.0+[fv Val 1]]
    lappend results [expr 0.0+[fv Val 2]]
    lappend results [expr 0.0+[fv Val 3]]
    set results
} 0 {16 1.0 -1.5 56789.1 65536.1 16 1.0 -1.5 56789.1 65536.1} -cleanup {otclDelete fv}

Test 2.2 {
    # Check that ints are really written big/little
    floatVec fv 2
    fv Set 0 222  
    fv Set 1 [expr 1+256*(2+256*(3))]
    set f [open $tmpfile "w"]  
    bwritefvec $f fv i4b
    bwritefvec $f fv i4l
    close $f
    set results [file size $tmpfile]
    fv Clear
    # Set to read in as unsigned bytes
    fv Resize 16
    set f [open $tmpfile "r"]
    breadfvec $f fv u1
    close $f
    for {set i 0} {$i < 16} {incr i} {
	lappend results [expr int([fv Val $i])]
    }
    set results
} 0 {16 0 0 0 222 0 3 2 1 222 0 0 0 1 2 3 0} -cleanup {otclDelete fv}

Test 3.1 {
    # Check access via non-unity-step floatRefs
    set results {}
    floatArray fa 3 3
    fa Set 0 0 1.0
    fa Set 0 1 2.0
    fa Set 0 2 3.0
    fa Set 1 2 6.0
    fa Set 2 2 9.0
    set f [open $tmpfile "w"]  
    bwritefvec $f [fa RowRef 0]
    bwritefvec $f [fa ColRef 2]
    close $f
    fa Clear
    set f [open $tmpfile "r"]
    breadfvec $f [fa ColRef 0]
    lappend results [expr int([fa Val 0 0])]
    lappend results [expr int([fa Val 1 0])]
    lappend results [expr int([fa Val 2 0])]
    breadfvec $f [fa RowRef 1]
    lappend results [expr int([fa Val 1 0])]
    lappend results [expr int([fa Val 1 1])]
    lappend results [expr int([fa Val 1 2])]
    close $f
    set results
} 0 {1 2 3 3 6 9} -cleanup {otclDelete fa}
    

puts "binary int/float floatvec IO tests completed"

