/*
 *  DATUtils.c - A set of functions to provide encapsulated access to DAT files
 *  after DATmex/ files
 *  dpwe@media.mit.edu 1995may10
 * 1998feb10 dpwe@icsi.berkeley.edu
 * $Header: /n/yam/da/dpwe/src/tcl/libdat/RCS/DATUtils.c,v 1.2 1998/02/11 02:02:33 dpwe Exp $
 */

#include <stdio.h>
#include <stdlib.h>
#include <time.h>

#include "mystdarg.h"

#include <dat.h>

#include "DATUtils.h"

void Error(char *format, ...)
{
    va_list args;
    
    va_start(args, format);
    vfprintf(stderr, format, args);
    va_end(args);
    fputc('\n', stderr);
}
    
void Warn(char *format, ...)
{
    va_list args;
    
    va_start(args, format);
    vfprintf(stderr, format, args);
    va_end(args);
    fputc('\n', stderr);
}

static struct { char fmt; int count; } fmtTab[] = {
    { 'b', sizeof(unsigned char)  }, 
    { 'u', sizeof(unsigned short) },
    { 'U', sizeof(unsigned long)  },
    { 'c', sizeof(char)   },
    { 'd', sizeof(short)  }, 
    { 'D', sizeof(long)   },
    { 'f', sizeof(float)  },
    { 'F', sizeof(double) },
    { 'z', 2*sizeof(float)  }, /* re,im complex */
    { 'Z', 2*sizeof(double) }, /* re,im complex */
    { 'x', sizeof(float)  },   /* always -1..1 (but 'f' does this for me) */
    { 'X', sizeof(double)  }
};

int Fmt2ByteCount(char fmt)
{   /* return the number of bytes per data point for the format 
       indicated by the char.  Zero if unrecognized */
    int i=0, bytes = 0;

    while(bytes == 0 && i<(sizeof(fmtTab)/sizeof(struct { char c; int i; }))) {
	if(fmt == fmtTab[i].fmt) {
	    bytes = fmtTab[i].count;
	}
	++i;
    }
    return bytes;
}

static int EnsureDimensions(int dsets, int chans, int rows, int cols, 
			  int *dsetPtr, int *chanPtr, int *rowPtr, int *colPtr,
			  int *xsizePtr, int *ysizePtr)
{	/* Take quartet specifying a 2D slice through a 4D DAT file
	   and ensure that exactly two of the specified dimensions are -1
	   i.e. it's a valid 2D slice.  Return 1 if changes were necessary. 
	   Uses the input actual outer dimensions of the volume to return 
	   the size of the specified slice too, where <x> is the slower
	   rotating dimension (i.e. the rows of a raster-scanned image). */
    int changed = 0;
    int nneg = (*dsetPtr<0)+(*chanPtr<0)+(*rowPtr<0)+(*colPtr<0);
    /* ensure exactly two dimensions are -1 */    
    switch(nneg)  {
    case 0:
	*colPtr = -1; 	*rowPtr = -1;
	changed =  1;
	break;
    case 1:
	if(*colPtr<0)	*rowPtr = -1;
	else		*colPtr = -1;
	changed = 1;
	break;
    case 2:			/* as we want it */
	break;
    case 3:
	if(*dsetPtr<0)	*dsetPtr = 0;
	else		*chanPtr = 0;
	changed = 1;
	break;
    case 4:			/* i.e. all of them negative! */
	*dsetPtr = 0;	*chanPtr = 0;
	changed = 1;
	break;
    }
    /* figure size of output data */
    if(*dsetPtr<0) {
	*xsizePtr = dsets;
	if(*chanPtr<0) {
	    *ysizePtr = chans;
	} else if(*rowPtr<0) {
	    *ysizePtr = rows;
	} else {
	    *ysizePtr = cols;
	}
    } else if(*chanPtr<0) {
	*xsizePtr = chans;
	if(*rowPtr<0)  {
	    *ysizePtr = rows;
	} else {
	    *ysizePtr = cols;
	}
    } else {
	*xsizePtr = rows;
	*ysizePtr = cols;
    }
    /* meaning of x,y is that array is read in transpose raster format, 
       i.e. every row (yval) of the first column (x=0), then x=1 etc */
    return changed;
}

int DATRead(char *fileName, int dset, int chan, int row, int col, 
	    void **dataPtrPtr, char format, 
	    int *rowsPtr, int *colsPtr, double *srPtr)
{   /* Read a 2D matrix of data from the DAT file specified by <fileName>.  
       The first two of the four dimensions (<dset>, <chan>, <row>, <col>)
       that are not (-1) are held constant, and the result scans across the 
       remaining two.  If fewer than two dimensions are not -1, then they 
       are arbitrarily taken to be zero starting with <dset>.
       <format> specifies the type of data to write into <**dataPtrPtr> -- 
       'f' for float, 'F' for doubles, 'c' for signed char, 'b' for 
       unsigned byte etc (these are the libdat format chars).  
       <*srPtr> receives the sample rate of the DAT file if non-null.  
       If <*dataPtrPtr> is NULL, it is allocated.  The return code is the 
       number of *data points* written to <**dataPtrPtr>, or -1 on 
       failure. <*rowsPtr> receives the number of rows read, and 
       <*colsPtr> the number of columns.  Data is stored with columns as 
       the fastest rotating dimension, image-style. */
    long got = 0;
    DATFILE *file;
    double sampleRate;
    int rows, cols;
    long    dsets,chans,dims[2];
    int tdims, bps;
    int x,y,xsize,ysize;
    const char *key;
    int infmt;
    struct d_addr  where;
    int bytesPerPt;
    int outputBytes;
    char dtype;
    
    /* Open the datfile */
    if( (file = d_open(fileName, "r"))==NULL)  {
	Error("error opening '%s'", fileName); 
 	return -1;
    }
    /* Inspect the number of dimensions */
    tdims=d_getdim(file, &dsets, &chans, dims, 2);
    if(tdims!=2)  {
	Warn("DATRead: warning: file '%s' has %d dimensions; last 2 used.", 
	     fileName, tdims);
	if(tdims<2)  {
	    Error("malformed DAT file.");
	    d_close(file);
	    return -1;
	}
    }
    rows = dims[0];    cols = dims[1];
    /* Make sure we have exactly a 2D slice selected and figure its size */
    EnsureDimensions(dsets, chans, rows, cols, &dset, &chan, &row, &col, 
		     &xsize, &ysize);
    if(dset >= dsets || chan >= chans || row >= rows || col >= cols)  {
	Error("DATRead: file has dsets=%d, chans=%d, rows=%d, cols=%d, not %d,%d,%d,%d", 
	      dsets, chans, rows, cols, dset, chan, row, col);
	d_close(file);
	return -1;
    }
    dtype = d_gettype(file, &bps);	/* e.g. 'b' or 'f' */
    
    if( (key=d_getkey(file, SAMPLERATE))!=NULL)  {
	sampleRate = atof(key);
    } else {	
	sampleRate = 0;
    }

    /* figure size of output data */
    /* How many bytes of data are we actually going to read? */
    bytesPerPt = Fmt2ByteCount(format);
    if(bytesPerPt == 0)  {
	Error("DATRead: '%c' is not a known format type", format);
	d_close(file);
	return -1;
    }
    outputBytes = bytesPerPt * xsize * ysize;

    if(*dataPtrPtr == NULL)  {
	*dataPtrPtr = (void *)malloc(outputBytes);
	if(*dataPtrPtr==NULL)  {
	    Error("DATRead : could not alloc %d bytes", outputBytes);
	    d_close(file);
	    return -1;
	}
    }

    /* We are either reading a whole dataset, a set of columns from 
       different datasets, or individual points from each dataset */
    if(row<0 && col<0)  {	
	/* just a straight frame */
	/* seek to appropriate pos */
	where.dset = dset;
	where.chan = chan;
	where.dim[0] = 0;
	where.dim[1] = 0;
	d_seek(file, &where, 0);	/* SEEK_SET */
	got = d_readc(file, (char *)(*dataPtrPtr), xsize*ysize, format);
/* Warn("DATread got a frame of %d elements", got); */
    } else if(col < 0)  {
	where.dset = dset;
	where.chan = chan;
	where.dim[0] = row;
	where.dim[1] = 0;
	/* can at least load a whole col at a time */
	got = 0;
	for(x = 0; x<xsize; ++x)  {
	    if(dset<0)	where.dset = x;
	    else if(chan<0)	where.chan = x;
	    else		where.dim[0] = x;
	    d_seek(file, &where, 0);	/* SEEK_SET */
	    got += d_readc(file, ((char *)*dataPtrPtr)
			          +bytesPerPt*(x*ysize), ysize, format);
/* Warn("DATread got a column of %d elements", got); */
	}	
    } else {
	where.dset = dset;
	where.chan = chan;
	where.dim[0] = row;
	where.dim[1] = col;
	got = 0;
	for(x = 0; x<xsize; ++x)  {		
	    if(dset<0)	where.dset = x;
	    else 		where.chan = x;
	    for(y = 0; y<ysize;++y)  {
		if(chan<0)	where.chan = y;
		else	where.dim[0] = y;
		d_seek(file, &where, 0);	/* SEEK_SET */
		got += d_readc(file, ((char *)*dataPtrPtr)
			              +bytesPerPt*(x*ysize+y), 1,format);
/* Warn("DATread got a point of %d elements", got); */
	    }
	}
    }		
    d_close(file);

    /* Write data to the output pointers */
    if (srPtr != NULL){
	*srPtr = sampleRate;
    }
    if (rowsPtr != NULL)  {
	*rowsPtr = xsize;
    }
    if (colsPtr != NULL)  {
	*colsPtr = ysize;
    }

    return got;
}

int DATWrite(char *fileName, int dset, int chan, int row, int col, 
	     void *dataPtr, char format, int nrows, int ncols, 
	     double sampleRate, char outFormat, int trunc)
{   /* Write a 2D matrix of data out to a DAT file.  The data is read from 
       <*dataPtr>, which is a pointer to <nrows>x<ncols> of the data point 
       indicated by the format character <format>.  (columns are the 
       fastest rotating dimension, image-style).  The DAT file is given 
       a samplingRate tag if <sr> is greater than zero.  If the DAT file 
       exists, attempt to replace or append the slice indicated by 
       <dset>, <chan>, <row>, <col> where a value of (-1) indicates that 
       the dimension is to be written over, not held constant.  
       <outFormat> is the output format for the file, or ' ' to copy an 
       existing file's format.  Return samples written, -1 on error. 
       If <trunc> is nonzero, any existing file is destroyed (truncated) 
       first. */
    long did = 0;
    DATFILE *file;
    int rows, cols;
    long    dsets,chans,dims[2];
    int tdims;
    int x,y,xsize,ysize;
    const char *key;
    int infmt;
    struct d_addr  where;
    int bytesPerPt;
    int outputBytes;
    int append = 0;
    int spos;

    memset(&where, 0, sizeof(struct d_addr));

    /* Open the datfile */
    if(trunc)  {
	file = d_open(fileName, "w");
    } else {
	file = d_open(fileName, "a+");
    }
    if( file == NULL )  {
	Error("error opening '%s'", fileName);
 	return 0;
    }
    if((tdims=d_getdim(file, &dsets, &chans, dims, 2))==0 || dims[0] == 0) {
	/* tdims == 0 indicates that this file is new - set up the dims */
	/* first, check we have a valid format char */
	if((bytesPerPt = Fmt2ByteCount(outFormat))==0)  {
	    Error("Creating new '%s' but '%c' is not a valid format\n", 
		  fileName, outFormat);
	    d_close(file);
	    return -1;
	}
	dims[0] = nrows;
	dims[1] = ncols;	   /* last dimension is most rapidly varying */
	d_setdim(file, dsets=0, chans=1, dims, tdims=2);	
	/* no datasets as yet, make 1 chan */
	d_settype(file, outFormat);
	if(sampleRate > 0)  {
	    char srateStr[32];
	    sprintf(srateStr, "%.1f", sampleRate);
	    d_setkey(file, SAMPLERATE, srateStr);
	}
    } else {
	outFormat = d_gettype(file, &bytesPerPt);	/* e.g. 'b' or 'f' */
    	if( (key=d_getkey(file, SAMPLERATE))!=NULL)  {
	    sampleRate = atof(key);
	}
    }
    rows = dims[0];    cols = dims[1];
    /* Make sure we have exactly a 2D slice selected and figure its size */
    EnsureDimensions(dsets, chans, rows, cols, &dset, &chan, &row, &col, 
		     &xsize, &ysize);
    if(nrows != xsize || ncols != ysize) {
	Error("DATWrite: write: input rxc (%dx%d) but file (%dx%d)\n", 
	      nrows, ncols, xsize, ysize);
	d_close(file);
	return -1;
    }

    if(dset==dsets)  {	/* attempting to write dataset just beyond end - app */
	append = 1;
    }
    if(append)  {
	/* can only write 1st channel for appending; no way to write 
	   one chan of a multichan dataset */
	if(chan != 0 && chans != 1)  {
	    Error("DATUpdate: append: must be exactly one dataset in 1 chan file");
	    d_close(file);
	    return -1;
	}
	if(rows != nrows || cols != ncols) {
	    Error("DATUpdate: append: input rxc (%dx%d) but file (%dx%d)\n", 
		  nrows, ncols, rows, cols);
	    d_close(file);
	    return -1;
	}
	dset = dsets;
	dsets += 1;	/* will be when we're done */
	d_setdim(file, dsets, chans, dims, 2);	/* 1 dataset, 1 channel */
	{
	    time_t tm;
	    char timestr[256];
	    time(&tm);
	    sprintf(timestr, 
		    "Written by DATUtils::DATAppend, %s", ctime(&tm));
	    /* clobber trailing CR */
	    timestr[strlen(timestr)-1] = '\0';
	    d_setkey(file, COMMENT, timestr);
	}
    }
    if(dset >= dsets || chan >= chans || row >= rows || col >= cols)  {
	Error("DATUpdate: file has dsets=%d, chans=%d, rows=%d, cols=%d, not %d,%d,%d,%d", 
	      dsets, chans, rows, cols, dset, chan, row, col);
	d_close(file);
	return -1;
    }

    if(Fmt2ByteCount(format) == 0)  {
	Error("DATUpdate: '%c' is not a known input format type", format);
	d_close(file);
	return -1;
    }
    /* We are either writing a whole dataset, a set of columns to
       different datasets, or individual points to each dataset */
    if(row<0 && col<0)  {
	/* just a straight frame */
	/* seek to appropriate pos */
	where.dset = dset;
	where.chan = chan;
	where.dim[0] = 0;
	where.dim[1] = 0;
	spos = d_seek(file, &where, 0);	/* SEEK_SET */
	did = d_writec(file, (char *)(dataPtr), xsize*ysize, format);
    } else if(col < 0)  {
	where.dset = dset;
	where.chan = chan;
	where.dim[0] = row;
	where.dim[1] = 0;
	/* can at least do a whole col at a time */
	did = 0;
	for(x = 0; x<xsize; ++x)  {
	    if(dset<0)	where.dset = x;
	    else if(chan<0)	where.chan = x;
	    else		where.dim[0] = x;
	    spos = d_seek(file, &where, 0);	/* SEEK_SET */
	    did += d_writec(file, ((char *)dataPtr)
			          +bytesPerPt*(x*ysize), ysize, format);
	}	
    } else {
	where.dset = dset;
	where.chan = chan;
	where.dim[0] = row;
	where.dim[1] = col;
	did = 0;
	for(x = 0; x<xsize; ++x)  {		
	    if(dset<0)	where.dset = x;
	    else 		where.chan = x;
	    for(y = 0; y<ysize;++y)  {
		if(chan<0)	where.chan = y;
		else	where.dim[0] = y;
		spos = d_seek(file, &where, 0);	/* SEEK_SET */
		did += d_writec(file, ((char *)dataPtr)
			              +bytesPerPt*(x*ysize+y), 1,format);
	    }
	}
    }		
    d_close(file);

    return did;
}

int DATInfo(char *fileName, int *dsetsPtr, int *chansPtr, int *rowsPtr, 
	    int *colsPtr, double *srPtr, char *fmtPtr)
{   /* Obtain information for the DATfile specified by <filename>.
       Return the number of dataset, channels, rows and columns 
       in the appropriate pointers.  Return the sampling rate (if any) 
       and format character too.  Returns 0 if file could not be opened, 
       else 1. */
    DATFILE *file;
    double sampleRate;
    int rows, cols;
    long    dsets,chans,dims[2];
    int tdims;
    const char *key;
    char dtype;
    int  bps;

    /* Open the datfile */
    if( (file = d_open(fileName, "r"))==NULL)  {
 	return 0;
    }
    if((tdims=d_getdim(file, &dsets, &chans, dims, 2))!=2)  {
	Warn("DATInfo: warning: file '%s' has %d dimensions; last 2 used", 
	     fileName, tdims);
    }
    rows = dims[0];    cols = dims[1]; /* cols are fastest dimension */
    dtype = d_gettype(file, &bps);	/* e.g. 'b' or 'f' */
    
    key=d_getkey(file, SAMPLERATE);
    if( key!=NULL)  {
	sampleRate = atof(key);
    } else {
	sampleRate = 0.0;
    }
    d_close(file);

    if (dsetsPtr != NULL) *dsetsPtr = dsets;
    if (chansPtr != NULL) *chansPtr = chans;
    if (rowsPtr  != NULL) *rowsPtr  = rows;
    if (colsPtr  != NULL) *colsPtr  = cols;
    if (srPtr    != NULL) *srPtr    = sampleRate;
    if (fmtPtr   != NULL) *fmtPtr   = dtype;
    return 1;
}

/* Additional util fns to access extra tags in descriptor.
   Added 1996jan25 dpwe@media.mit.edu */

int DATTagRead(char *fileName, char *tagname, char *tagval, int tagvallen)
{   /* Read a descriptor field for the DATfile specified by <filename>.
       Opens the named DAT file, attempts to retrieve a field specified 
       by *tagname, and returns as much of the associated data in 
       tagval, which is assumed preallocated to tagvallen.  *tagval 
       is null-terminated. 
       Return 1 if *fileName successfully opened as a DATfile (even 
       if the tagname not found) else 0.
     */
    DATFILE *file;
    const char *key;

    /* Open the datfile */
    if( (file = d_open(fileName, "r"))==NULL)  {
 	return 0;
    }
    key=d_getkey(file, tagname);
    if( key!=NULL)  {
	char c='\1';
	int n = tagvallen;
	while(n>0 && c!='\0') {
	    c = *key++;
	    *tagval++ = c;
	    --n;
	}
    } else {
	if(tagvallen>0)	*tagval = '\0';
    }
    d_close(file);

    return 1;
}

int DATTagWrite(char *fileName, char *tagname, char *tagval)
{   /* Write a descriptor field for the DATfile specified by <filename>.
       Opens the named DAT file (which must exist), write an entry 
       in the descriptor for *tagname as the null-terminated string 
       *tagval.
       Return 1 if *fileName successfully opened as a DATfile else 0.
     */
    DATFILE *file;
    const char *key;

    /* Open the datfile */
    if( (file = d_open(fileName, "a+"))==NULL)  {
 	return 0;
    }
    d_setkey(file, tagname, tagval);
    d_close(file);

    return 1;
}

#ifdef MAIN
/* test code */

static void PrintShortArray(char *tag, short *d, int rs, int cs)
{   /* print a 2d array of shorts with <rs> rows and <cs> cols.  
       Cols is fastest rotating dimension. */
    int i,j;
    fprintf(stderr , "'%s':\n", tag);
    for(i=0; i<rs; ++i)  {
	for(j=0; j<cs; ++j)  {
	    fprintf(stderr, "%d ", *d++);
	}
	fprintf(stderr, "\n");
    }
}

static void testDATinfo(char *filename)
{
    int ds, ch, rw, cl;
    double sr;
    char fm;

    if(DATInfo(filename, &ds, &ch, &rw, &cl, &sr, &fm)==0)  {
	fprintf(stderr, "testDATinfo: couldn't open '%s'\n", filename);
    } else {
	fprintf(stderr, "DATinfo for %s: ds %d, ch %d, rw %d, co %d\n\
sr %f fmt %c\n", filename, ds, ch, rw, cl, sr, fm);
    }
}

static void testDATwrite(char *filename)
{
    int i,j,rc;
    int sx = 5, sy = 10;
    short dd[5][10];
    fprintf(stderr, "Write a single frame 5 row, 10 col dat file\n");
    for(i=0; i<5; ++i)  {
	for(j=0; j<10; ++j)  {
	    dd[i][j] = 1000*((i+j)%10);
	}
    }
    rc = DATWrite(filename, -1, -1, -1, -1, dd, 'd', 5, 10, 44100, 'd', 1);
    fprintf(stderr, "Write returned %d\n", rc);
}

static void testDATread(char *filename)
{
    int i,j,rc;
    int sx, sy;
    short *dd=NULL;
    double sr=-1;

    fprintf(stderr, "Read a single frame from a DAT file\n");
    rc = DATRead(filename, -1, -1, -1, -1, (void **)&dd, 'd', &sx, &sy, &sr);
    fprintf(stderr, "DATRead returned %d (sr = %f)\n", rc, sr);
    if(rc > 0)  {
	PrintShortArray("", dd, sx, sy);
    }
}

static void testDATwriteMulti(char *filename)
{
    int i,j,k,rc;
    int sx = 5, sy = 10;
    short dd[5][10];
    fprintf(stderr, "Write 3 datasets of 5 row, 10 col dat file\n");
    for(k = 0; k<3; ++k) {
	fprintf(stderr, "%d..\n", k);
	for(i=0; i<5; ++i)  {
	    for(j=0; j<10; ++j)  {
		dd[i][j] = 10*(k+1)+((i+j)%10);
	    }
	}
	rc = DATWrite(filename, k, 0, -1, -1, dd, 'd', 5, 10, 44100, 'd', 0);
    }
    fprintf(stderr, "Write returned %d\n", rc);
}

static void testDATreadMulti(char *filename)
{
    int i,j,k=0,rc=1;
    int sx, sy;
    short *dd=NULL;
    double sr=-1;

    fprintf(stderr, "Read every frame from a DAT file\n");
    while(rc>0)  {
	rc = DATRead(filename, k, -1, -1, -1, (void **)&dd, 'd', &sx, &sy, &sr);
	fprintf(stderr, "frm %d: DATRead returned %d (sr = %f)\n", k, rc, sr);
	if(rc > 0)  {
	    PrintShortArray("", dd, sx, sy);
	}
	++k;
    }
}

static void testDATwriteSlice(char *filename)
{
    int ds, ch, rw, cl;
    double sr;
    char fm;
    short *dd;
    int i,rc;

    fprintf(stderr, "Testing writing a slice of data\n");
    if(DATInfo(filename, &ds, &ch, &rw, &cl, &sr, &fm)!=0)  {
	fprintf(stderr, "DATinfo for %s: ds %d, ch %d, rw %d, co %d\n\
sr %f fmt %c\n", filename, ds, ch, rw, cl, sr, fm);
	/* write data sets by cols */
	dd = (short *)malloc(ds*cl*sizeof(short));
	for(i = 0; i<ds*cl; ++i)  { dd[i] = i; }
	rc = DATWrite(filename, -1, 0, 1, -1, dd, 'd', ds, cl, 44100, 'd', 0);
	fprintf(stderr, "DATWrite returned %d\n", rc);
	free(dd);

	/* write data sets by rows */
	dd = (short *)malloc(ds*cl*sizeof(short));
	for(i = 0; i<ds*rw; ++i)  { dd[i] = i; }
	rc = DATWrite(filename, -1, 0, -1, 2, dd, 'd', ds, rw, 44100, 'd', 0);
	fprintf(stderr, "DATWrite returned %d\n", rc);
	free(dd);
    }
}

static void testDATreadSlice(char *filename)
{
    int ds, ch, rw, cl;
    double sr;
    char fm;
    short *dd=NULL;
    int i,rc,sx,sy;

    fprintf(stderr, "Testing reading a slice of data\n");
    if(DATInfo(filename, &ds, &ch, &rw, &cl, &sr, &fm)!=0)  {
	fprintf(stderr, "DATinfo for %s: ds %d, ch %d, rw %d, co %d\n\
sr %f fmt %c\n", filename, ds, ch, rw, cl, sr, fm);
	/* read data sets by cols */
	rc = DATRead(filename, -1, 0, 1, -1, (void **)&dd, 'd', &sx, &sy, &sr);
	fprintf(stderr, "DATRead returned %d\n", rc);
	PrintShortArray("", dd, sx, sy);
	free(dd); dd = NULL;

	/* read data sets by rows */
	rc = DATRead(filename, -1, 0, -1, 2, (void **)&dd, 'd', &sx, &sy, &sr);
	fprintf(stderr, "DATRead returned %d\n", rc);
	PrintShortArray("", dd, sx, sy);
	free(dd); dd = NULL;
    }
}

int main(int argc, char **argv)
{
    int test;
    char *filename;
    if(argc != 3)  {
	fprintf(stderr, "usage: %s nn filename\n\
where:\n\
  nn  specifies the test (0 = write, 1 = read,\n\
                          2 = write multi, 3 = read multi, \n\
                          4 = write slice, 5 = read slice, \n\
                          6 = info)\n\
  filename  is name of datfile to mess with\n", argv[0]);
	exit(1);
    }
    filename = argv[2];
    test = atoi(argv[1]);
    switch(test)  {
    case 0:  testDATwrite(filename); break;
    case 1:  testDATread(filename); break;
    case 2:  testDATwriteMulti(filename); break;
    case 3:  testDATreadMulti(filename); break;
    case 4:  testDATwriteSlice(filename); break;
    case 5:  testDATreadSlice(filename); break;
    case 6:  testDATinfo(filename); break;
    default: fprintf(stderr, "Unknown test %d\n", test);
    }
    /* exit code */
    return 0;
}

#endif /* MAIN */
