/*
 *  pfread.C - A Matlab C Extension (MEX) to read ICSI "pfile" data files
 *  after DATRead.c
 *  Matlab 5.0 version
 *  1997jun11 dpwe@icsi.berkeley.edu
 *  $Header: /u/drspeech/src/SPRACHcore/guitools/pfif_otcl/RCS/pfread.C,v 1.2 1998/11/17 01:17:02 dpwe Exp $
 */

#include    <stdio.h>
#include    <math.h>
#include    <ctype.h>
#include    "mex.h"

#include    "pfif.H"

#define MAX_FNAME 300
#define MAX_TYPE  32

/* tokens for type */
#define TYPE_DATA   0
#define TYPE_LABELS 1

extern "C"
void
mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[])
{
    char    filename[MAX_FNAME];
    char    typename[MAX_TYPE];
    int     i, j;
    /* Arguments with default vals */
    int	    utnum = 0;
    int     start = 0;
    int     len   = -1;
    int     type  = TYPE_DATA;
    
    if (nrhs < 1){
	mexPrintf("pfread    D = pfread(FILE [, UTNUM [, START [, LEN [, TYPE]]]])  Read a pfile\n");
	mexPrintf("         specified by filename FILE, and returns a two-dimensional array D.\n");
	mexPrintf("         UTNUM specifies which utterance to read (default 0)\n");
	mexPrintf("         START is the starting frame within the utterance (default 0)\n");
	mexPrintf("         LEN is number of frames to read (default -1 -> read to end of utt)\n");
	mexPrintf("         TYPE is 'd' for data or 'l' for labels (default 'd')\n");
	mexPrintf("1997jun11 dpwe@icsi.berkeley.edu\n");
	return;
    }
    if (!mxIsChar(prhs[0])){	/* FILENAME */
	mexErrMsgTxt("pfread: first argument (FILENAME) must be a string");
	return;
    } else {
	/* Convert to String */
	mxGetString(prhs[0], filename, MAX_FNAME);
    }
    if(nrhs>1)  {		/* UTNUM */
	utnum = mxGetScalar(prhs[1]);
    }
    if(nrhs>2)  {		/* START */
	start = mxGetScalar(prhs[2]);
    }
    if(nrhs>3)  {		/* LEN */
	len   = mxGetScalar(prhs[3]);
    }
    if(nrhs>4)  {		/* TYPE */
	if (!mxIsChar(prhs[4])) {
	    mexErrMsgTxt("pfread: fourth argument (TYPE) must be a char");
	    return;
	}
	mxGetString(prhs[4], typename, MAX_TYPE);
	if (typename[0] == 'd') {
	    type = TYPE_DATA;
	} else if (typename[0] == 'l') {
	    type = TYPE_LABELS;
	} else {
	    mexErrMsgTxt("pfread: TYPE must be 'l' or 'd'");
	    return;
	}
    }
    /* arguments all set up */
    /* open the file and find out what it has */
    pfileif pfif;
    if (!pfif.open(filename, "r")) {
    	mexErrMsgTxt("pfread: unable to open pfile");
    	return;
    }
    int totalutts = pfif.getNumUtts();
    int nfeats    = pfif.getNumFeatures();
    if(utnum >= totalutts) {
    	mexErrMsgTxt("pfread: requested UTNUM not in file");
    	return;
    }
    /* get size of requested utt */
    int uttframes = pfif.numFramesForUtt(utnum);
    if(start > uttframes) {
    	mexErrMsgTxt("pfread: requested START exceeds utt length");
    	return;
    }
    if(len < 0) {
    	len = uttframes - start;
    } else if (start + len > uttframes) {
    	mexErrMsgTxt("pfread: request LEN exceeds utt length");
    	return;
    }
    /* OK, allocate the output array */
    mxArray  *dataMatrix;
    double *pd;
    int rows = nfeats;
    int cols = len;
    mxClassID mClass = mxDOUBLE_CLASS;
    if(type == TYPE_LABELS) {
	rows = pfif.getNumLabels();
	mClass = mxINT32_CLASS;
	assert(sizeof(int)==4);
    }
    /* dataMatrix = mxCreateDoubleMatrix(rows, cols, mxREAL); */
    int dims[2];
    dims[0] = rows;  dims[1] = cols;
    dataMatrix = mxCreateNumericArray(2, dims, mClass, mxREAL);
    plhs[0] = dataMatrix;

    /* Read the data from the pfile */
    pfif.seek(utnum, start);
    if(type == TYPE_LABELS) {
	int *pl   = (int *)mxGetPr(dataMatrix);
	pfif.readLabels(rows*cols, pl);
    } else {	/* type == TYPE_DATA */
	floatArray fa = pfif.readFrames(cols);
	double *pd = mxGetPr(dataMatrix);
	float *pf = &(fa.El(0,0));    
	/* copy to the matlab array */
	for (i = 0; i < cols; ++i) {
	    for (j = 0; j < rows; ++j) {
		*pd++ = *pf++;
	    }
	}
    }
    
    /* all done */
    /* pfif will be closed and fa deleted automatically */
}
