// $Header: /n/lasagna/db/davidj/spert/src/quicknet/H-spert0/../RCS/QN_SentDB_Narrow.cc,v 1.3 1996/03/01 03:37:40 davidj Exp $

/* Must include the config.h file first */
#include <QN_config.h>
#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include "intvec.h"
#include "fltvec.h"
#include "QN_types.h"
#include "QN_Logger.h"
#include "QN_SentDB_Narrow.h"

#ifndef EXIT_SUCCESS
#define EXIT_SUCCESS (0)
#define EXIT_FAILURE (1)
#endif

// The name of this class

const char* QN_SentDB_Narrow::CLASSNAME = "QN_SentDB_Narrow";

// Some useful inline routines

static inline unsigned long
min(unsigned long a, unsigned long b)
{
    if (a<=b)
	return a;
    else
	return b;
}

// Constructor.

QN_SentDB_Narrow::QN_SentDB_Narrow(int dbg, QN_SentDB& a_db,
				   size_t a_first_ftr,
				   size_t a_num_ftrs,
				   size_t a_first_label,
				   size_t a_num_labels,
				   size_t a_first_tgt,
				   size_t a_num_tgts,
				   size_t a_buf_size)
  : debug(dbg),
    db (a_db),
    first_ftr(a_first_ftr),
    n_ftrs(a_num_ftrs),
    first_lbl(a_first_label),
    n_lbls(a_num_labels),
    first_tgt(a_first_tgt),
    n_tgts(a_num_tgts),
    n_db_ftrs(a_db.num_features_in_frame()),
    n_db_lbls(a_db.num_labels_in_frame()),
    n_db_tgts(a_db.num_targets_in_frame()),
    buf_size(a_buf_size),
    n_sents(a_db.num_sentences()),
    ftr_buf(NULL),
    lbl_buf(NULL)
{

    // Allow user to select all features, labels etc.
    if (n_ftrs==ALL_FEATURES)
	n_ftrs = n_db_ftrs - first_ftr;
    if (n_lbls==ALL_LABELS)
	n_lbls = n_db_lbls - first_lbl;
    if (n_tgts==ALL_LABELS)
	n_tgts = n_db_tgts - first_tgt;

    // TODO - should handle targets properly
    assert(n_tgts==db.num_targets_in_frame() || n_tgts==0);
    assert(first_tgt==0);

    // Check for some error conditions
    if (first_ftr > n_db_ftrs
	|| (first_ftr + n_ftrs) > n_db_ftrs)
    {
	QN_ERROR(CLASSNAME, "features %i to %i not in DB.", (int) first_ftr,
		(int) (first_ftr + n_ftrs - 1));
    }
    if (first_lbl > n_db_lbls
	|| (first_lbl + n_lbls) > n_db_lbls)
    {
	QN_ERROR(CLASSNAME, "features %i to %i not in DB.",(int) first_lbl,
		 (int) (first_lbl + n_lbls - 1));
    }

    if (debug>=1)
    {
	QN_LOG(CLASSNAME, "Using features %i to %i.", (int) first_ftr,
	       (int) (first_ftr + n_ftrs - 1));
	QN_LOG(CLASSNAME, "Using labels %i to %i.", (int) first_lbl,
	      (int) (first_lbl + n_lbls - 1));
	QN_LOG(CLASSNAME, "Using targets %u to %u.",(unsigned) first_tgt,
	       (unsigned) (first_tgt + n_tgts - 1));
    }

    narrow_ftrs = (db.num_features_in_frame()==n_ftrs) ? 0 : 1;
    narrow_lbls = (db.num_labels_in_frame()==n_lbls) ? 0 : 1;
    size_t buf_sents = min(buf_size, n_sents);
    buf_frames = db.num_frames_in_sents(0, buf_sents);

    if (narrow_ftrs)
    {
	ftr_buf = malloc_vf(buf_frames * n_db_ftrs);
	if (debug>=3)
	{
	    QN_LOG(CLASSNAME, "Allocating ftr buf for %lu sentences "
		   "(%lu frames).", (unsigned long) buf_size,
		   (unsigned long) buf_frames);
	}
    }
    if (narrow_lbls)
    {
	lbl_buf = (QNUInt32 *) malloc_vi32(buf_frames * n_db_lbls);
	if (debug>=3)
	{
	    QN_LOG(CLASSNAME, "Allocating ftr buf for %lu labels.",
		   (unsigned long) buf_frames);
	}
    }
}


QN_SentDB_Narrow::~QN_SentDB_Narrow()
{
    if (narrow_ftrs)
	free_vf(ftr_buf);
    if (narrow_lbls)
	free_vf(lbl_buf);
}

size_t
QN_SentDB_Narrow::get_sentences(size_t sentno, size_t num_sents,
				  QNUInt32* sentnos, QNUInt32* framenos,
				  float* ftrs, QNUInt32* lbls, float* tgts)
{
    size_t cur_buf_sentno;	// The sentence number of first sentence in
				// ..buffer
    size_t total_frames = 0;	// Total number of frames read
    size_t limit_sent = sentno+num_sents; // One after last sentence wanted

    assert(sentno < n_sents);
    assert(sentno+num_sents-1 < n_sents);

    // Loop over all the buffer fulls

    size_t cur_buf_sents;      // Number of sentences in current buffer full
    size_t cur_buf_frames;     // Number of frames in current buffer full
    for (cur_buf_sentno = sentno;
	 cur_buf_sentno < sentno + num_sents;
	 cur_buf_sentno += buf_size)
    {
	// Work out how many sentences this time
	cur_buf_sents = min(limit_sent - cur_buf_sentno, buf_size);
	// Work out how much space we need
	cur_buf_frames = db.num_frames_in_sents(cur_buf_sentno,cur_buf_sents);
	// Keep track of total number of frames
	total_frames += cur_buf_frames;

	// Enlarge buffers if necessary
	if (cur_buf_frames > buf_frames)
	{
	    if (narrow_ftrs)
	    {
		ftr_buf = realloc_vf(ftr_buf, n_db_ftrs * cur_buf_frames);
		if (debug>=3)
		{
		    QN_LOG(CLASSNAME, "Growing ftr buf to %lu frames.",
			   (unsigned long) cur_buf_frames);
		}
	    }
	    if (narrow_lbls)
	    {
		lbl_buf = (QNUInt32*) realloc_vi32(lbl_buf,
						   n_db_lbls * cur_buf_frames);
		if (debug>=3)
		{
		    QN_LOG(CLASSNAME, "Growing lbl buf to %lu frames.",
			   (unsigned long) cur_buf_frames);
		}
	    }
	    buf_frames = cur_buf_frames;
	}

	float* ftr_ptr;		// Where we put features
	QNUInt32* lbl_ptr;	// Where we put labels

	// Work out where we want to put stuff
	if (ftrs!=NULL && narrow_ftrs)
	    ftr_ptr = ftr_buf;
	else
	    ftr_ptr = ftrs;
	if (lbls!=NULL && narrow_lbls)
	    lbl_ptr = lbl_buf;
	else
	    lbl_ptr = lbls;

	// Get sentence data
	db.get_sentences(cur_buf_sentno, cur_buf_sents, sentnos, framenos,
			 ftr_ptr, lbl_ptr, tgts);

	// Copy features from buffer to destination if necessary
	if (narrow_ftrs && ftrs!=NULL)
	{
	    float* from_ptr = &ftr_buf[first_ftr];
	    float* to_ptr = ftrs;
	    for (size_t i=0; i<cur_buf_frames; i++)
	    {
		copy_vf_vf(n_ftrs, from_ptr, to_ptr);
		from_ptr += n_db_ftrs;
		to_ptr += n_ftrs;
	    }
	}
	if (narrow_lbls && lbls!=NULL)
	{
	    QNUInt32* from_ptr = &lbl_buf[first_lbl];
	    QNUInt32* to_ptr = lbls;
	    for (size_t i=0; i<cur_buf_frames; i++)
	    {
		copy_vi32_vi32(n_lbls, (intv_int32_t*) from_ptr,
			       (intv_int32_t*) to_ptr);
		from_ptr += n_db_lbls;
		to_ptr += n_lbls;
	    }
	}

	if (sentnos!=NULL)
	    sentnos += cur_buf_frames;
	if (framenos!=NULL)
	    framenos += cur_buf_frames;
	if (lbls!=NULL)
	    lbls += cur_buf_frames * n_lbls;
	if (ftrs!=NULL)
	    ftrs += cur_buf_frames * n_ftrs;
	if (tgts!=NULL)
	    tgts += cur_buf_frames * n_tgts;
	
    }

    if (debug>=3)
    {
	QN_LOG(CLASSNAME, "Read sentences %lu to %lu, %lu frames.",
	       (unsigned long) sentno, (unsigned long) sentno+num_sents-1,
	       (unsigned long) total_frames);
    }
    return total_frames;
}

size_t
QN_SentDB_Narrow::num_features_in_frame()
{
    return n_ftrs;
}

size_t
QN_SentDB_Narrow::num_labels_in_frame()
{
    return n_lbls;
}

size_t
QN_SentDB_Narrow::num_targets_in_frame()
{
    return n_tgts;
}

size_t
QN_SentDB_Narrow::num_frames_in_sents(size_t sentno, size_t num_sents)
{
    return db.num_frames_in_sents(sentno, num_sents);
}

size_t
QN_SentDB_Narrow::num_frames()
{
    return db.num_frames();
}

// Return the total number of sentences in the database

size_t
QN_SentDB_Narrow::num_sentences()
{
    return n_sents;
}
