/*
 *	Dynamic programming class for aligning strings
 */

#ifndef __DPROG_H__
#define __DPROG_H__

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "defines.h"
#include "sentence.h"

#define SUB_PENALTY	4	// substitution penalty
#define INS_PENALTY	3	// insertion penalty
#define DEL_PENALTY	3	// deletion penalty

enum { HORIZONTAL, VERTICAL, DIAGONAL, START } ;

class dProg {

private:

  // statistics stuff
  int totWords ;		// total number of words aligned
  int totIns ;			// total number of insertion errors
  int totDel ;			// total number of deletion errors
  int totSub ;			// total number of substitution errors
  int numWords ;		// number of words aligned in sentence
  int numSub ;			// number of subsitutions in sentence
  int numDel ;			// number of deletions in sentence
  int numIns ;			// number of insertions in sentence
  int numSents ;		// number of sentences
  int sentCorr ;		// number of correct sentences


  // back trace scores
  short **dist;
  short **dir;

  // back trace word data
  int *backTrace_row ;
  int *backTrace_col ;
  int backCnt ;

  int max_num_words;

public:

  // constructor --------------------------------------------------------------
  dProg() {

    // initialize counts
    totWords = 0 ;
    totIns = 0 ;
    totDel = 0 ;
    totSub = 0 ;
    numSents = 0 ;
    sentCorr = 0 ;
    max_num_words = MAX_NUM_WORDS;

    dist = new short *[max_num_words];
    dir = new short *[max_num_words];
    backTrace_row = new int[max_num_words*2];
    backTrace_col = new int[max_num_words*2];
    for( int row = 0 ; row < max_num_words ; row++ ) {
      dist[row] = new short[max_num_words];
      dir[row] = new short[max_num_words];
    }
  }

  ~dProg() {
    for( int row = 0 ; row < max_num_words ; row++ ) {
      delete[] dist[row];
      delete[] dir[row];
    }
    delete[] dist;
    delete[] dir;
    delete[] backTrace_row;
    delete[] backTrace_col;
  }

  // run the alignmemt --------------------------------------------------------
  void align( sentence *recs, sentence *refs ) {

    // initialize parameters
    numWords = refs->getNum() ;
    numSub = 0 ;
    numDel = 0 ;
    numIns = 0 ;

    /*
     *		Initialize the local distance 
     */

    while (recs->getNum() >= max_num_words || 
	   refs->getNum() >= max_num_words) {
      for( int row = 0 ; row < max_num_words ; row++ ) {
	delete[] dist[row];
	delete[] dir[row];
      }
      delete[] dist;
      delete[] dir;
      delete[] backTrace_row;
      delete[] backTrace_col;
      
      max_num_words *= 2;

      dist = new short *[max_num_words];
      dir = new short *[max_num_words];
      backTrace_row = new int[max_num_words*2];
      backTrace_col = new int[max_num_words*2];
      for( int row = 0 ; row < max_num_words ; row++ ) {
	dist[row] = new short[max_num_words];
	dir[row] = new short[max_num_words];
      }
    }

    // initialize the starting distance
    for( int row = 0 ; row <= recs->getNum() ; row++ ) {
      for( int col = 0 ; col <= refs->getNum() ; col++ )
	if( row && col &&
	    strcmp( recs->getWord( row - 1 ), refs->getWord( col - 1 ) ) )
	  dist[row][col] = SUB_PENALTY ;
	else
	  dist[row][col] = 0 ;
    }

    /*
     *		Find the best path through the words
     */

    // step through all the possible alignments
    for( int row = 0 ; row <= recs->getNum() ; row++ ) {
      for( int col = 0 ; col <= refs->getNum() ; col++ ) {
	
	// first row is only through deletions
	if( row == 0 ) {
	  if( col == 0 ) {
	    dir[row][col] = START ;
	  }
	  else {
	    dist[row][col] += dist[row][col-1] + DEL_PENALTY ;
	    dir[row][col] = HORIZONTAL ;
	  }
	}
      
	// first column is only through insertions -- note don't penalize 
	// for insertion with a substitution error
	else if( col == 0 ) {
	  dist[row][col] = dist[row-1][col] + INS_PENALTY ;
	  dir[row][col] = VERTICAL ;
	}
     
	// else, search over different possibilities
	else {
	
	  // initialize for no insertion/deletions
	  short tmpDist  = dist[row-1][col-1] + dist[row][col] ;
	  dir[row][col] = DIAGONAL ;
	
	  // check for deletions
	  if( dist[row][col-1] + DEL_PENALTY < tmpDist ) {
	    tmpDist = dist[row][col-1] + DEL_PENALTY ;
	    dir[row][col] = HORIZONTAL ;
	  }
       
	  // check for insertions -- note don't penalize for insertion and
	  // substitutions
	  if( dist[row-1][col] + INS_PENALTY < tmpDist ) {
	    tmpDist = dist[row-1][col] + INS_PENALTY ;
	    dir[row][col] = VERTICAL ;
	  }
       
	  // update the distance
	  dist[row][col] = tmpDist ;
	}
      }
    }

    // initialize backtrace
    int row = recs->getNum() ;
    int col = refs->getNum() ;
    short tmpDir = dir[row][col] ;
    backCnt = 0 ;

    // do the backtrace
    while( tmpDir != START ) {

      // save the alignment
      backTrace_row[backCnt] = row ;
      backTrace_col[backCnt] = col ;
      backCnt++ ;
      
      // process local backtrace
      switch( tmpDir ) {

      // no insertion or deletion (possible substitution)
      case DIAGONAL:
	if( dist[row][col] != dist[row-1][col-1] ) numSub++ ;
	row-- ;
	col-- ;
	break ;

      // deletion
      case HORIZONTAL:
	numDel++ ;
	col-- ;
	break ;

      // insertion 
      case VERTICAL:
	numIns++ ;
	row-- ;
	break ;
      }

      // check for errors 
      if( row < 0 || col < 0 ) {
	fprintf( stderr, "dProg::align : error in backtrace\n" ) ;
	exit( 1 ) ;
      }

      // update the backpointer
      tmpDir = dir[row][col] ;
    }

    // save the last backpointer
    backTrace_row[backCnt] = 0 ;
    backTrace_col[backCnt] = 0 ;

    // update cummulative terms
    updateStats() ;
  }
    
  // update the cummulative statistics ----------------------------------------
  void updateStats() {

    // word rates
    totWords += numWords ;	
    totSub += numSub ;
    totDel += numDel ;
    totIns += numIns ;

    // sentence rates
    numSents++ ;
    if( numSub + numDel + numIns == 0 ) sentCorr++ ;
  }
  
  // print sentence statistics ------------------------------------------------
  void printSentStats( FILE *fpOut ) {
    fprintf( fpOut, "\twords        : %d\n", numWords ) ;
    fprintf( fpOut, "\tsubstitutions: %d\t%4.1f%%\n", numSub, 
	     numSub * 100. / numWords ) ;
    fprintf( fpOut, "\tdeletions    : %d\t%4.1f%%\n", numDel, 
	     numDel * 100. / numWords ) ;
    fprintf( fpOut, "\tinsertions   : %d\t%4.1f%%\n", numIns, 
	     numIns * 100. / numWords ) ;
    fprintf( fpOut, "\terror        : %d\t%4.1f%%\n", numSub + numDel + numIns,
	     ( numSub + numDel + numIns ) * 100. / numWords ) ;

  }

  // print full statatistics --------------------------------------------------
  void printStats( FILE *fpOut ) {
    fprintf( fpOut, "\twords            : %d\n", totWords ) ;
    fprintf( fpOut, "\tsubstitutions    : %d\t%4.1f%%\n", totSub, 
	     totSub * 100. / totWords ) ;
    fprintf( fpOut, "\tdeletions        : %d\t%4.1f%%\n", totDel, 
	     totDel * 100. / totWords ) ;
    fprintf( fpOut, "\tinsertions       : %d\t%4.1f%%\n", totIns, 
	     totIns * 100. / totWords ) ;
    fprintf( fpOut, "\terror            : %d\t%4.1f%%\n", 
	     totSub + totDel + totIns, 
	     ( totSub + totDel + totIns ) * 100. / totWords ) ;
    fprintf( fpOut, "\tsentences        : %d\n", numSents ) ;
    fprintf( fpOut, "\tsentence errors  : %d\t%4.1f%%\n", numSents - sentCorr,
	     ( numSents - sentCorr ) * 100. / numSents ) ;

  }

  // print alignment data -----------------------------------------------------
  void printAlign( FILE *fpOut, sentence *recs, sentence *refs ) {

    // print the aligned refernce string
    fprintf( fpOut, "ref:\t" ) ; 
    for( int i = backCnt - 1 ; i >= 0 ; i-- ) {
      if( backTrace_col[i] != backTrace_col[i+1] ) {
	fprintf( fpOut, "%s\t", refs->getWord( backTrace_col[i] - 1 ) ) ;
      }
      else {
	fprintf( fpOut, "<ins>\t" ) ;
      }
    }
    fprintf( fpOut, "\n" ) ;

    // print the aligned recognized string
    fprintf( fpOut, "rec:\t" ) ; 
    for( int i = backCnt - 1 ; i >= 0 ; i-- ) {
      if( backTrace_row[i] == 0 ) {
	fprintf( fpOut, "<del>\t" ) ;
      }
      else if( backTrace_row[i] != backTrace_row[i+1] ) {
	fprintf( fpOut, "%s\t", recs->getWord( backTrace_row[i] - 1 ) ) ;
      }
      else {
	fprintf( fpOut, "<del>\t" ) ;
      }
    }
    fprintf( fpOut, "\n\n" ) ;

  }
} ;
#endif
